let family_name = "Cyclic automata"
let description = "States are arranged in a cycle. A cell updates its state to the next in the cycle if the number of occurrences of this next state is above some threshold."

open Acml

exception Pb

let _ = add_shortcut "cyclic" (fun _ ->
  ["radius","1"; "threshold","3"; "states", "3"; "dimension", "2"; "neighborhood","Moore"; "Greenberg-Hastings","false"]
)

let _ = declare_object "cyclic"

let _ = add_shortcut "^R[0-9]+/T[0-9]+/C[0-9]+/N[NM]\\(/GH\\)?" 
  (fun s -> let ll = Str.split (Str.regexp "/") s in
     (match ll with 
       | r::(t::(c::(n::q))) -> let iofs p = (String.sub p 1 (String.length p - 1)) in
	   ["radius", (iofs r); "threshold", (iofs t); "states", (iofs c); "dimension", "2";
	    "neighborhood", (match n with "NM" -> "Moore" | _ -> "vonNeumann");
	    "Greenberg-Hastings", (if List.length ll = 5 then "true" else "false")]
       | _ -> raise Pb
     )
  )

let maker a = (
  let (d,n,r,neig) = std_params ~states:3 ~neighborhood:"Moore" ~radius:1 ~dimension:2 a in
  let e = param a "threshold" 0 300 3 in
  let gh = bool_param a "Greenberg-Hastings" false in
    set_multiset_rule a (fun c x -> 
			   if gh && c>0 then (c+1) mod n
			   else if x.((c+1) mod n) >= e then (c+1) mod n 
			   else c);
    set_name a family_name;
    add_description a "Rule family" description;
    doc_param a "threshold" "Number of occurrences of 'next state' in the neighborhood above which the cell state is updated";
    doc_param a "Greenberg-Hastings" "Restrict the threshold condition to state 0. For any other state, the value is updated to next state systematically."
)

let _ = valid maker

  
  
