let name = "Elementary Cellular Automata"
let description = "All one-dimensional two-states radius 1 cellular automata"

open Acml

let eca i t =
  if ((1 lsl (4*t.(0)+2*t.(1)+t.(2))) land i) = 0 then 0 else 1

let codes = [|"A"; "B"; "E"; "F"; "C"; "D"; "G"; "H"|]
let order = [|0;1;4;5;2;3;6;7|]

let _ = add_shortcut "^[0-9][0-9]?[0-9]?$" (fun s -> ["define by", "number"; "number", s; "second order", "false"])
let _ = add_shortcut "^[0-9][0-9]?[0-9]?R$" (fun s -> ["define by", "number"; "number", String.sub s 0 (String.length s - 1); "second order", "true"])
let _ = add_shortcut "^[ABCDEFGH]+$" (fun s -> ["define by", "t-code"; "t-code", s])

let maker a =
  (
    set_dimension a 1;
    set_moore_neighbourhood a 1;
    set_name a name;
    add_description a "Family of rules" description;
    let thebit i c = (i land (1 lsl c)) lsr c in
    let name_trans i = (string_of_int (thebit i 2))^(string_of_int (thebit i 1))^(string_of_int (thebit i 0)) in
    let show_trans v i = param a (name_trans i) 0 1 v in
    let rec set_all_trans w i = if i >= 0 then (
      set_param a (name_trans i) (thebit w i); set_all_trans w (i-1)) in
    let tab = Array.init 8 (fun i -> show_trans (ca_random 2) i) in 
    let st = choice_param a "define by" ["number"; "t-code"; "transitions"] in
    let w = param a "number" 0 255 (ca_random 256) in
    let tcod = string_param a "t-code" "BDE" in
    let alpha = float_param a "alpha" 0. 1. 1. in
    let is_active i rul = eca rul [|thebit i 2; thebit i 1; thebit i 0|] <> thebit i 1 in 
    let tcode_from_rule rul =
      let rec doit acc i =
	if i = 8 then acc else if is_active order.(i) rul then doit (acc^codes.(order.(i))) (i+1) else doit acc (i+1)
      in doit "" 0 in
    let rule_from_tcode cod =
      let rec isin l s i = if i = String.length s then false else if String.sub s i 1 <> l then isin l s (i+1) else true in
      let rec doit acc i =
	if i=8 then acc else 
	  let up = if isin codes.(i) cod 0 then 1 - (thebit i 1) else thebit i 1 in
	  doit (up*(1 lsl i) + acc) (i+1)
      in doit 0 0
    in
    let thenum = match st with 
	"number" -> (set_all_trans w 7; set_string_param a "t-code" (tcode_from_rule w); w)
      | "t-code" -> (let w = rule_from_tcode tcod in set_all_trans w 7; set_param a "number" w; w)
      | _ -> let r = Array.fold_right (fun x acc -> 2*acc +x) tab  0 in set_param a "number" r; set_string_param a "t-code" (tcode_from_rule w); r
    in
    let rec nl i = if i>=0 then (let nm = name_trans i in doc_param a nm ("Result of transition "^nm); 
				 nm)::(nl (i-1)) else [] in
    group_param a "Transitions" (nl 7);
    let sndorder = bool_param a "second order" false in
    let r = if sndorder then "(R)" else "" in
    let basename = (string_of_int thenum)^r^" / "^(get_string_param a "t-code")^r in
    set_cardinality a (if sndorder then 4 else 2);
    let thiswillbethecorerule =
      if sndorder then (fun x -> 
	x.(centercell a)/2 + 2*((x.(centercell a) + eca thenum (Array.map (fun a -> a / 2) x)) mod 2)
      )
      else eca thenum
    in
    if alpha = 1. then (
      set_rule a thiswillbethecorerule;
      set_object_name a basename;
    ) else (
      declare_stochastic a;
      set_rule a (fun x -> if ca_random_float 1. < alpha then  thiswillbethecorerule x else x.(centercell a));
      set_object_name a (basename^" - asynchronous");
    );
    if sndorder then add_init_conf a ~set:true "Second order valid" (General (fun i j -> 2*(conf_random 2)));
    doc_param a "alpha" "Rate of synchronism. 1 means a synchronous and deterministic CA";
    doc_param a "second order" "Define the 4-states second order rule instead of the standard 2-states one";
    doc_param a "number" "Wolfram number of the rule";
    doc_param a "t-code" "T-code of the rule: list of transitions that change the value of the central cell, transitions are named A=000,B=001,C=100,D=101,E=010,F=011,G=110,H=111";
    doc_param a "define by" "Whether Wolfram number or transition list will be used to define the new rule";
  )
    
let _ = valid maker
