let name = "The cellular fire effect"
let description = "This is a deterministic discrete directional diffusion rule with a random update of the bottom line."

open Acml

let maker a = (
  let (d,n,r,neighb) = std_params ~dimension:2 ~states:500 ~neighborhood:"Moore" a in
  let acc=ref 0 in
  let dividvon = r+(2*r+1) in
    declare_deterministic a;
    set_rule a (fun x -> 
		  acc:=0;
		  for i=(-r) to r do
		    acc:=!acc + x.(relativecell a i (-1))
		  done;
		  acc:=!acc + x.(relativecell a 0 0);
		  for j = 2 to r do
		    acc:=!acc + x.(relativecell a 0 (-j))
		  done; (!acc / dividvon);
	       );
    add_init_conf a ~set:true "Fire conf" (General (fun i j -> if j=0 then (n-1) else 0));
    let (_,h) = get_universe_size a in
      set_name a name;
      add_description a "General idea" description;
      add_description a "Local rule" "The new value of a cell is the average of the neighbouring cells which are not positionned above it.";
      add_description a "Fire kept alive" "At each time step some 'hot spots' are generated at the bottom line of the configuration.";
      add_colormap a  ~set:true "Fire colormap" (color_gradient a [|(0,0,0);(250,0,0);(250,250,0); (250,250,250)|]);
      add_view a "Clean fire view" (fun x y read -> 
					  if y > 10 then
					    color_gradient a [|(0,0,0);(250,0,0);(250,250,0); (250,250,250)|] (read x (h-1-y))
					  else Unicolor (y*7,y*7,y*7)
				       );
      let sw = param a "source power" 1 100 3 in
      let failure = bool_param a "random failures" true in
      let ffreq = param a "mean failure interval" 1 5000 10 in
      let pmotion = param a "random shift" 0 100 10 in
	if (bool_param a "keep alive" true) then add_measurement a "Keep fire alive" 
	  (fun t x y read write ->
	     let fail = if failure then (if conf_random ffreq = 0 then 0 else (n-1)) else (n-1) in
	     let middle =  x/2 + (conf_random ((x*pmotion)/100)) - (x*pmotion)/200 in
	     let source i = min fail ((n-1)*(sw*x/10-(abs (i-middle))*(abs (i-middle)))) in
	       for i=0 to x-1 do
 		 write i 0 (min (((n-1)*(min t 100))/100) (max 0(* (conf_random (max 1 (n/10))) *) (source i)));
	       done
	  );
	doc_param a "source power" "How large are the 'hot spots' generated at each step.";
	doc_param a "random failures" "Whether the 'hot spots' generation fails from time to time.";
	doc_param a "mean failure interval" "The expected number of steps between two failures.";
	doc_param a "random shift" "How far the 'hot spots' can be placed from the middle of the bottom line.";
	doc_param a "keep alive" "Whether 'hot spots' generation is on.";
)
  
  
let _ = valid maker
