let family_name =  "Generations CA"
let description = "Rules are like life birth/survival rules but with history: cells are getting older instead of dying directly."

open Acml

exception DecodeFailure

let _ = add_shortcut "^[Bb]rian\\('s\\)?[- ][Bb]rain$" (fun s -> ["define by", "Bx/Sy code"; "states", "3"; 
								  "radius", "1"; "code", "B2/S"])
let _ = add_shortcut "^[0-9]*/[0-9]*/[0-9]+$" (fun s -> (match Str.split_delim (Str.regexp "/") s with
							     [su;bi;st] -> ["define by", "Bx/Sy code"; "states", st; 
									      "radius", "1"; "code", (Printf.sprintf "B%s/S%s" bi su)]
							   | _ -> raise DecodeFailure
 							))

let read_code code =
  if Str.string_match (Str.regexp "^B[0-9]*/S[0-9]*$") code 0 then
    let rec advance in_birth bacc sacc s =
      if String.length s = 0 then (bacc,sacc)
      else if Str.string_match (Str.regexp "[0-9]") s 0 then
	try
	  let n = int_of_string (Str.first_chars s 1) in
	    if in_birth then advance in_birth (n::bacc) sacc (Str.string_after s 1)
	    else advance in_birth bacc (n::sacc) (Str.string_after s 1)
	with _ -> raise DecodeFailure
      else advance false bacc sacc (Str.string_after s 1) in (
	advance true [] [] (Str.string_after code 1)
      )
  else raise DecodeFailure

let produce_code n birth survival =
  let grow i t acc = if t.(i) then acc^(string_of_int i) else acc in
  let rec doit bacc sacc i = 
    if i >= Array.length birth then ("B"^bacc^"/S"^sacc)
    else doit (grow i birth bacc) (grow i survival sacc) (i+1)
  in doit "" "" 0
      

let maker a =
  (
    let n = param a "states" 1 500 3 in set_cardinality a n;
      set_dimension a 2;
      let r = param a "radius" 1 50 1 in set_moore_neighbourhood a r;
	set_name a family_name;
	add_description a "Rules" description;
	let birth = Array.make (arity a) false in
	let survival = Array.make (arity a) false in
	let bl = [2] in
	let sl = [] in
	let st = choice_param a "define by" ["Bx/Sy code"; "transitions"] in
	let rec add_param myfun acc tab str list i = if i >= 0 then (
	  let nam = str^(string_of_int i) in
	    tab.(i) <- myfun a nam (if List.mem i list then true else false);
	    add_param myfun (nam::acc) tab str list (i-1)
	) else acc in
	  group_param a "Birth" (add_param bool_param [] birth "birth if " bl (arity a -1));
	  group_param a "Survival" (add_param bool_param [] survival "survive if " sl (arity a -1));
	  let code = string_param a "code" "B2/S" in
	    (match st with
		 "Bx/Sy code" ->  
		   let (bili, suli) = try 
		     read_code code
		   with DecodeFailure -> (bl,sl)
		   in (
		       ignore (add_param (fun u d t -> set_bool_param u d t; t) [] birth "birth if " bili (arity a -1));
		       ignore (add_param  (fun u d t -> set_bool_param u d t; t) [] survival "survive if " suli (arity a -1));
		     )
	       | _ -> let toto =  (produce_code n birth survival) in
		   set_string_param a "code" toto
	    );
	    declare_deterministic a;
 	    doc_param a "Bx/Sy code" "x=list of numbers of cells in state 1 for a dead cell to become 1; y= list of numbers of cells of the same age for the current cell to keep unchanged";  
	    if get_string_param a "code" = "B2/S" && get_param a "states" = 3 && get_param a "radius" = 1 then
	      set_object_name a "Brian's Brain";
	    set_multiset_rule a (fun state x -> 
				   let s = x.(1) + (if state=1 then -1 else 0) in 
				     if (state=0 && birth.(s)) then 1
				     else if (state>0 && survival.(s))  then state
				     else if (state>0) then ((state+1) mod (cardinality a))
				     else 0
				)
  )
    
let _ = valid maker


