let name = "Langton's Ant"
let description = "The land is made of cells which can be empty or full (of some pheromone). Moreover, an ant can be there with a given orientation (N,S,E or W).\n  When the ant is on some cell in some orientation, it goes forward one cell in its direction and then turns 90° right or left depending on whether the cell it comes from was empty or not (pheromone). Moreover, it changes the state (pheromone) of the cell it comes from."

open Acml;;

type ant =
    L
    | R
    | U
    | D

type state =
    Empty
    | Pheromone

type set =
    Nothere of state
    | Ishere of state*ant

let name_dir = function
  | L -> "left"
  | R -> "ritght"
  | U -> "top"
  | D -> "bottom" 

let name_p = function
  | Empty -> "nothing"
  | Pheromone -> "pheromone"

let nameof = function
  | Ishere (p,d) -> Printf.sprintf "Ant pointing to the %s over %s" (name_dir d) (name_p p)
  | Nothere p -> name_p p


let stts = [|Nothere Empty; Nothere Pheromone;
			    Ishere (Empty,L); Ishere (Pheromone,L);
			    Ishere (Empty,R); Ishere (Pheromone,R);
			    Ishere (Empty,U); Ishere (Pheromone,U);
			    Ishere (Empty,D); Ishere (Pheromone,D);
			  |]

let alphabet = AlphaArray stts 

let _ = add_shortcut "Ant" (fun _ -> []); declare_object "Ant"


let maker a =
  (
    set_name a name;
    add_description a "The rule" description;
    set_dimension a 2;
    set_vonneumann_neighbourhood a 1;
    let local x =
      match x.(centercell a) with
	  Ishere (s,_) -> Nothere (if s=Empty then Pheromone else Empty)
	| Nothere s -> (match (x.(relativecell a (-1) 0), x.(relativecell a 1 0), 
			       x.(relativecell a 0 (-1)),x.(relativecell a 0 1)) with
			    (Nothere _,Nothere _,(Ishere (Empty,R)),Nothere _) -> (Ishere (s,U))
			  | (Nothere _,Nothere _,(Ishere (Pheromone,L)),Nothere _) -> Ishere (s,U)
			  | (Nothere _,Nothere _,Nothere _,Ishere (Pheromone,R)) -> Ishere (s,D)
			  | (Nothere _,Nothere _,Nothere _,Ishere (Empty,L)) -> Ishere (s,D)
			  | (Ishere (Pheromone,U),Nothere _,Nothere _,Nothere _) -> Ishere (s,R)
			  | (Ishere (Empty,D),Nothere _,Nothere _,Nothere _) -> Ishere (s,R)
			  | (Nothere _,Ishere (Pheromone,D),Nothere _,Nothere _) -> Ishere (s,L)
			  | (Nothere _,Ishere (Empty,U),Nothere _,Nothere _) -> Ishere (s,L)
			  | _ -> Nothere s
		       ) in
    poly_set_rule a ~alphabet ~active:(function Ishere _ -> true | _ -> false) local;
    Array.iter (fun x -> poly_set_state_name a ~alphabet x (nameof x)) stts;
      let rep_st = function 
	  Pheromone -> Unicolor (50,250,50)
	| _ -> Unicolor (50,50,50)
      in
      let rep_an = function 
	  L  -> WestArrow
	| R  -> EastArrow
	| U  -> NorthArrow
	| D  -> SouthArrow
      in
      let st_change = function
	  Empty -> Pheromone
	| _ -> Empty
      in
      let ant_change = function
	  L -> U
	| U -> R
	| R -> D
	| D -> L
      in
	poly_add_colormap a ~set:true "Ant coloring" ~alphabet 
	  (function
	       Nothere s -> rep_st s
	     | Ishere (s,ant) -> Multiple [rep_st s; rep_an ant]
	  );
	poly_add_modif a "Blank" ~alphabet (Forced (Nothere Empty));
	poly_add_modif a ~set:true ~left:true "Random Pheromon" ~alphabet 
	  (Map (function
	       Nothere x -> Nothere (if conf_random 2 = 1 then st_change x else x)
	     | Ishere (x,a) -> Ishere ((if conf_random 2 = 1 then st_change x else x), a)
	  ));
	poly_add_modif a ~set:true ~left:true "Pheromon" ~alphabet 
	  (Map (function
	       Nothere x -> Nothere (st_change x)
	     | Ishere (x,a) -> Ishere (st_change x, a)
	  ));
	poly_add_modif a "Ant" ~set:true ~left:false ~alphabet
	  (Map (function
	       Nothere x -> Ishere (x,L)
	     | Ishere (x, D) -> Nothere x
	     | Ishere (x,a) -> Ishere (x, (ant_change a))
	  ));
	poly_add_init_conf a ~alphabet ~set:true "Solo Ant in void land" 
	  (General 
	     (let x,y = get_universe_size a in 
		(fun a b -> if a = x/2 && b = y/2 then Ishere (Empty,U) else Nothere Empty))
	  );
  )

let _ =  valid maker
    
