let family_name = "Life-like birth/survival CA"
let description = "These are two states (dead/alive) rules with Moore neighborhood specified by two sets B (birth) and S (survive).\n A dead cell becomes alive if the number of neighboring alive cells is in B. An alive cell becomes dead if the number of neighboring dead cells is not in S. In any other case, the cell is unchanged."

open Acml

let additionnal_names = [
  "anneal","B4678/S35678";
  "life","B3/S23";
] 

exception DecodeFailure

let _ = add_shortcut "^[0-9]*/[0-9]*$" (fun s -> (match Str.split_delim (Str.regexp "/") s with
							     [su;bi] -> ["define by", "Bx/Sy code";
									 "code", (Printf.sprintf "B%s/S%s" bi su)]
							   | _ -> raise DecodeFailure
 							))

let _ =
  List.iter
    (fun (x,y) -> 
       let rege = ("^"^(Str.global_replace (Str.regexp " ") "[ -]?" x)^"$") in
	 add_shortcut rege (fun s -> ["define by", "Bx/Sy code"; "code", y]);
	 declare_object x
    )
    additionnal_names;
  add_shortcut "^B[0-9]*/S[0-9]*$" (fun s -> ["define by", "Bx/Sy code"; "code", s])


let read_code code =
  if Str.string_match (Str.regexp "^B[0-9]*/S[0-9]*$") code 0 then
    let rec advance in_birth bacc sacc s =
      if String.length s = 0 then (bacc,sacc)
      else if Str.string_match (Str.regexp "[0-9]") s 0 then
	try
	  let n = int_of_string (Str.first_chars s 1) in
	    if in_birth then advance in_birth (n::bacc) sacc (Str.string_after s 1)
	    else advance in_birth bacc (n::sacc) (Str.string_after s 1)
	with _ -> raise DecodeFailure
      else advance false bacc sacc (Str.string_after s 1) in
      advance true [] [] (Str.string_after code 1)
  else raise DecodeFailure

let produce_code birth survival =
  let grow i t acc = if t.(i) then acc^(string_of_int i) else acc in
  let rec doit bacc sacc i = 
    if i >= Array.length birth then ("B"^bacc^"/S"^sacc)
    else doit (grow i birth bacc) (grow i survival sacc) (i+1)
  in doit "" "" 0
      

let maker a =
  (
    set_cardinality a 2;
    set_dimension a 2;
    let st = choice_param a "define by" ["Bx/Sy code"; "transitions"] in
      set_moore_neighbourhood a 1;
      set_name a family_name;
      add_description a "General idea" description;
      let birth = Array.make (arity a) false in
      let survival = Array.make (arity a) false in
      let bl = [3] in
      let sl = [2; 3] in
      let code = string_param a "code" "B3/S23" in
      let rec add_param myfun acc tab str list i = if i >= 0 then (
	let nam = str^(string_of_int i) in
	  tab.(i) <- myfun a nam (if List.mem i list then true else false);
	  add_param myfun (nam::acc) tab str list (i-1)
      ) else acc in
      let do_by_code code =
	let (bili, suli) = try 
	  read_code code
	with DecodeFailure -> (bl,sl)
	in (
	    ignore (add_param (fun u d t -> set_bool_param u d t; t) [] birth "birth if " bili (arity a -1));
	    ignore (add_param  (fun u d t -> set_bool_param u d t; t) [] survival "survive if " suli (arity a -1));
	  )
      in
	group_param a "Birth" (add_param bool_param [] birth "birth if " bl (arity a -1));
	group_param a "Survival" (add_param bool_param [] survival "survive if " sl (arity a -1));
	(match st with
	     "Bx/Sy code" ->  do_by_code code
 	   | _ -> let toto =  (produce_code birth survival) in (
	       set_string_param a "code" toto;
	     )	       
	);
	declare_deterministic a;
	set_object_name a code;
	set_multiset_rule a (fun state x -> 
			       let s = max 0 (x.(1) - state) in 
				 if (state=0 && birth.(s)) || (state=1 && survival.(s))  then 1
				 else 0
			    )
  )
    
let _ = valid maker


