let family_name = "Larger Than Life"
let description = "These are two states (dead/alive) rules with arbitrary radius specified by two intervals B (birth) and S (survive).\n A dead cell becomes alive if the number of neighboring alive cells is in B. An alive cell becomes dead if the number of neighboring dead cells is not in S. In any other case, the cell is unchanged."

open Acml;;

let uple_read s =
  Scanf.sscanf s "(%u,%u,%u,%u,%u)" (fun a b c d e -> (a,b,c,d,e))

let names = 
  [("bosco", "(5,34,45,34,58)");
   ("globe", "(8,74,252,163,223)");
   ("bugsmovie","(10,123,170,123,212)");
   ("waffle","(10,100,200,75,170)")
  ]

let _ =
  List.iter
    (fun (x,y) -> 
       let rege = ("^"^(Str.global_replace (Str.regexp " ") "[ -]?" x)^"$") in
	 declare_object x;
	 add_shortcut rege (fun s -> ["define by", "5-uple"; "5-uple", y]))
    names

let _ = declare_object "bugs"; add_shortcut "bugs" (fun s -> ["define by", "name"; "name", "bosco"])

let default_name = "Anonymous"

let my_name code =
  let rec doit = function
      t::q -> if snd t = code then fst t else doit q
    | [] -> default_name
  in
    doit names

let maker a =
  (
    let r = param a "radius" 1 50 1 in
      set_cardinality a 2;
      set_dimension a 2;
      let st = choice_param a "define by" ["5-uple"; "name"; "intervals"] in
      let uple = string_param a "5-uple" "(5,34,45,34,58)" in
      let name = choice_param a "name" (default_name::(List.map fst names)) in
	set_name a family_name;
	let borninf = param a "min to born" 0 (arity a) 3 in
	let bornsup = param a "max to born" 0 (arity a) 3 in
	let liveinf  = param a "min to survive" 0 (arity a) 2 in
	let livesup  = param a "max to survive" 0 (arity a) 3 in
	let check x sum i s = if sum<i || sum>s then x else 1-x in
	let do_by_code s = 
	  let (ra,b,c,d,e) = try
	    uple_read s
	  with _ -> (5,34,45,34,58)
	  in
 	    set_param a "radius" ra;
	    set_param a "min to born" b;
	    set_param a "max to born" c;
	    set_param a "min to survive" d;
	    set_param a "max to survive" e;
	    set_moore_neighbourhood a ra;
	    ra,b,c,d,e	      
	in
	let (r,bi,bs,li,ls) = (
	  match st with
	      "intervals" -> ( 
		let new_uple = (Printf.sprintf "(%d,%d,%d,%d,%d)" r borninf bornsup liveinf livesup) in
		  set_string_param a "5-uple" new_uple;
		  set_choice_param a "name" (my_name new_uple);							 
		  set_moore_neighbourhood a r;
		  r,borninf,bornsup,liveinf,livesup;
	      )
	    | "name" -> (
		try
		  let code = List.assoc name names in
		    set_string_param a "5-uple" code;
		    do_by_code code
		with Not_found -> do_by_code uple
	      )
	    | _ ->  (
		set_choice_param a "name" (my_name uple);							 
		do_by_code uple
	      )
	) in
 	  declare_deterministic a;
	  set_object_name a (my_name (get_string_param a "5-uple"));
	  set_multiset_rule a (fun state x -> 
				 let s = x.(1) in
				   if state=0 then check state s bi bs
				   else check (1-state) s li ls
			      );
	  group_param a "Intervals" ["min to born"; "max to born"; "min to survive"; "max to survive"];
	  doc_param a "5-uple" "K. M. Evans' notation as a 5-uple: (radius, min to born, max to born, min to survive, max to survive). Count of living neighbours includes the cell itself."
  )

let _ =  valid maker
