open Acml
open Num

let rule_from_number a n reverse number x =
  let m = Array.length x -1 in
  let access = (fun i -> x.(m-i)) in
  let rec conf_array_to_int acc i =
    if i<0 then acc
    else conf_array_to_int ((Int (access i))+/(Int n)*/acc) (i-1)
  in 
  let num_of_conf = 
    let right = conf_array_to_int (Int 0) m in
      if reverse then 
	let max = ((Int n) **/ (Int (m+1))) -/ (Int 1) in power_num (Int n) (max -/ right)
      else power_num (Int n) right
  in
    int_of_num (mod_num (quo_num number num_of_conf) (Int n))


let char_to_int = function
'0' -> num_of_int 0
| '1' -> num_of_int 1
| '2' -> num_of_int 2
| '3' -> num_of_int 3
| '4' -> num_of_int 4
| '5' -> num_of_int 5
| '6' -> num_of_int 6
| '7' -> num_of_int 7
| '8' -> num_of_int 8
| '9' -> num_of_int 9
| 'A' -> num_of_int 10
| 'B' -> num_of_int 11
| 'C' -> num_of_int 12
| 'D' -> num_of_int 13
| 'E' -> num_of_int 14
| _ -> num_of_int 15

let rec string_to_int base acc accbase i nb =
if i = -1 then acc
else string_to_int base ((char_to_int (String.get nb i))*/accbase +/  acc ) (accbase*/(num_of_int base)) (i-1) nb

let transform a =
  set_name a "Rule by number";
  let d = param a "dimension" 1 2 1 in
  let r = param a "radius" 1 50 1 in
  let n = param a "states" 2 100 2 in
  let base = param a "base" 2 16 10 in
  let reverse = bool_param a "reverse" false in
  let nb =  string_param a "number" "110" in
  let number = string_to_int base (num_of_int 0) (num_of_int 1) (String.length nb - 1) nb in
    declare_deterministic a;
    set_dimension a d;
    doc_param a "base" "the base in which the integer [number] is written";
    doc_param a "number" "when all possible transition tables are sorted lexicographically (see param [reverse]), this number gives the position of the desired one in the list";
    doc_param a "reverse" "when false (standard Wolfram notation), the least significant bits of [number] correspond to transition 0...0; when true, the most significant bits of [number] correspond to transition 0...0"; 
    set_cardinality a n;
    (match choice_param a "neighborhood" ["von Neumann"; "Moore"]  
     with "von Neumann" -> set_vonneumann_neighbourhood a r
       | _ -> set_moore_neighbourhood a r);
    set_rule a (rule_from_number a n reverse number);
    set_object_name a nb

let _ = valid transform

    
    
