let name = "Timer"

let description = "Add a measurement of the steps and cell update speeds. It is not very useful outside batch mode."

open Acml

let (raz, count, message) =
  let cnt = ref 0 in
  let ticks = ref 0. in (
      (fun () -> (cnt := 0; ticks := float_of_int (Sdltimer.get_ticks ()))),
      (fun () -> incr cnt),
      (fun size -> let nt = float_of_int (Sdltimer.get_ticks ()) in
	 Printf.printf "%.2f\t%.2f\n" ((float_of_int (!cnt *1000)) /. (nt-. !ticks)) ((float_of_int (!cnt*size))/.(1000.*.(nt-. !ticks)));
	 flush stdout
      )
    )

let maker a =
  add_measurement a "chronos" (fun t x y read write -> 
				 if t = 1 then raz ()
				 else if t = get_stop_time () then message (if dimension a = 2 then x*y else x)
				 else count ()
			      );
  doc_action a "chronos" "Measure the time interval between step 1 and stopping step. Then outputs the step rate (steps per second) and cell update rate (millions of cells per second)."

let _ = valid_plugin ~name ~description ~executed_at:EachNewObject_after maker
  
