let name = "Mirek's Cellebration names"
let description = "Recognize some CA names from the MCell online database."

open Acml

let data_cyclic = [
"Turbulent phase","R2/T5/C8/NM","When the threshold is high enough that wave fronts cannot wind, but still low enough that they can advance, CCA rules exhibit a chaotic equilibrium phase which combines the small-scale structure of failed cores with large-scale disordered fronts. Our animations give only a glimpse of this behavior, since the typical final length scale of several hundred cells means that large arrays are needed to support a viable steady state. Small systems inevitably fixate. Note how impossible it is to predict, until the very end, which color will predominate.\nA rule by David Griffeath.";
"Perfect","R1/T3/C4/NM","A particularly interesting excitable system. From a uniform random configuration it quickly self-organizes into a chaotic soup with large length scale. But later on, often after more than one hundred updates, perfect, widely separated stable spiral cores emerge and slowly take over the lattice.\nA rule by David Griffeath.";
"LavaLamp","R2/T10/C3/NM","The rule produces blobs that seperate and combine resembling a lava lamp.\nA rule by Jason Rampe.";
"Imperfect","R1/T2/C4/NM","From Elliott's description: \"A little magic carpet generator of cellsize 4. One thing that sets this rule apart from other CCAs I know of is a phenomenon that occurs in the occasional orbit, in which a garish living stain forms and grows till it blots out the entire carpet. This cancerous growth (to mix my metaphors) starts from a small anomalous patch of cells - though not inevitably, for some such tumors are benign. So, in a way this rule reverses the situation we encounter in its higher threshold cousin, Perfect. In the latter rule, regular spirals invade a \"broken\" but seemingly stable dynamic, whereas here the tables are turned and it is an orderly spiral regime that finds itself overthrown.\"\nA rule by John Elliott, May 2000.";
"GH","R3/T5/C8/NM/GH","A basic Greenberg-Hastings model.\nA rule by David Griffeath.";
"GH Percolation mix","R5/T10/C8/NM/GH","With larger neighbor sets and low thresholds, stable periodic cycles occur by chance in the initial random soup. Thus one encounters a mixture of partial self-organization and percolation effects.\nA rule by David Griffeath.";
"GH Macaroni","R2/T4/C5/NM/GH","The threshold is sufficiently high that wave fragments have a hard time bending inwards to make spiral cores. Consequently, cyclic bands self-organize over time into weakly aligned parades of wave fragments. The pattern is not unlike that of the monolayer of macaroni that adheres to a colander during the preparation of a $.059 box of macaroni and cheese. CAM6 experiments have revealed that if these dynamics are allowed to run for thousands of updates on a larger array, say 2048 by 2048, then the wave fragments continue to align into ever longer parades. But more than this, the pasta also merge end to end until they transform into a new variety: spaghetti. We expect that on an infinite array the process would continue to enormous length scales, giving rise to linguine, then angel hair, then ...\nA rule by David Griffeath.";
"Cubism","R2/T5/C3/NN","This interesting rule creates rectangular regions.\nA rule by Jason Rampe.";
"CCA","R1/T1/C14/NN","A basic CCA. Starting from a uniform random distribution over 14 colors, droplets of color waves nucleate fairly quickly. Soon virtually all of the initial \"debris\" are overrun by the droplets. As the last vestiges of debris are eliminated, vortices emerge from the disordered wave fronts, creating diamond-shaped spirals. By about time 300 the array is completely covered with periodic spirals, out of phase with one another and not all of minimal period 14. Typically it takes much longer for the period 14 spiral cores to displace their feebler competitors.\nA rule by David Griffeath.";
"Black vs White","R5/T23/C2/NN","Well-balanced rule where both death (black) and life (white) have equal chances. Seed patterns with 50% density.\nA rule by Jason Rampe.";
"Amoeba","R3/T10/C2/NN","Similar to Vote 4/5, this rule smoothes patterns and leaves irregular islands with characteristic oscillating edges.\nA rule by Jason Rampe.";
"3-color bootstrap","R2/T11/C3/NM","The term bootstrap refers to systems with local clusters which can only propagate by means of external support from random noise or other clusters. Such critical dynamics, and related near-critical rules, often fixate in complex 'fossilized' final configurations. This rule shows a 3-color CCA with competing bootstrap growth.\nA rule by David Griffeath.";
"313","R1/T3/C3/NM","\"This is the only three-color rule in the range-1 CCA rulespace that exhibits significant self-organization. And it's no slouch in that regard, either. It sports a number of distinct spiral types, its demographics are eminently watchable, and in general it has a flavor all its own.\" - J.Elliott\nA rule by David Griffeath.";
]

let data_generation = [
"Xtasy","1456/2356/16","From simple seeds in a closed universe this is a lovely kaleidoscopic wave rule that tends to do extended meditations on the letter X.\nA rule by John Elliott, July 2000";
"Worms","3467/25/6","A very vigorous rule producing crowds of disgusting worms.\nA rule by Mirek Wojtowicz.";
"Wanderers","345/34678/5","A rule very similar to \"Prairie on fire\", but creates interesting dynamically moving patterns.\nA rule by Mirek Wojtowicz.";
"TransersII","0345/26/6","A rule by Michael Sweney.";
"Transers","345/26/5","A beautiful and dynamic rule with many distinctive entities that occur \"in the wild\", with remarkable inter-entity transformations.\nA rule by John Elliott.";
"ThrillGrill","1234/34/48","Sparkling dynamics inhabit and rearrange a grill-like background of one-cells.\nA rule by John Elliott, July 2000.";
"Swirl","23/34/8","While spirals are common in CAs (see for example RainZha or Cyclic CAs), Swirl's whirlers are of the rarer discrete variety - isolated \"whirlpools\" in an otherwise dead sea. There are two sizes of whirlpools that occur regularly. Ladd calls the smaller the \"swirlpool\" and the larger the \"flamewhirl\". Also there is a glider with a flickering double tail which Ladd aptly refers to as a fire or flame glider.\nA rule by Scott Robert Ladd.";
"Sticks","3456/2/6","This interesting rule builds a variety of sticks, with ships floating among them.\nA rule by Rudy Rucker.";
"Star Wars","345/2/4","A very interesting and beautiful rule producing deep space battle scenes; a paradise for patterns' creators in the tradition of Conway's Life. It's a successful combination of famous Brian's Brain with a stabilizing factor.\n\"StarWars is also an interesting rule from the low-investment standpoint. In particular, it is notable for the abundance, intricacy, and variety of its naturally-occurring glider guns. In general when running \"au naturel\" StarWars is active to a degree almost reminiscent of Brain, with the distinction that it likes to build fixed lego-like skeletal structures. In some orbits this latter penchant reaches its zenith, and we get large universe-spanning \"shells\" possessing an intricate radiate geometry.\" - J.E.\nThe rule is also known as \"MirekGro\".\nA rule by Mirek Wojtowicz, March 1999.";
"Spirals","2/234/5","The rule produces beautiful stable spirals that unroll forever.\nAuthor unknown.";
"SoftFreeze","13458/38/6","\"The rule is delicately poised between \"solid\" and \"liquid\" phases. From many simple seeds it forms \"living crystals\" - the crystals form, melt, and reform seemingly indefinitely, in balance with an active phase. The areas of melting can shrink to a point where you're sure everything will freeze over, and then they stage a comeback.\nFrom random starts, this rule resembles a nice poised rule from Wojtowicz called \"Living on the Edge\", only this one is even more \"on the edge\". (BTW, though the rule definition looks as though I took LOTE as a starting point, I actually came at it from a different direction entirely, and only afterwards realized I had converged on the B3S345 core.)\" - J.E.\nA rule by John Elliott.";
"Snake","03467/25/6","This rule is a simple variation of the Worms rule. It contains two interesting oscillators, a 3*3 square and a shuttle, among its basic elements. By combining copies of the square, many beautiful oscillators can be generated. The rule also generates many puffers one of which generates a sequence of shuttles. The tail generated by the shuttle puffer resembles a snake.\nA rule by Brian Prentice.";
"SediMental","45678/25678/4","From a random seed state of 30% ones, the rule tends to form stable \"islands\", with active coastlines and fascinating inter-island \"commerce\". But an orbit's fate is quite sensitive to the starting percentage. Below 25% or so the world remains sparsely Brain-like, whereas somewhere above 30% the islands begin to accrete into larger \"landmasses\", so that by 40% you nearly always end up with a single huge \"continent\". As for simple seeds, they often lead to surprising results. And the rule clearly has great potential from the \"engineering\" standpoint.\nA rule by John Elliott.";
"Rake","3467/2678/6","This rule is a simple variation of the Worms rule and is characterized by the numerous rake puffers that it generates. Together with the puffers, it generates many ships and expanding wings. These elements can easily be arranged to collide and form ships and puffers of arbitrary complexity.\nA rule by Brian Prentice.";
"RainZha","2/23/8","The simplest \"Zhabotinsky\" style cellular automaton. It spontaneously generates spirals from an initial random pattern.\nA rule by Rudy Rucker and John Walker";
"Prairie on fire","345/34/6","This rule covers the lattice with fire. For best results use yellow-red cells color settings.\nA rule by Mirek Wojtowicz.";
"OrthoGo","3/2/4","This simple rule supports spontaneous orthogonally productive \"glider colliders\". But otherwise it seems fairly quiescent.\nJohn Elliott, December 1999";
"Nova","45678/2478/25","Very lovely rule characterized by vibrant yet stately streaming growth from solid seeds. The growth patterns and active border have a fascinating geometry, the particulars of which are sensitive to the size and shape of the initial solid.\nA rule by Mirek Wojtowicz.";
"Meteor Guns","01245678/3/8","Almost any design above a certain level of complexity results in a slowly exploding ball of fire which then spits fairly large numbers of generational gliders \"meteors\".There are easily discovered stable oscillators and static patterns. A truly satisfying environment (if only finding a limited growth meteor gun were as easy...).\nA rule by Charles A. Rockafellor.";
"LivingOn TheEdge","345/3/6","LOTE. In this very chaotic rule it's hard to tell if patterns will survive or die out.\nA rule by Mirek Wojtowicz, February 2000";
"Lines","012345/458/3","The rule quickly self-organizes into linear structures, with separate orthogonally hatched areas. To see most interesting behaviour seed the universe with a circle with r >= 40 and a density of 20%.\nA rule by Anders Starmark, September 2000";
"Lava","12345/45678/8","All patterns are getting very slowly filled with hot lava.\nA rule by Mirek Wojtowicz.";
"Glissergy","035678/245678/5","\"This rule differs from Glisserati only in its cellsize (5 rather than 7). As would be expected, this change is an activity booster - so much so, that, unlike Glisserati, this rule has long transients from random as well as simple starts. Think of it as the pin-iferous answer to Brian's Brain.\" - J.E.\nA rule by John Elliott, May 2000";
"Glisserati","035678/245678/7","\"From simple seeds this rule is often highly glideriferous, generating striking kaleidoscopic patterns. Sometimes, though, as in the Glidathon orbit, it is more sparse, in the vein of rules such as Transers, but with much longer transients, providing an uncanny sense of \"perpetual motion\".\nSince I found this rule I've become aware of other recently discovered kaleidoscopic glider rules, such as Snake and TransersII, which share many of the same objects, though I've yet to see a Glidathon style orbit in them (which doesn't mean they aren't there - no one's really looked, AFAIK).\nIn the present version of Glisserati, orbits seeded randomly, as opposed to simply, collapse almost immediately, but that's not the case with the C=5 variation, which is very delicately poised between expansion and contraction. So I've spun that one off as 'Glissergy'.\" - J.E.\nA rule by John Elliott, May 2000.";
"Frozen spirals","356/23/6","The rule features a popular Zhabotinsky reaction, but with one distinction - spirals have icicles in their vertices.\nA rule by Mirek Wojtowicz.";
"Frogs","12/34/3","The name comes from very commonly occurring frog-like gliders.\nA rule by Scott Robert Ladd.";
"Flaming Starbows","347/23/8","In this world beautiful expanding diamonds and oddly off-parallelograms sweep across the screen leaving trails of fire and spirals of autogenesis. Reminiscent of a mutant spawn of Faders and mad Zhabotinsky spirals.\nA rule by Charles A. Rockafellor.";
"Fireworks","2/13/21","Nearly perfect fireworks simulation. It produces interesting results from both random and prepared initial states. Even several cells scattered over the lattice will produce long-running ravishing pictures.\nA rule by John Elliott";
"Faders","2/2/25","A \"genetic\" cross between Life and Brian's Brain. A dead Faders cell requires exactly 2 firing neighbors to get turned on. A firing Faders cell keeps firing if it has exactly 2 firing neighbors. And when a Faders cell leaves the firing state it goes into a sequence of refractory states.\nA rule by Rudy Rucker and John Walker";
"Ebb and Flow II","012468/37/18","A close variant of Ebb&amp;Flow that supports simple gliders.\nA rule by Michael Sweney.";
"Ebb and Flow","012478/36/18","A slow exploding rule with interesting internal dynamics within the growth boundaries. Different types of symmetries will often determine whether a pattern grows indefinitely or dies out. Presents challenging engineering problems. Best viewed with MCell standard color palette at speeds of 10 or slower.\nA rule by Michael Sweney.";
"Cooties","23/2/8","A very vivacious rule, with cells crawling around like mad lice.\nA rule by Rudy Rucker.";
"Circuit Genesis","2345/1234/8","Circuitry which evolves far beyond its programming and continues to do so long after the initial intput have taken the borders elsewhere.\nA rule by Charles A. Rockafellor.";
"Chenille","05678/24567/6","Author unknown.";
"Caterpillars","124567/378/4","The name comes from very commonly occurring caterpillars-like patterns.\nA rule by Mirek Wojtowicz.";
"BurstII","235678/3468/9","Basically it is Burst without the single cell still life. It supports slow and fast spaceships, oscillators, and methuselahs as well as a very characteristic p500+ oscillator.\nA rule by Michael Sweney.";
"Burst","0235678/3468/9","A generations rule related to Nova, Prairie Fire and others which features large areas of cells that slowly die out or decay into oscillators. Named after a characteristic exploding oscillator. Similar to Day&amp;Night in that constructions can be built both with inside and outside oscillations and growth/decay. Has still lifes similar to Life.\nA rule by Michael Sweney.";
"Brian's Brain","/2/3","Brian's Brain (BB, Brain), unquestionably one of the best known and most beautiful CA rules. If we name the possible cell values based on a simplistic neural analogy, viz. 0 = \"ready\", 1 = \"firing\", 2 = \"refractory\", then this simple rule can be stated thusly: Only a cell in the ready state may fire and it will only do so if exactly 2 of its neighbors are firing. After firing for one step, a cell spends a step in the refractory state before regaining readiness.\nAlthough the rule is relatively old, it has never been systematically explored. The rule is so live that it's not easy to construct stable patterns in it. In December 1999 Michael Sweney discovered first BB oscillators.\nA rule by Brian Silverman.";
"Brain 6","6/246/3","An interesting variation of immortal Brian's Brain, with many patterns constructed.\nA rule by Michael Sweney.";
"Bombers","345/24/25","A close variant of StarWars rule with much more history steps, what results in beautiful ships, puffers and oscillators.\nA rule by Mirek Wojtowicz.";
"Bloomerang","234/34678/24","A wave rule that rebounds into a kinder, gentler form of kaleidoscopics - rounded, soft, full, and slow.\nA rule by John Elliott, July 2000";
"BelZhab Sediment","145678/23/8","The rule creates BelZhab like patterns that have growing regions of \"sedimentary deposits\".\nA rule by Jason Rampe.";
"BelZhab","23/23/8","Another case of Zhabotinsky reaction.\nAuthor unknown.";
"Banners","2367/3457/5","The rule often produces ships with growing fluttering banners.\nA rule by Mirek Wojtowicz.";
]

let data_life = [
"WalledCities","2345/45678","The rule creates walled cities of activity. Once the field has stabilized, one can draw lines to connect the cities and the patterns expand to create an even larger city. But once the wall is complete, the city never grows, even though there is near-random activity inside it.\nA rule by David Macfarlane.";
"Stains","235678/3678","Most close variations of these rules expand forever, but this one curiously does not. Why?\nAuthor unknown.";
"Serviettes","/234","Like /2, every living cell dies every generation. This rule is picked for the exceptional fabric-like beauty of the patterns that it produces.\nAuthor unknown.";
"Seeds (2)","/2","Every living cell dies every generation, but most patterns explode anyway. It's a challenge to build new patterns that don't explode. Arguably the simplest challenging rule.\nA rule by Brian Silverman.";
"Replicator","1357/1357","In this remarkable universe every pattern is a replicator. After 32 steps every starting pattern is replicated 8 times.\nAuthor unknown.";
"Pseudo life","238/357","In this close variation of Conway's Life, the chaos is remarkably similar, but almost none of the engineered patterns work.\nAuthor unknown.";
"Move","245/368","A very calm universe, which nonetheless has a very commonly occurring slow spaceship and a slow puffer.\nAuthor unknown.";
"Mazectric","1234/3","\"Mazectric\" and \"Corrosion of Conformity\". An interesting variation of the Maze rule which produces longer halls and a highly linear format. Adding B7 to maze (keeping S5) allows some \"mice\" to run back and forth in the halls. Switching the B3 to B45 though, electrifies the mazes. Dropping S3 gives \"Corrosion of Conformity\", a slow burn from almost any starting pattern, resulting in a rusting away of the local continuum.\nA rule by Charles A. Rockafellor.";
"Maze","12345/3","An \"a-maze-ing\" universe - crystallizes into maze-like patterns. Interesting variations: try removing 5 from the \"Survival\" list. To produce mice running in the maze, add 7 to the \"Births\" list.\nAuthor unknown.";
"Long life","5/345","This rule is called \"Long life\" because of the extremely high period patterns that can be produced in this universe.\nA rule by Andrew Trevorrow.";
"InverseLife","34678/0123478/2","The rule shows similar oscillators and gliders to GOL, but dead cells create the patterns amongst live cells in the background.\nA rule by Jason Rampe.";
"HighLife","23/36","This rule is very similar to Conway's Life, but it has a surprise replicator pattern. There is no known replicator in Conway's Life.\nA rule by Nathan Thompson.";
"Gnarl","1/1","A simple rule provided by Kellie Evans. To see its beauty start with simple patterns, for example with a single dot.";
"Flakes","012345678/3","Also known as Life without Death (LwoD).\nThe rule produces beautiful flakes, starting from simple groups of cells. Try for example various filled circles with radius > 20 cells. The rule produces also ladders, what allowed David Griffeath and Cris Moore to prove that the rule is P-complete.\nA rule by Janko Gravner.";
"Diamoeba","5678/35678","Creates solid diamond-shaped \"amoeba\" patterns that are surprisingly unpredictable. For a long time it was not known whether any diamonds expand forever, or if the tendency toward the catastrophic destruction of corners is too strong. Finally in March 1999 David Eppstein found the c/7 spaceship, and David Bell made a 100% spacefiller out of it.\nA rule by Dean Hickerson.";
"Day & Night","34678/3678","So named because dead cells in fields of live cells act by the same rules as live cells in fields of dead cells. There are obviously other rules, which have this symmetrical property, but this rule was chosen because it has some interesting high period spaceships and oscillators. The properties of the rule were explored by David Bell.\nA rule by Nathan Thompson.";
"Coral","45678/3","This rule produces patterns with a surprisingly slow rate of expansion and an interesting coral-like texture.\nAuthor unknown.";
"Conway's Life","23/3","This is the most famous cellular automata ever invented. People have been discovering patterns for this rule since around 1970. Large collections are available on the Internet.\nThe rule definition is very simple: a living cell remains alive only when surrounded by 2 or 3 living neighbors,  otherwise it dies of loneliness or overcrowding. A dead cell comes to life when it has exactly 3 living neighbors.\nA rule by John Conway.";
"Coagulations","235678/378","Creates gooey coagulations as it expands forever. Best viewed at zoom=1. Notice that this is a close variation of the previous rule, 235678/3678, except that there is one less condition for a dead cell to come to life on the next generation. In general this should make a universe less active, but this is an exception.\nAuthor unknown.";
"Assimilation","4567/345","Rule similar to Diamoeba, but much more stable. The diamond-shaped patterns are filled in 70-85% and never die.\nAuthor unknown.";
"Amoeba","1358/357","An \"amoeba\" universe - forms large random areas that resemble amoebas. Internal to a random area is chaos. The edge vacillates wildly, and patterns tend to grow more than shrink. The more they grow, the more certain their survival. This is a fairly well-balanced rule.\nAuthor unknown.";
"34 Life","34/34","One of the first explored alternatives to Conway's Life, back in the early 1970's. Computing power was so low back then, it was months before anyone noticed that this is an exploding universe. What makes this universe interesting is the variety of small oscillators and the period-3 orthogonal spaceship.\nAuthor unknown.";
"2x2","125/36","Similar in character to Conway's Life, but creates completely different patterns. Many different oscillators occur at random, and a rare glider. Simple block seeds usually lead to oscillators of various periods.\nThis rule is also a 2x2 block universe. This means that patterns consisting entirely of 2x2 blocks, all aligned, will continue to consist of 2x2 blocks.\nAuthor unknown.";
]

let h_table = Hashtbl.create 1024

let _ = List.iter (fun (a,b,c) -> declare_object ~family:"Cyclic CA" a; 
		     Hashtbl.add h_table a (b,c)) data_cyclic

let _ = List.iter (fun (a,b,c) -> declare_object ~family:"Generations CA" a; 
		     Hashtbl.add h_table a (b,c)) data_generation

let _ = List.iter (fun (a,b,c) -> declare_object ~family:"Life CA" a; 
		     Hashtbl.add h_table a (b,c)) data_life


let _ = valid_hook_plugin ~activate:true ~pattern:".*" ~name 
  ~description
  (fun str ac (w,h) read write ->
     try
       let (expanded,blabla) = Hashtbl.find h_table str in
       add_description ac (str^" (from Mirek's Cellebration website)") blabla;
       set_object_name ac str;
       (Nothing,[expanded])
     with _ -> (Skip,[])
  )
