let name = "Misc file formats"
let description = "Adds minimal support for some file formats"

open Acml

let _ = declare_object "toto"

exception Pb

let _ = valid_hook_plugin ~activate:true ~pattern:"\\(.*\\.icons$\\)\\|\\(.*\\.colors$\\)\\|\\(.*\\.[tT][xX][tT]$\\)" ~name ~description
  (fun str ac (w,h) read write ->
       if Str.string_match (Str.regexp ".*\\.[tT][xX][tT]$") str 0 then (
	 (* Add description from TXT file *)
	   let file = open_in str in
 	   let rec read_all acc =
	     let (nline,ok) = try 
	       (input_line file),true
	     with End_of_file -> ("",false) in
	       if ok then read_all (nline::acc) else String.concat "\n" (List.rev acc)
	   in add_description ac (Filename.basename str) (read_all []); (Nothing,[])
       ) 
       else if Str.string_match (Str.regexp ".*\\.[cC][oO][lL][oO][rR][sS]$") str 0 then (
	 let palette = Hashtbl.create 100 in
	 let colordef = Str.regexp "^[ \t]*color[ \t]*=[ \t]*\\([0-9]+\\)[ \t]+\\([0-9]+\\)[ \t]+\\([0-9]+\\)[ \t]+\\([0-9]+\\)" in
	 let file = open_in str in
 	 let rec read_all () =
	   let rl,continue = try (input_line file),true with End_of_file -> "",false in
	     if continue then (
	       if Str.string_match colordef rl 0 then (
		 let i = int_of_string (Str.matched_group 1 rl) in
		   Hashtbl.add palette i (Unicolor ((int_of_string (Str.matched_group 2 rl)),
						    (int_of_string (Str.matched_group 3 rl)), 
						    (int_of_string (Str.matched_group 4 rl))))
	       );
	       read_all ()
	     ) else ()
	 in 
	   read_all ();
	   add_colormap ac ~set:true (Filename.basename str) 
	     (fun i -> try Hashtbl.find palette i with Not_found -> Unicolor (0,0,0));
	   (Nothing,[])
       )
       else (Nothing,[])
  )
