open Acml

let name = "Quick configuration definition"
let description = "Wait for some piece of OCaml code describing what to put at position (x,y) in the initial configuration. Then register this as an generic initial configuration, which will be available in any new object. The piece of code can (should) use x and y, but there are other bound variable available: w (the width of the current universe), h (the height), n (the number of states of the current CA), r (its radius) and d (the dimension).

Technically it must be used like this:
confNAME:PIECE_OF_CODE
where NAME is an optional name and PIECE_OF_CODE is the OCaml code mentionned above. For instance the following shortcut defines a periodic configuration of period the sequence of all states (in 2D, the configuration is diagonally periodic):
confPeriod all state:(x+y) mod n
"

let pre_source ="open Acml
let maker a =
  (
    let (w,h) = get_universe_size a in
    let r = radius a in
    let n = cardinality a in
    let d = dimension a in
"

let pattern = "conf\\([^:]*\\):\\(.*\\)"

let _ = valid_hook_plugin ~activate:true ~pattern ~name ~description
  (fun str ac (w,h) read write ->
     try
       if Str.string_match (Str.regexp pattern) str 0 then (
	 let name = Str.matched_group 1 str in
	 let core = Str.matched_group 2 str in
	 let ln = min 20 (String.length core) in 
	 let cntd = if ln = String.length core then "" else "..." in
	 let name = if name="" then (String.sub core 0 ln)^cntd else name in
	 let (tmp_name,tmp_file) = Filename.open_temp_file "quickconf" ".ml" in
	 output_string tmp_file pre_source;
	 Printf.fprintf tmp_file  "add_init_conf a ~set:true \"%s\" (General (fun x y -> %s));\n" name core;
	 Printf.fprintf tmp_file  ")\n let _ = valid_plugin ~executed_at:EachNewObject_after ~name:\"conf:%s\" ~activate:true ~description:\"It adds configuration \\\"%s\\\".\"  maker\n" name name;
 	 close_out tmp_file;
	 (Nothing,[tmp_name])
       ) else (Nothing,[])
     with _ -> (infomsg ("Problem in creating source file. Aborted."); (Nothing,[]))
  )
