let name = "Quick rule definition"
let description = "Wait an OCaml function definition as input and consider it as the local rule of a new family. 
The function must have one or two arguments and depending on their name, different rule kind will be used (cf. API documentation): 
\"fun cell ->...\" will trigger a set_rule_fun, 
\"fun any_ident howamny ->...\" will trigger a set_multiset_rule_fun,
\"fun any_ident any_ident ->...\" will trigger a set_multiset_rule, and
\"fun any_ident ->...\" will trigger a set_rule.

Moreover, the function definition has the following variables defined in its scope and can use them:
\"a\" is the current ACML object,
\"d\" gives the dimension, 
\"n\" the number of states, 
\"r\" the radius and 
\"neighb\" the name of the neighbourhood type. 

As an example, the following definition gives the Game of Life (if correct dimension, state, radius, etc are chosen):
\"fun c howmany -> if (c=0 && howmany 1 = 3) || (c=1 && (howmany 1 = 3 or howmany 1 = 4)) then 1 else 0\"
"

open Acml

let pre_source ="open Acml
let maker a =
  (
    let (d,n,r,neighb)=std_params a in
"

let post_source="
  )
let _ =  valid maker
"

let _ = valid_hook_plugin ~activate:true ~pattern:"\\(fun +[^ ]+ *->.*\\)\\|\\(fun +[^ ]+ *[^ ]+ *->.*\\)" ~name 
  ~description
  (fun str ac (w,h) read write ->
     try
       let test_match s = Str.string_match (Str.regexp s) str 0 in
       let def_type = if test_match "fun +cell *->" then "set_rule_fun" 
       else if test_match "fun +[^ ]+ *howmany *->" then "set_multiset_rule_fun"
       else if test_match "fun +[^ ]+ *[^ ]+ *->" then "set_multiset_rule"
       else "set_rule" in
       let (tmp_name,tmp_file) = Filename.open_temp_file "quickdef" ".ml" in
	 output_string tmp_file pre_source;
	 Printf.fprintf tmp_file  "%s a (%s);\n" def_type str;
	 let ln = min 30 (String.length str) in
	 let cntd = if ln = String.length str then "" else "..." in
	   Printf.fprintf tmp_file  "set_name a \"(see source code) %s\";\n" ((String.sub str 0 ln)^cntd);	   
	   output_string tmp_file post_source;
 	   close_out tmp_file;
	   (Nothing,[tmp_name])
     with _ -> (infomsg ("Problem in creating source file. Aborted."); (Nothing,[]))
  )
