open Acml


let aspectratio w h =
  let w = float_of_int w in
  let h = float_of_int h in
  let arbitrary_constant = 15. in
  if w > h then (arbitrary_constant, (h*.arbitrary_constant)/.w)
  else ((w*.arbitrary_constant)/.h, arbitrary_constant)


let write_svg a name reader n w h =
  let f = open_out name in
  let fw,fh = aspectratio w h in
  let color i = 
    let intens = if n=1 then 0 else (255*i)/(n-1) in
    Printf.sprintf "rgb(%d,%d,%d)" intens intens intens
  in
  let label i = Printf.sprintf "state%d" i in
  Printf.fprintf f "<?xml version=\"1.0\" standalone=\"no\"?>
<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" 
  \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">
<svg width=\"%.5fcm\" height=\"%.5fcm\" viewBox=\"0 0 %d %d\"
     xmlns=\"http://www.w3.org/2000/svg\" 
     xmlns:xlink=\"http://www.w3.org/1999/xlink\"
     version=\"1.1\">
  <title>Saved configuration</title>
  <desc>Generated by ACML</desc>
\n" fw fh w h;
  Printf.fprintf f "<defs>\n";
  for i=0 to n-1 do
    Printf.fprintf f "<rect id=\"%s\" width=\"1\" height=\"1\" fill=\"%s\" stroke-width=\"0\"/>\n" (label i) (color i)
  done;
  Printf.fprintf f "</defs>\n";
  for x=0 to w-1 do
    for y=0 to h-1 do
      Printf.fprintf f "<use x=\"%d\" y=\"%d\" xlink:href=\"#%s\" />\n" x (h-1-y) (label (reader x y))
    done
  done;
  Printf.fprintf f "</svg>\n";
  close_out f

let fresh_file_name base ext =
  let ithname i = base^"_"^(string_of_int i)^ext in
  let i = ref 0 in 
    while Sys.file_exists (ithname !i) do
      incr i
    done;
    ithname !i

let maker a=
  let n = cardinality a in
  let d = dimension a in
    add_modif a "save as SVG" (FreeZone
      (fun (ax,ay,bx,by) read write ->
	 let (w,h) = (bx-ax+1),(by-ay+1) in
	 let reader a b = read (ax+a) (ay+b) in
	 let name = fresh_file_name "region" ".svg" in
	   write_svg a name reader n w (if d=1 then 1 else h);
	   infomsg ("File '"^name^"' successfully written.")
      ));
    add_snapshot a "Configuration as SVG"
      (fun w h reader ->
	 let name = fresh_file_name "configuration" ".svg" in
	   write_svg a name reader n w (if d=1 then 1 else h);
	   infomsg ("File '"^name^"' successfully written.")
      )

let _ = valid_plugin ~executed_at:EachNewObject_after ~name:"SVG configuration/region saver" ~activate:true ~description:"It adds actions to save a configuration or a region as a SVG file."  maker
    
