let name = "LaTeX/TikZ output"

let description = "It adds a snapshot/modif that display a window containing source code to copy and paste in a LaTeX source file inside a TikZ figure environment.\n The LaTeX code produces a representation of the universe in grayscale. 
***CAUTION WITH LARGE UNIVERSES*** each cell of the universe produces a line of LaTeX code!"

open Acml

let maker a =
  let n = cardinality a in
  let digit_to_str a = if a=0 then "Z" else "O" in
  let rec state_to_str i = 
    if i < 2 then digit_to_str i 
    else (digit_to_str (i land 1))^(state_to_str (i lsr 1))
  in
  let newcommand_state i = 
    try let na = get_state_name a i in
	Printf.sprintf "\\newcommand\\state%s[2]{\\draw[fill=white!%d!black] (#1,#2) rectangle +(1,1);} %% %s\n" (state_to_str i) ((100*i)/n) na
    with Not_found -> Printf.sprintf "\\newcommand\\state%s[2]{\\draw[fill=white!%d!black] (#1,#2) rectangle +(1,1);}\n" (state_to_str i) ((100*i)/n) in 
  let action = ref (fun () -> ()) in
  let windowexists = ref false in
  let (buff,buffer_add) =
    let (tmbuff,tmbuffer_add) = gui_textbuffer "" in
    (ref tmbuff), (ref tmbuffer_add) in
  let addsometimes_newlines =
    let c = ref 0 in
    (fun () -> incr c; if !c mod 100 = 0 then "\n" else "") in
  let out_f read i j = Printf.sprintf "\\state%s{%d}{%d}%s" (state_to_str (read i j)) i j (addsometimes_newlines ()) in
  let out_full read buffer_add x y = 
    let acc = ref "" in
    for i=0 to x-1 do
      for j=0 to y-1 do
	acc := !acc^(out_f read i j)
      done; 
      buffer_add !acc;
      acc := "";
    done
  in
  let rec enum acc i = if i = 0 then i::acc else enum (i::acc) (i-1) in
  let do_the_job justazone x y read =
    if not !windowexists then (
      let (tmbuff,tmbuffer_add) = gui_textbuffer "" in
      buff:= tmbuff;
      buffer_add:=tmbuffer_add;
      let  w_killer = gui_window ~width:700 ~height:450 "LaTeX output" 
	(gui_vbox[gui_textbox (List.fold_left (fun acc i -> acc^(newcommand_state i)) "" (enum [] (n-1)));
		  !buff;
		  gui_hbox [gui_button "Close" (fun () -> windowexists := false; !action ()); gui_button "Separator" (fun () ->  !buffer_add "\n%%%%%%%%%%\n");];
		 ]) in
      windowexists := true;
      action := w_killer);
      if justazone then !buffer_add (out_f read x y) else
	out_full read !buffer_add x y
  in (
    add_snapshot a ~set:false "Configuration as LaTeX/TikZ" (do_the_job false);
    add_modif a ~set:false "LaTeX/TikZ" (SimpleZone (fun st x y -> do_the_job true x y (fun _ _ -> st); st))
  )

let _ = valid_plugin ~name ~activate:true ~executed_at:EachNewObject_after ~description maker
