(* compilation_options: -package zip zip.cmxa *)
let name = "ZIP archive reader"
let description = "Opens a ZIP archive and recursively opens everything inside."

open Acml

exception Pb

let _ = declare_file_type "ZIP archive" ["zip"; "ZIP"]
    
let _ = valid_hook_plugin ~activate:true ~pattern:".*\\.[zZ][iI][pP]$" ~name ~description
  (fun str ac (w,h) read write ->
     try 
       let rec mk_tmp_dir name =
	 let r = string_of_int (Random.int (1 lsl 20)) in
	 let candidate = Filename.concat (Filename.get_temp_dir_name ()) ((Filename.basename name)^"_"^r) in
	   if Sys.file_exists candidate then mk_tmp_dir name else (
	     Unix.mkdir candidate 0o770; candidate )
       in
       let archive = Zip.open_in str in
       let dirname = mk_tmp_dir str in
       let skip = (Str.regexp "^\\..*") in
       let names = List.rev (List.fold_left (fun acc x -> 
					       if (Str.string_match skip  (Filename.dirname x.Zip.filename) 0 && 
						 x.Zip.is_directory) ||
						 (Str.string_match skip  (Filename.basename x.Zip.filename) 0 && 
						    (not x.Zip.is_directory))
					       then acc
					       else
						 let tmpname = Filename.concat dirname x.Zip.filename in
 						   if x.Zip.is_directory then (
						     Unix.mkdir tmpname 0o770; acc)
						   else (
						       Zip.copy_entry_to_file archive x tmpname; tmpname::acc)
					    ) [] (Zip.entries archive))
       in
	 Zip.close_in archive; (Nothing,names)
     with Zip.Error (arch, entry, msg) -> 
       (infomsg (Printf.sprintf "In entry '%s' of archive '%s': %s" entry arch msg); raise Pb)
  )
  

