let name = "Useful defaults"
let description = "It adds some usefull actions/views/initial configurations."

open Acml

let maker a=
  let n = cardinality a in
  let x,y = get_universe_size a in
  let r = max 2 ((min x y)/15)  in
  let d = dimension a in
  let mem = ref (Array.make_matrix 1 1 0) in
    add_modif a ~key:"c" ~oriented:true "copy"
      (FreeZone (fun zone read write -> 
		   let (x,y,xb,yb)=zone in
		   let (xmi,ymi,xma,yma) = (min x xb),(min y yb),(max x xb),(max y yb) in
		   mem := Array.make_matrix (xma-xmi+1) (yma-ymi+1) 0;
		   let mem i j c = 
		     let mi = if x<=xb then i-x else x-i in
		     let mj = if y<=yb then j-y else y-j in
		     (!mem).(mi).(mj) <- c
		   in
		   for i= xmi to xma do
		     for j= ymi to yma do
		       mem i j (read i j)
		     done
		   done
       ));
    add_modif a ~key:"v" ~oriented:true "paste"
      (FreeZone (fun zone read write -> 
		   let (x,y,xb,yb)=zone in
		   let lx,ly = (Array.length !mem), (Array.length !mem.(0)) in
		   let copy i j = 
		     let mi = if x<=xb then i-x else x-i in
		     let mj = if y<=yb then j-y else y-j in
		     write i j (!mem).(mi mod lx).(mj mod ly)
		   in
		   for i=min x xb to max x xb do
		     for j= min y yb to max y yb do
		       copy i j
		     done
		   done
       ));
    add_modif a ~key:"i" "info"
      (FreeZone (fun zone read write -> 
	   let (x,y,xb,yb)=zone in
	   let vmax = ref (read x y) in
	   let vmin = ref (read x y) in
	   let vsum = ref 0 in
	   let readc = ref 0 in
	   for i=x to xb do
	     for j=y to yb do
	       let v = read i j in
	       vmax := max !vmax v;
	       vmin := min !vmin v;
	       vsum := !vsum + v;
	       incr readc
	     done
	   done;
           let gui_content = if (x=xb && y=yb) then
			       gui_vbox [
				   (gui_label (Printf.sprintf "Cell %dx%d" x y));
				   (gui_label (Printf.sprintf "val=%d%s" !vmax
					         (try Printf.sprintf "\n%s" (get_state_name a !vmax) with Not_found -> "")
				   ));
				 ]
                             else
			       gui_vbox [
				   (gui_label (Printf.sprintf "Region %dx%d -> %dx%d" x y xb yb));
				   (gui_label (Printf.sprintf "max= %d" !vmax));
				   (gui_label (Printf.sprintf "min= %d" !vmin));
				   (gui_label (Printf.sprintf "average: %.2f" ((float_of_int !vsum)/.(float_of_int !readc))))
				 ]
           in
	   let kw = gui_window ~width:300 "Info" gui_content in
           ignore kw
      ));
    add_modif a ~key:"z" "min (0)" (Forced 0);
    add_modif a ~key:"m" (Printf.sprintf "max (%u)" (n-1)) (Forced (n-1));
    add_modif a ~set:true ~left:false ~key:"r" "random" (Map (fun x -> conf_random n));
    add_modif a ~set:true ~left:true ~key:"y" "cycle" (Map (fun x -> (x+1) mod n));
    let back = conf_random n in
      add_init_conf a "Single cell" (General (fun a b -> if a = x/2 && 
						  (d=1 || b = y/2) then conf_random n else back));
      add_init_conf a "Everywhere zero" (Uniform 0);
      add_init_conf a "Random finite" (General (fun a b -> if abs (a - x/2) < r && 
						  (d=1 || abs (b - y/2) <r) then conf_random n else back));
    if d = 1 then 
      add_view a "Time goes down (grayscale)" (fun i j read -> color_gradient a [|(0,0,0); (255,255,255)|] 
					       (read i (y-j-1))
					    )
	     
let _ = valid_plugin ~executed_at:EachNewObject_after ~name ~activate:true ~description maker

