let description = "Cellular automata having a local rule which is invariant by any permutation of the state set."

open Acml;;

let objects = [
"colorblind1D", ["dimension", "1"; "states", "4"; "radius", "3"];
"colorblind2D", ["dimension", "2"; "states", "6"; "neighborhood", "vonNeumann"];
]

let declare = 
  function name, params -> (
    add_shortcut ("^"^name^"$") (fun _ -> params);
    declare_object name
  )

let _ = 
  add_shortcut "^\\([Rr]andom\\)?[- ]?\\(\\([Mm]ulti-?set[- ][Cc]olor-?blind\\)\\|\\([Cc]olor-?blind[- ][Mm]ulti-?set\\)\\)$" 
    (fun _ -> ["multiset", "true"]);
  add_shortcut "^\\([Rr]andom\\)?[- ]?[Cc]olor-?blind?$" (fun _ -> []);
  List.iter declare objects



let maker a =
  (
    let d,n,r,_ = std_params a in
    let h = Hashtbl.create 1024  in
    let mu = bool_param a "multiset" false in
    if mu then (
      set_name a "Multiset color-blind CA";
      set_multiset_rule a (fun me x -> 
	let charac_tab = Array.make (arity a + 1) 0 in
	for i=0 to Array.length x - 1 do
	  charac_tab.(x.(i)) <- charac_tab.(x.(i)) + 1
	done;
	let rec state_from_occurences nb  p =
	  if p = Array.length x then failwith "BUG"
	  else if x.(p) = nb then p
	  else state_from_occurences nb (p+1)
	in
	try
	  state_from_occurences (Hashtbl.find h charac_tab) 0
	with Not_found -> 
	  let uniq_occurence_nbs = 
	    let rec doit acc i = 
	      if i = Array.length x then acc
	      else if charac_tab.(x.(i)) = 1 then doit (x.(i)::acc) (i+1) 
	      else doit acc (i+1) 
	    in Array.of_list (doit [] 0) in
	  if Array.length uniq_occurence_nbs = 0 then failwith "There is no colorblind multiset CA with that parameters";
	  let res = uniq_occurence_nbs.(ca_random (Array.length uniq_occurence_nbs)) in
	  (Hashtbl.add h (Array.copy charac_tab) res; state_from_occurences res 0)			    
      )
    )
    else (
      set_name a "Captive color-blind CA";
      set_rule a (fun x -> 
	let charac_tab = Array.make (arity a) 0 in
	for i=0 to Array.length x - 1 do
	  for j=i to Array.length x - 1 do
	    if charac_tab.(j) = 0 && x.(j)=x.(i) then charac_tab.(j) <- n+i
	  done
	done;
	try
	  x.(Hashtbl.find h charac_tab)
	with Not_found -> let res = ca_random (arity a) in
			  (Hashtbl.add h (Array.copy charac_tab) res; x.(res))
      );
    );
    add_description a "Family" description;
    doc_param a "multiset" "Whether the rule depends only on the multi-set of neighborhing states";

  )

let _ =
  valid maker
