let description = "The resulting state of each transition is chosen at random"

open Acml

let objects = [
"captive1D", ["captive", "true"; "dimension", "1"; "states", "6"; "radius", "2"];
"captive2D", ["captive", "true"; "dimension", "2"; "states", "4"; "neighborhood", "vonNeumann"];
]

let declare = 
  function name, params -> (
    add_shortcut ("^"^name^"$") (fun _ -> params);
    declare_object name
  )

let _ = (
  add_shortcut "^\\([Rr]andom\\)?[- ]?[Cc]aptive$" (fun _ -> ["captive", "true"]);
  add_shortcut "^[Rr]andom$" (fun _ -> ["captive", "false"]);
  List.iter declare objects
)

let maker a = 
  (
    let (d,n,r,neig) = std_params a in
    let b = bool_param a "captive" true in
    ensure_determinism a;
    set_rule a (fun x -> if b then x.(ca_random (Array.length x)) else ca_random n);
    set_name a ("Random "^(if b then "captive" else ""));
    add_description a "Family" (description^(if b then " among states present in the neighbourhood." else " among all states."));
    doc_param a "captive" "Whether all transitions are forced to lead to a state already present in the neighborhood"
  )

let _ =  valid maker
