let description = "cellular automata having a local rule which is invariant by any permutation of neighbors"

open Acml

let objects = [
"multiset1D", ["captive", "false"; "dimension", "1"; "states", "4";];
"multiset2D", ["captive", "true"; "outer", "true"; "dimension", "2"; "states", "4";];
]

let declare = 
  function name, params -> (
    add_shortcut ("^"^name^"$") (fun _ -> params);
    declare_object name
  )

let _ = (
  add_shortcut "^\\([Rr]andom\\)?[- ]?\\(\\([Mm]ulti-?set[- ][Cc]aptive\\)\\|\\([Cc]aptive[- ][Mm]ulti-?set\\)\\)$" 
    (fun _ -> ["captive", "true"]);
  add_shortcut "^\\([Rr]andom\\)?[- ]?[Mm]ulti-?set$" (fun _ -> []);
  List.iter declare objects
)

module OrdTab=
  struct
    type t = int * (int list)
    let compare = compare
  end

module TabMap = Map.Make(OrdTab)

let maker a = 
  (
    let (d,n,r,neig) = std_params ~dimension:2 ~states:3 ~radius:1 ~neighborhood:"Moore" a in
    let b = bool_param a "captive" true in
    let o = bool_param a "outer" false in
    let h = ref TabMap.empty in
    let transmotif c m =
      let rec minilist a x acc =
 	if a > 0 then minilist (a-1) x (x::acc)
 	else acc
      in
      let rec izisont acc i =
	if i = -1 then acc 
 	else if m.(i)>0 then izisont (minilist m.(i) i acc) (i-1) 
	else izisont acc (i-1)
      in 
	(* present states *)
      let lili = (izisont [] (Array.length m - 1)) in
      let tab = Array.of_list  lili in
      let cuple = if o then (c,lili) else (0,lili) in
	try
	  TabMap.find cuple !h
	with Not_found -> 
	  let koi = if b then tab.(ca_random (Array.length tab)) else (ca_random n) in
	    (h:=TabMap.add cuple koi !h; 
	     add_mem_info (Array.length tab + (if o then 1 else 0)); koi) 	  
    in
      set_name a "Random Multiset";
      add_description a "Family" description;
      set_multiset_rule a transmotif;
      doc_param a "captive" "Whether all transitions are forced to lead to a state already present in the neighborhood";
      doc_param a "outer" "Whether the rule depends on the central cell AND the multi-set of neighborhing states (otherwise it depends only on the multi-set of neighborhing states)";
  )
    
let _ =  valid maker
