let description = "cellular automata whose transitions depend only on the set of states present in the neighbourhood."

open Acml

let objects = [
"set1D", ["captive", "false"; "dimension", "1"; "states", "6";];
"set2D", ["captive", "true"; "dimension", "2"; "states", "7"; "neighborhood", "vonNeumann"];
]

let declare = 
  function name, params -> (
    add_shortcut ("^"^name^"$") (fun _ -> params);
    declare_object name
  )


let _ = (
  add_shortcut "^\\([Rr]andom\\)?[- ]?\\(\\([Ss]et[- ][Cc]aptive\\)\\|\\([Cc]aptive[- ][Ss]et\\)\\)$" (fun _ -> ["captive", "true"]);
  add_shortcut "^\\([Rr]andom\\)?[- ]?[Ss]et$" (fun _ -> ["captive", "false"]);
  List.iter declare objects
)

module OrdTab=
  struct
    type t = int * (int list)
    let compare = compare
  end

module TabMap = Map.Make(OrdTab)

let maker a =
    let (d,n,r,neig) = std_params ~dimension:2 ~states:6 ~radius:1 ~neighborhood:"Moore" a in
  let b = bool_param a "captive" true in
  let o = bool_param a "outer" false in
  let h = ref TabMap.empty in
  let transmotif c m =
    let rec izisont acc i =
      if i = -1 then acc 
      else if m.(i)>0 then izisont (i::acc) (i-1) 
      else izisont acc (i-1)
    in 
    let lili = (izisont [] (Array.length m - 1)) in
    let tab = (Array.of_list lili) in
    let cuple = ((if o then c else 0), lili) in
      try
	TabMap.find cuple !h
      with Not_found -> 
	let koi = if b then tab.(ca_random (Array.length tab)) else ca_random n in  
	  (h:=TabMap.add cuple koi !h; add_mem_info (Array.length tab + 1); koi) 
  in
    set_multiset_rule a transmotif;
  add_description a "Family" description;
    doc_param a "captive" "Whether all transitions are forced to lead to a state already present in the neighborhood";
    doc_param a "outer" "Whether the rule depends on the central cell AND the set of neighborhing states (otherwise it depends only on the set of neighborhing states)";
    set_name a ("Random "^(if b then "Set Captive CA" else "Set CA"))

let _ = valid maker

