open Acml;;

let _ = (
  add_shortcut "^\\([Rr]andom\\)?[- ]?\\(\\([Tt]otalistic[- ][Cc]aptive\\)\\|\\([Cc]aptive[- ][Tt]otalistic\\)\\)$" (fun _ -> ["captive", "true"]);
  add_shortcut "^\\([Rr]andom\\)?[- ]?[Tt]otalistic$" (fun _ -> ["captive", "false"])
)

let maker a = 
  (
    let (d,n,r,neig) = std_params ~dimension:2 ~states:3 ~radius:1 ~neighborhood:"Moore" a in
    let b = bool_param a "captive" false in
    let h = Hashtbl.create 1024 in
    let rec sum x acc i = if i>0 then sum x (acc+i*x.(i)) (i-1) else acc in
    let transmotif x m =
      let rec izisont acc i =
	if i = -1 then acc 
	else if x.(i)>0 then izisont (i::acc) (i-1) 
	else izisont acc (i-1)
      in 
	try
	  Hashtbl.find h m 
	with Not_found -> 
	  let tab = (Array.of_list (izisont [] (Array.length x - 1))) in
	  let koi = if b then 
	    tab.(ca_random (Array.length tab)) 
	  else ca_random n in  
	    (Hashtbl.add h m koi; add_mem_info 2; koi) 
    in
    let outer = bool_param a "outer-totalistic" false in
      set_name a ("Random "^(if b then "Captive " else "")^(if outer then "Outer-" else "")^"Totalistic");
      doc_param a "captive" "Whether all transitions are forced to lead to a state already present in the neighborhood";
      doc_param a "outer-totalistic" "Whether next state depends on the sum of states AND the current state of the cell (otherwise it depends only on the sum of states).";
      set_multiset_rule a (fun c x -> transmotif x ((if outer then c else 0),(sum x 0 (n-1)))
			  )
  )

let _ =  valid maker
