let family_name = "Threshold rules"
let description = "The state of a cell is changed depending on whether the sum of neighbouring states is above some threshold or not."

open Acml

let _ =
  add_shortcut "^spin[ -]?loco$" (fun s -> ["states", "2"; "anti-self", "true"; "freezing", "false"]);
  declare_object "spinloco";
  add_shortcut "^bootstrap[ -]?\\(CA\\)?$" (fun s -> ["states", "2"; "anti-self", "false"; "freezing", "true"]);
  declare_object "bootstrap"


let maker a =
  let  (d,n,r,_) = std_params ~states:2 a in
  set_name a family_name;
  add_description a "Rule" description;
  let t = param a "threshold" 0 (n*(arity a)) ((arity a)/2) in
  let freez = bool_param a "freezing" false in
  let selfponderation = bool_param a "anti-self" false in
  set_multiset_rule a (fun c x ->
    let rec cumule acc i = if i = n then acc else cumule (acc+i*x.(i)) (i+1) in
    let sum = cumule 0 1 + 2*(if selfponderation then -1 else 0)*c in
    if freez then (if sum >= t then min (n-1) (c+1) else c)
    else if sum >= t then min (n-1) (c+1) else max 0 (c-1)
  );
  doc_param a "threshold" "The threshold to compare to the sum of neighbours";
  doc_param a "freezing" "If true, the rule is freezing: the state can only increase. If false the cell changes by increasing or decreasing depending on the comparison of the sum of neighbors with the threshold.";
  doc_param a "anti-self" "If true, the cell itself counts negatively in computing the sum of neighbours. If false it counts positively."


let _ =  valid maker 
