open Debug
open Sdl
open Sdlvideo
open Sdlevent


type zoomstate =
    {
      mutable z_orig : int * int;
      mutable z_size : int * int;
      mutable z_cell : int * int;
    }


type kernelrequest =
    HaraKiri

(* global flag for activation of function of this module *)
let interactive= ref true

(* All functions prefixed by "gtk" must test this flag *)
let gtk_enabled = ref true

(* gtk can be usefull even without the control window *)
let gtk_control = ref true

let interactivity_delay = ref 0.
let max_i_cursor = 1000
let interactivity_cursor = ref 

let cadenced = ref false
let cadence = ref 60.
let gTKsolidarity = ref true
let showrate = ref 1
let rate_changed = ref false

(* let blur = ref 0 *)

let infobordery = 20
let infoborderx = 20
let sidepannelsize = 200
let itempannelsep = 10
let thickness = 10

let thesurface = ref None
let thebigarraysurface = ref None

let dorefresh = ref true
let do1Dscroll = ref true

let ready_msg = "Ready."

let colorofsource = function
    Kernel.SrcLeftclick -> (50,50,250)
  | Kernel.SrcRightclick -> (250,50,50)
  | Kernel.SrcMiddleclick -> (50,250,50)
  | Kernel.SrcKeyClick ks -> let i = int_of_char (ks.[0]) in (((i mod 5)*50),((i mod 7)* 30),((i mod 3) * 80))			     
  | _ -> (100,100,100)

(********  file extensions filters  ***********)

let file_filters = ref []
let identity = fun (x:string) -> x
let acml_source_filter = ("ACML source file", (["*.ml"; "*.acml"], identity))

let complete_all_file_filters () = 
  let pre_all = (acml_source_filter::(!file_filters))  in
  let recognized = List.fold_left (fun acc (a,(b,c)) -> b@acc) [] pre_all in
    (("Any known file type", (recognized,identity))::pre_all)@ [("Any file", (["*"], identity))]

let all_file_filters () = List.map (fun (a,(b,c)) -> (a,b)) (complete_all_file_filters ())

let add_file_filter trans name exts = file_filters:=((name,(exts,trans))::(!file_filters))

(********************************************)
(* Geometry and co.                         *)
(********************************************)

(* universe *)
let geomem = ref (0,0)

(* last recorded mouse position *)
let mousex = ref 0 
let mousey = ref 0

(* cell zoom info *)
let zx = ref 0
let zy = ref 0

(* number of displayed cells *)
let x = ref 0
let y = ref 0
(* offset of displayed cells inside the universe *)
let ox = ref 0
let oy = ref 0
(* size of graphic window *)
let sx = ref 0
let sy = ref 0
(* size of cells in pixels *)
let scx = ref 0
let scy = ref 0
(* offset in pixels of the displayed cell grid inside the graphic window *)
let offx = ref 0
let offy = ref 0
(* proportion of matrix cells really displayed *)        
let ratex = ref 1
let ratey = ref 1
        
(* default size of video window *)
let dx = ref 600
let dy = ref 600

let zone_of_coord a b c d =
  let xop u = !ox + ((min (max (!ratex*u - !offx) 0) ((!x-1) * !scx))) / !scx in
  let yop u = !oy + (!y-1) - ((min (max (!ratey*(u- !offy)) 0) ((!y-1) * !scy)))/ !scy in
  let (ux, uy, vx, vy) = (xop a, yop b, xop c, yop d) in
    (
      debugline Dbgdev (Printf.sprintf "multiple modif (%d,%d,%d,%d) %d %d %d %d" 
			  a b c d ux uy vx vy);
      (ux,uy),(vx,vy)
    )
    
let recalibrate () = 
  let s = max 1 (min (!sx / !x) ((!sy / !y))) in
    (      
      ratex := max 1 (!x / !sx);
      ratey := max 1 (!y / !sy);
      scx := s;
      scy := s;
      offx := max 0 ((!sx - (!scx)*(!x)) / 2);
      offy := max 0 ((!sy - (!scy)*(!y)) / 2);
    )
	  
let recenter () =
  (		     
    offx := max 0 ((!sx - (!scx)*(!x)) / 2);
    offy := max 0 ((!sy - (!scy)*(!y)) / 2)
  )
  
(* Read cell size, zoom center, screen size and change the rest *)
let zfocus () =
  let ux,uy = !geomem in
    (
      x := min (ux-1) ((!sx * !ratex) / !scx);
      y := min (uy-1) ((!sy * !ratey) / !scy);
      offx := max 0 ((!sx - (!scx)*(!x)) / 2);
      offy := max 0 ((!sy - (!scy)*(!y)) / 2);      
      ox:= if !ratex > 1 then 0 else min (ux- !x) (max 0 ((!zx - (!mousex - !offx)/ !scx)));
      oy:= if !ratey > 1 then 0 else min (uy- !y) (max 0 ((!zy - (!mousey - !offy)/ !scy)));
      debugline Dbgdev 
	(Printf.sprintf "screen=%dx%d (shift=%dx%d) universe=%dx%d (shift=%dx%d) cells=%dx%d (1/%dx1/%d displayed) zoomat=%dx%d" 
	   !sx !sy !offx !offy !x !y !ox !oy !scx !scy !ratex !ratey !zx !zy)
    )		     

let full_zoom_out () =   
  let ux,uy = !geomem in
    ((!sx * !ratex) / !scx) >= ux && ((!sy * !ratey) / !scy) >= uy

let zoomout () =
  let ux,uy = !geomem in
    if not (full_zoom_out ())  then  (
      scx:= if !scx = 1 then (incr ratex; !scx) else  max (!sx/ ux) (!scx - 1 - (!scx lsr 2));
      scy:= if !scy = 1 then (incr ratey; !scy) else max (!sy/ uy) (!scy - 1 - (!scy lsr 2));
      if (!sx / !scx) >= ux || (!sy / !scy) >= uy  then (
	x:=ux;
	y:=uy;
	ox:=0;
	oy:=0;
	recalibrate ()
      )
      else zfocus ()
    )
      
let calibratezoom () = ()

let zoomat u v =
  if !scx >= !sx || !scy >= !sy then ()
  else (
    if !mousex <> u || !mousey <> v then (
      mousex:=max !offx (min (!sx- !offx) u);
      mousey:=max !offy (min (!sy- !offy) v);
      zx:= (!ratex*(u - !offx))/ !scx + !ox;
      zy:= (!ratey*(v- !offy))/ !scy + !oy
    );
    scx:= if !ratex > 1 then (decr ratex; !scx) else min !sx (!scx + 1+ !scx lsr 2);
    scy:= if !ratey > 1 then (decr ratey; !scy) else min !sy (!scy + 1+ !scy lsr 2);
    zfocus ();
  )


let centerzoom () = zoomat (!x/2) (!y/2)

let geometry ac =
  let (e,f) = ac.Kernel.universe in 
    if !geomem <> (e,f) then 
      (
	x := e;
	y := f;
	geomem := (e,f);
	zx:= e/2;
	zy:=f/2;
	ox:=0;
	oy:=0;
	recalibrate ();
	calibratezoom ();
	debugline Dbgdev 
	  (Printf.sprintf "screen=%dx%d (shift=%dx%d) universe=%dx%d (shift=%dx%d) cells=%dx%d (1/%dx1/%d displayed)" 
	     !sx !sy !offx !offy !x !y !ox !oy !scx !scy !ratex !ratey)
      )
    
(********************************************)
(* SDL stuff                                *)
(********************************************)

let view_window_caption = "ACML view"

let caption_info info = 
  let tit = view_window_caption^" - " ^(info) in
    Sdlwm.set_caption ~title:tit ~icon:tit

(* mode of area selection *)
type mmode = MZone | MDiag | MSide
      
let string_of_mode = function
  | MZone -> "zone"
  | MDiag -> "diagonal"
  | MSide -> "side"

let sdl_curmode = ref MZone

let caption_action ac src = caption_info (Printf.sprintf "action=%s(%s)" (Kernel.getactionname ac src) (string_of_mode !sdl_curmode))

let sdl_stress_mode = ref false

let stress_sdl () =
(*   debugline Dbgdevsdl "SDL is now stressed!"; *)
  (* wake up SDL thread by sending a useless event *)
  add [USER 0]
(*   sdl_stress_mode := true; *)

let agression_sdl () =
  (* signal used to implement interruption of kernel during computation *)
  add [USER 1]

let sdl_curkey = ref None

let select_area x y src ac =
  let screensurf = get_video_surface () in  
  let scopy = create_RGB_surface_format screensurf [] ~w:(!sx) ~h:(!sy) in
  let oversurface = create_RGB_surface_format screensurf [] ~w:(!sx) ~h:(!sy) in
  let overcolor color = map_RGB screensurf color in
  let overcolorline = function (r,g,b) -> map_RGB oversurface ((255-r),(255-g),(255-b)) in
  let recolor color =
    unlock oversurface;
    fill_rect oversurface (overcolor color);
    lock oversurface;
    for i = 0 to  min x y do
      put_pixel oversurface ~x:(x-i) ~y:(y-i) (overcolorline color);
      put_pixel oversurface ~x:(x-i) ~y:(max 0 (y-i-1)) (overcolorline color);
      put_pixel oversurface ~x:(x-i) ~y:(min !sy (y-i+1)) (overcolorline color)
    done;
    for i = 0 to  min (!sx-x-1) (!sy-y-1) do
      put_pixel oversurface ~x:(x+i) ~y:(y+i) (overcolorline color);
      put_pixel oversurface ~x:(x+i) ~y:(max 0 (y+i-1)) (overcolorline color);
      put_pixel oversurface ~x:(x+i) ~y:(min !sy (y+i+1)) (overcolorline color)
    done;
    for i = 0 to  min (x) (!sy-y-1) do
      put_pixel oversurface ~x:(x-i) ~y:(y+i) (overcolorline color);
      put_pixel oversurface ~x:(x-i) ~y:(max 0 (y+i-1)) (overcolorline color);
      put_pixel oversurface ~x:(x-i) ~y:(min !sy (y+i+1)) (overcolorline color)
    done;
    for i = 0 to  min (!sx-x-1) (y) do
      put_pixel oversurface ~x:(x+i) ~y:(y-i) (overcolorline color);
      put_pixel oversurface ~x:(x+i) ~y:(max 0 (y-i-1)) (overcolorline color);
      put_pixel oversurface ~x:(x+i) ~y:(min !sy (y-i+1)) (overcolorline color)
    done;
    unlock oversurface
  in
  set_alpha oversurface ~rle:true 100;
  if must_lock screensurf then unlock screensurf;
  if must_lock scopy then unlock scopy;
  recolor (colorofsource src);
  let mx = ref x in
  let my = ref y in
  let overdraw () =
    let px,py = !mx, !my in
    let (cx,cy) = (min x px, min y py) in
    let (dx,dy) = (max x px, max y py) in
    let rect = {r_x=cx;r_y=cy;r_w=dx-cx;r_h=dy-cy} in
      blit_surface ~src:oversurface ~src_rect:rect ~dst:screensurf ~dst_rect:rect ();
      update_rect screensurf;
      blit_surface ~src:scopy ~dst:screensurf ()
  in
  let rec doit () =
    match wait_event () with
      | MOUSEMOTION mmx -> (
	      let mmo =
		(
		  let rec takelast = function
		      [] -> mmx
		    | (MOUSEMOTION xx)::[] -> xx
		    | t::q -> takelast q
		  in
		    takelast (Sdlevent.get ~mask:mousemotion_mask 5)
		) in
	(* MOUSEMOTION mmo -> ( *)
	      mx := mmo.mme_x;
	      my := mmo.mme_y;
	      overdraw (); doit ())
      | MOUSEBUTTONUP mb -> (disable_events (make_mask [MOUSEMOTION_EVENT]);
			     let coo = zone_of_coord x y mb.mbe_x mb.mbe_y in
			     let tosend = 
			       match !sdl_curkey with
			       | None -> src,coo
			       | Some ks -> (Kernel.SrcKeyClick (Sdlkey.name ks)),coo
			     in match !sdl_curmode with
			     | MZone -> Keymap.Zone tosend
			     | MDiag -> Keymap.ZoneDiag tosend
			     | MSide -> Keymap.ZoneSide tosend
      )
      | KEYDOWN k -> ((match k.keysym with
	| Sdlkey.KEY_LSHIFT -> sdl_curmode := MDiag
	| Sdlkey.KEY_LCTRL -> sdl_curmode := MSide
	| _ -> sdl_curkey:= Some k.keysym);
		      let act = match !sdl_curkey with
			  None -> src
			| Some x -> Kernel.SrcKeyClick (Sdlkey.name x) in 
		      recolor (colorofsource act);
		      overdraw ();
		      caption_action ac act; 
		      doit ())
      | KEYUP k -> ((match k.keysym with
	| Sdlkey.KEY_LSHIFT | Sdlkey.KEY_LCTRL -> sdl_curmode := MZone
	| _ -> sdl_curkey:= None);
		      let act = match !sdl_curkey with
			  None -> src
			| Some x -> Kernel.SrcKeyClick (Sdlkey.name x) in 
		    recolor (colorofsource act);
		    overdraw ();
		    caption_action ac act; 
		    doit ())
      | _ -> doit ()
  in
(*     set_alpha screensurf ~rle:true 100; *)
(*     set_alpha scopy ~rle:true 100; *)
    enable_events (make_mask [MOUSEMOTION_EVENT]);
    blit_surface ~src:screensurf ~dst:scopy ();
    doit ()

let relax_sdl () =
  sdl_stress_mode := false

let toggle_fullscreen =
  let is_full = ref false in
    (function () ->
      if !is_full then (
	let s = (set_video_mode ~w:(!sx) ~h:(!sy) [ `RESIZABLE ]) in
	  thesurface:= Some s;
	  thebigarraysurface:= Some (pixel_data_32 s);
	  is_full := not !is_full;
	  debugline Dbgsdl "leaved fullscreen.";
	  if must_lock (get_video_surface ()) then debugline Dbgsdl "SURFACE MUST LOCK.";
	) else ( 
	let s = (set_video_mode ~w:(!sx) ~h:(!sy) [ `FULLSCREEN ]) in
	  thesurface:= Some s;
	  thebigarraysurface:= Some (pixel_data_32 s);
	  is_full := not !is_full;
	  if must_lock (get_video_surface ()) then debugline Dbgsdl "SURFACE MUST LOCK.";
	  debugline Dbgsdl "switched to fullscreen."
	)
    )

let window_resize w h =
  let s = (set_video_mode ~w ~h [ `RESIZABLE ; `DOUBLEBUF ]) in
    thesurface:= Some s;
    thebigarraysurface:= Some (pixel_data_32 s);
    sx:=w;
    sy:=h;
    debugline Dbgdev (Printf.sprintf "window resized to %dx%d" w h)

let sdl_event_handler ac outsidevent sender = function
    VIDEOEXPOSE -> sender (Keymap.Redraw) 
  | VIDEORESIZE (w,h) -> sender (Keymap.Resize (w,h))
  | MOUSEBUTTONUP but -> (
      match but.mbe_button with
	    (* Left/Right buttonup will be generated when exiting select_area by keypress*)
	  | _ -> sender Keymap.Nothing
    )    
  | MOUSEBUTTONDOWN but -> (
      let doit src ac =
	pump ();
	sender (select_area but.mbe_x but.mbe_y src ac)
      in
	match but.mbe_button with
	    Sdlmouse.BUTTON_LEFT -> (sdl_curmode := MZone; caption_action ac Kernel.SrcLeftclick; doit Kernel.SrcLeftclick ac)
	  | Sdlmouse.BUTTON_RIGHT -> (sdl_curmode := MZone; caption_action  ac Kernel.SrcRightclick; doit Kernel.SrcRightclick ac)
	  (* | Sdlmouse.BUTTON_MIDDLE -> doit (colorofsource Kernel.SrcMiddleclick)  *)
	  (*     (fun x -> Keymap.Zone(Kernel.SrcMiddleclick, x)) (Printf.sprintf "action=%s" (Kernel.getactionname ac Kernel.SrcMiddleclick)) *)
	  | Sdlmouse.BUTTON_WHEELUP -> (
	      let a,b = ((max 0 (min !sx but.mbe_x)),(max 0 (min !sy (!sy-but.mbe_y)))) in
		zoomat a b;
		sender (Keymap.Zoomat (a,b))
	    )
	  | Sdlmouse.BUTTON_WHEELDOWN -> (
	      if not (full_zoom_out ()) then (
		zoomout ();
		sender (Keymap.Zoomout)
	      ) else sender Keymap.Nothing
	    )
	  | _ -> sender Keymap.Nothing
    )
  | QUIT -> sender Keymap.Askexit
  | KEYUP k -> (
    let kc = k.keysym in (* debugline Dbgdev ("key ["^(Sdlkey.name kc)^"]") *)
    let dokey ksym =
      try
	(* Cas spéciaux où l'on doit générer un événement dépendant de paramètres *)
	if List.mem kc (List.map fst Keymap.reserved) then
	  match kc with
	    Sdlkey.KEY_a -> sender (Keymap.Newconfsize (!sx, !sy))
	  | Sdlkey.KEY_p -> sender (Keymap.Newconfsize ((!sx/5), (!sy/5)))
	  | Sdlkey.KEY_d -> sender (Keymap.Newconfsize ((!sx/2), (!sy/2)))
	  | _ -> internalbug "reserved key not known in interactive.ml"
	else let (a,_) = List.assoc kc Keymap.keymap in sender a
      with Not_found -> sender Keymap.Nothing
    in
    match !sdl_curkey with
    | Some ks -> if ks = kc then (sdl_curkey:=None; sender Keymap.Nothing) else dokey kc
    | None -> (
      if k.keymod land Sdlkey.kmod_ctrl <> 0 then sender (Keymap.Zone ((Kernel.SrcKeyClick (Sdlkey.name kc)),
									((!ox, !oy),((!ox+ !x-1),(!oy+ !y-1))))) 
      else if k.keymod land (Sdlkey.kmod_ctrl lor Sdlkey.kmod_shift) = 0 then dokey kc
      else sender Keymap.Nothing
    )
  )
  | USER _ -> (
    debugline Dbgdevsdl "Gtk speaking to SDL";
    sender (Event.sync outsidevent)
  )
  | e -> debugline Dbgdevsdl (string_of_event e)(* ; internalbug "unsollicited event from sdl" *); sender Keymap.Nothing


(* Minimalistic poll function to handle kernel work interruption *)
let sdl_mini_treat outsidevent =
  pump ();
  match poll () with
      Some QUIT -> raise Kernel.ComputationInterrupted
    | Some (USER _) -> ignore (Event.poll outsidevent); raise Kernel.ComputationInterrupted
    | _ -> Keymap.Nothing

(* The poll function for sdl *)
let sdl_treat_events ac blocking outsidevent =
  if not blocking then pump ();
  match (if blocking then       
      let e = (wait_event ()) in
      (* 			  debugline Dbgdev "EVENT"; *)
      Some e
    else poll ()) with
      None -> Keymap.Nothing
    | Some e ->  (pump (); sdl_event_handler ac outsidevent (fun x -> x) e)
      
let debug_sdl_info () =
  let sob = function
      true -> "YES"
    | _ -> "NO"
  in
  let info = get_video_info () in
    debugline Dbgsdl (Printf.sprintf "info on hardware: hw_surf %s, hw_blit %s, sw_blit %s, blit_fill %s, mem %i"
			(sob info.hw_available) (sob info.blit_hw) (sob info.blit_sw) (sob info.blit_fill) info.video_mem
		     )

let sdl_initialization fake =
  (* Sdl.putenv "SDL_VIDEO_FULLSCREEN_HEAD" "1"; *)
  Sdl.init [`VIDEO];
  if not fake then (
    enable_events (make_mask [KEYUP_EVENT; EXPOSE_EVENT; RESIZE_EVENT; QUIT_EVENT; 
			      MOUSEBUTTONUP_EVENT; MOUSEBUTTONDOWN_EVENT; USER_EVENT]);
    disable_events (make_mask [MOUSEMOTION_EVENT]);
    let asurf = (set_video_mode ~w:(!dx) ~h:(!dy) ~bpp:32 [ `RESIZABLE; `DOUBLEBUF]) in
      thesurface := Some asurf;
      thebigarraysurface:= Some (pixel_data_32 asurf);
      sx:=!dx;
      sy:=!dy;
      Sdlwm.set_caption ~title:view_window_caption ~icon:view_window_caption;
      (*       enable_events (make_mask eventList); *)
      debugline Dbgsdl ("init ok (v"^(string_of_version (version ()))^" driver="^(driver_name ())
			^" bpp="^(string_of_int (surface_bpp asurf))^" pitch="^(string_of_int ((surface_info asurf).pitch)) ^")")
  )

let sdlticks = Sdltimer.get_ticks

(********************************************)
(* GTK SECTION START                        *)
(********************************************)

let disable_gtk () = gtk_enabled := false
let disable_control () = gtk_control := false
let use_gtk () = !gtk_enabled

(* Uninitialized gtk variables *)
let window = ref None
let windowbis = ref None
let vbox = ref None
let vboxbis = ref None
let stepinfolabel = ref None
let acmeminfolabel = ref None
let hmeminfolabel = ref None
let gcmeminfolabel = ref None
let toolbarBot = ref None
let toolbarAux = ref None
let delayadjustment = ref None
let interactivitydelayadj = ref None
let encorehbox = ref None
let delayrange = ref None
let refreshadjustment = ref None
let refreshhbox = ref None
let refreshrange = ref None
let toolbarMain = ref None
let colorbox = ref None
let leftactionbox = ref None
let rightactionbox = ref None
let img = ref None
let bss = ref None
let hboxProg = ref None
let pbar = ref None
let sbar = ref None
let adjustmentx = ref None
let adjustmenty = ref None
let fitwin = ref None
let adjustmentcell = ref None
let scrolledParam = ref None
let vboxParams = ref None
let displayedParams = ref []
let vboxPlug = ref None
let displayedPlug = ref []
let vboxAct = ref None
let displayedAct = ref []
let labacname = ref None
let labacspec = ref None
let labacsubname = ref None
let colorselect = ref None
let leftaction = ref None
let rightaction = ref None
let initchoice = ref None
let snapchoice = ref None
let savechoice = ref None
let initchoicebox = ref None
let snapchoicebox = ref None
let savechoicebox = ref None
let acinfobutton = ref None
let mysig = ref None

let quick_completion_update = ref (fun (l:string list) -> ())

let gtk_sender = ref (function Keymap.Nothing|_ -> ())

exception NothingToGet

let getoption = function
    None -> raise NothingToGet
  | Some x -> x

let gtk_switch_show_hide =
  let shown = ref false in
    (fun () -> (
       assert !gtk_enabled;
       if !shown then (
	 GtkThread.async (getoption !windowbis)#misc#hide ()
       ) else
	 (
	   GtkThread.async (getoption !windowbis)#show ()
	 );
       shown := not !shown
     ))
  
(* let stock_img_button txt img button = *)
(*   let imag = (GMisc.image ~stock:(img : GtkStock.id) ()) in *)
(*     button#set_label txt; *)
(*     button#set_image imag#coerce *)
      


let gtk_build_aux_interface () = (
  assert !gtk_enabled;
  (* Aux window *)
  windowbis := Some (GWindow.window ~border_width:2 ~title:"ACML auxiliary" ());
  vboxbis := Some (GPack.vbox ~border_width:1 ~packing:(getoption !windowbis)#add ());
  let align = GBin.alignment ~xalign:0.5 ~yalign:0.5
    ~xscale:0.0 ~yscale:0.0 ~packing:((getoption !vboxbis)#pack ~expand:false ~padding:20) ()
  in  let bxit = GButton.button  ~stock:`CLOSE 
      ~packing:(align#add) () in
    ignore (bxit#connect#clicked ~callback:gtk_switch_show_hide)     
)

let gtk_build_interface () =
  assert !gtk_enabled;
  gtk_build_aux_interface ();
  (* Main window *)
  window := Some (GWindow.window ~border_width:2 ~type_hint:`NORMAL ~title:"ACML control" ());
  vbox := Some (GPack.vbox ~border_width:1 ~packing:(getoption !window)#add ());
  (* CA frame *)
  (
    let hbb = GPack.hbox ~border_width:5 ~packing:((getoption !vbox)#pack ~expand:false ~fill:false) () in
      stepinfolabel := Some (GMisc.label ~markup:"unknown step" 
			       ~packing:(hbb#pack ~expand:false ~fill:true ~padding:2) ());
      let namesbax = GPack.vbox ~border_width:1 ~packing:((hbb)#pack ~expand:true) () in
      let bnbox =  GPack.hbox ~border_width:2 ~packing:((namesbax)#pack ~expand:true) () in
	ignore (GPack.hbox ~border_width:1 ~packing:((bnbox)#pack ~expand:true) ());
	labacname := Some (GMisc.label ~selectable:true ~markup:"unknown name" ~packing:(bnbox#pack ~expand:false ~padding:3) ());
	let snbox =  GPack.hbox ~border_width:1 ~packing:((namesbax)#pack ~expand:true) () in
	  ignore (GPack.hbox ~border_width:1 ~packing:((snbox)#pack ~expand:true) ());
	  labacsubname := Some (GMisc.label ~selectable:true ~markup:"unknown name" 
				  ~packing:((snbox)#pack ~expand:false ~padding:3) ());
	  let snbox =  GPack.hbox ~border_width:1 ~packing:((namesbax)#pack ~expand:true) () in
	    ignore (GPack.hbox ~border_width:1 ~packing:((snbox)#pack ~expand:true) ());
	    labacspec := Some (GMisc.label ~selectable:true ~markup:"unknown specs" 
				 ~packing:((snbox)#pack ~expand:false ~padding:3) ());
	    let imag = (GMisc.image ~stock:(`INFO : GtkStock.id) ()) in
	      acinfobutton := Some (GButton.button ~packing:((hbb)#pack 
							       ~expand:false ~fill:false ~padding:2) ());
	      (getoption !acinfobutton)#set_image imag#coerce	      
  );
  
  (* Quick bar *)
  (  
    let shbox = GPack.hbox ~packing:((getoption !vbox)#pack ~expand:false ~fill:false ~padding:10) () in
      ignore (GMisc.label ~markup:"<i><big>Enter something:</big></i>" ~packing:(shbox#pack ~expand:false ~padding:10) ());
      let cols = new GTree.column_list in
      let col_string = cols#add Gobject.Data.string in
      let store = GTree.list_store cols in
      let fill str =
	let iter = store#append () in
	  store#set ~row:iter ~column:col_string str;
      in
	quick_completion_update := (fun l -> store#clear (); List.iter fill l);
	let entry = GEdit.entry ~text:"" ~packing:(shbox#pack ~expand:true ~padding:10) () in
	let completion = GEdit.entry_completion ~model:store ~entry () in
	  completion#set_text_column col_string;
	  ignore (entry#connect#activate ~callback:(fun () -> !gtk_sender (Keymap.Shortcut entry#text)));
	  entry#misc#grab_focus ();
	  let clean_button = GButton.button ~stock:`CLEAR ~packing:(shbox#pack ~expand:false ~padding:10) () in
	    ignore (clean_button#connect#clicked ~callback:(fun () ->entry#set_text ""))
  );
  
  (  
    let fr = GBin.frame ~packing:((getoption !vbox)#pack ~expand:false ~fill:false ~padding:10) () in
    let vhbox =  GPack.vbox ~border_width:0 ~packing:(fr#add) () in
    let hhbox =  GPack.hbox ~border_width:5 ~packing:(vhbox#pack) () in
    let momobox =  GPack.vbox ~border_width:5 ~packing:((hhbox)#pack ~expand:false ~fill:false) () in
      ignore (GMisc.label ~text:"Left click:" 
		~packing:((momobox)#pack ~expand:false ~fill:true ~padding:1) ());
      leftactionbox := Some (GPack.vbox ~border_width:1 ~packing:(momobox#pack ~expand:false
 								    ~fill:false ~padding:1) ());
      let align = GBin.alignment ~xalign:0.5 ~yalign:0.5
	~xscale:0.0 ~yscale:0.0 ~packing:((hhbox)#pack ~expand:true ~padding:2) () in
	(* Aux toolbar *)
	toolbarAux := Some (GButton.toolbar
			      ~orientation:`HORIZONTAL
			      ~style:`BOTH
			      ~border_width:2 (* ~space_size:5 *)
			      ~packing:((align)#add) ());
	
	(* Start/stop button  *)
	img := Some (GMisc.image ~stock:(`MEDIA_PLAY : GtkStock.id) ());
	bss := Some ((getoption !toolbarAux)#insert_button ~icon:((getoption !img)#coerce) 
	 	       ~text:"Simulation" ~tooltip:"Run/pause simulation" ());
	let momobox =  GPack.vbox ~border_width:5 ~packing:((hhbox)#pack ~expand:false ~fill:false) () in
	  ignore (GMisc.label ~text:"Right click:" 
		    ~packing:((momobox)#pack ~expand:false ~fill:true ~padding:1) ());
	  rightactionbox := Some (GPack.vbox ~border_width:1 ~packing:(momobox#pack ~expand:false
 									 ~fill:false ~padding:1) ());
	  let hhhbox = GPack.hbox ~border_width:1 ~packing:(vhbox#pack ~expand:true ~fill:true ~padding:10) () in
	    refreshhbox := Some hhhbox;
(* 	    ignore (GPack.hbox ~border_width:1 ~packing:(hhhbox#pack ~expand:true ~fill:true ~padding:1) ()); *)
	    refreshadjustment := Some (GData.adjustment ~value:(float_of_int !showrate) ~lower:1. ~upper:1000. ~step_incr:1. 
					 ~page_incr:1. ~page_size:0. ());
	    ignore (GMisc.label ~markup:"<i>Steps/refresh:</i>" ~packing:((getoption !refreshhbox)#pack ~expand:false ~padding:2) ());
	    refreshrange := Some (GEdit.spin_button ~numeric:true ~update_policy:`IF_VALID ~digits:0 
				    ~adjustment:(getoption !refreshadjustment )
				    ~packing:((getoption !refreshhbox)#pack ~expand:false ~padding:2) ());
	    let dodoadushorate x = (
	      showrate := x;
	      rate_changed:=true;
	      (getoption !refreshadjustment)#set_value (float_of_int !showrate)	   
	    ) in
	    let bb = GButton.button ~label:"1 step/refresh" 
	      ~packing:((hhhbox)#pack ~expand:false ~fill:true ~padding:2) () in
  	      ignore (bb#connect#clicked ~callback:(fun () -> dodoadushorate 1));
	      let bb = GButton.button ~label:"Many steps/refresh" 
		~packing:((hhhbox)#pack ~expand:false ~fill:true ~padding:2) () in
  		ignore (bb#connect#clicked ~callback:(fun () -> dodoadushorate (max 1 (!y lsr 2))));
		ignore (GPack.hbox ~border_width:1 ~packing:(hhhbox#pack ~expand:true ~fill:true ~padding:1) ());
		let xxxbox = GPack.hbox ~border_width:1 ~packing:(hhhbox#pack ~expand:true ~fill:true ~padding:1) () in
		ignore (GMisc.label ~markup:"<i>video refresh</i>:" ~packing:((xxxbox)#pack ~expand:false ~fill:false ~padding:2) ());
		let lab_of_act b = if b then "ON " else "OFF" in
		let tbou= GButton.toggle_button ~label:(lab_of_act true) ~active:true 
		  ~packing:((xxxbox)#pack ~expand:false ~fill:false ~padding:2) () in
 		  ignore (tbou#connect#toggled ~callback:(fun () -> !gtk_sender Keymap.Redraw; 
							    tbou#set_label (lab_of_act tbou#active);
							    dorefresh := not !dorefresh));

  );
  


  let centerhbox = GPack.hbox ~border_width:5 ~packing:((getoption !vbox)#pack ~expand:true ~fill:true) () in

    (* Main toolbar *)
    toolbarMain := Some (GButton.toolbar
			   ~orientation:`VERTICAL
			   ~style:`BOTH
			   ~border_width:10
			   ~packing:(centerhbox#pack ~expand:false ~fill:false) ());

    
    (* notebook *)
    let notbook = GPack.notebook ~tab_pos:`TOP ~tab_border:2
      ~packing:(centerhbox#pack ~expand:true ~fill:true) () in
      notbook#set_homogeneous_tabs true;
      
      (* snapshot, view and saver *)
      (
 	let momobox =  GPack.hbox ~border_width:1 ~packing:((getoption !vbox)#pack ~expand:false ~fill:false ~padding:3) () in
	  ignore (GMisc.label ~text:"View:" ~packing:((momobox)#pack ~expand:false ~padding:3) ());
	  colorbox := Some (GPack.vbox ~border_width:1 ~packing:(momobox#pack ~expand:false
 								   ~fill:false ~padding:3) ());
	  ignore (GPack.hbox ~border_width:1 ~packing:(momobox#pack ~expand:true
 							 ~fill:true ~padding:3) ());
	  ignore (GMisc.label ~text:"Initconf:" ~packing:((momobox)#pack ~expand:false ~padding:3) ());
	  initchoicebox := Some (GPack.vbox ~border_width:1 ~packing:(momobox#pack ~expand:false
 								   ~fill:false ~padding:3) ());
 	  let momobox =  GPack.hbox ~border_width:1 ~packing:((getoption !vbox)#pack ~expand:false ~fill:false ~padding:3) () in
	  ignore (GMisc.label ~text:"Snap:" 
		    ~packing:((momobox)#pack ~expand:false ~fill:true ~padding:3) ());
	  snapchoicebox := Some (GPack.vbox ~border_width:1 ~packing:(momobox#pack ~expand:false
 									~fill:false ~padding:3) ());
	  ignore (GPack.hbox ~border_width:1 ~packing:(momobox#pack ~expand:true
 							 ~fill:true ~padding:3) ());
	  ignore (GMisc.label ~text:"Save as:" 
		    ~packing:((momobox)#pack ~expand:false ~fill:true ~padding:3) ());
 	  savechoicebox := Some (GPack.vbox ~border_width:1 ~packing:(momobox#pack ~expand:false
 									~fill:false ~padding:3) ());
      );
      
      (******* contents of notebook **************)
      (* CA parameters *)
      (
	let label = GMisc.label  ~text:"Parameters" () in
	  scrolledParam := Some (GBin.scrolled_window ~border_width:10 ~height:130
				   ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC 
				   ~packing:(fun x -> 
					       ignore (notbook#append_page ~tab_label:label#coerce x)) ());
	  vboxParams := Some (GPack.vbox ~border_width:5  ~packing:(getoption !scrolledParam)#add_with_viewport ());
	  displayedParams := [GPack.hbox ~border_width:1 ~packing:((getoption !vboxParams)#pack) ()];
	  
      );
      (* Actions *)
      (
	let label = GMisc.label ~text:"Actions" () in
	let sw = (GBin.scrolled_window ~border_width:10 ~height:130
		    ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC 
		    ~packing:(fun x -> 
				ignore (notbook#append_page ~tab_label:label#coerce x)) ()) in
	  vboxAct := Some (GPack.vbox ~border_width:5  ~packing:sw#add_with_viewport ());
	  displayedAct := [GPack.hbox ~border_width:1 ~packing:((getoption !vboxAct)#pack) ()];
	  (* ignore (GMisc.label ~markup:"<i>Not implemented yet...</i>" *)
	  (* 	    ~packing:(prebox#pack ~expand:false ~fill:true ~padding:10) ()) *)
      );

      (* Universe *)
      (
	let label = GMisc.label ~text:"Space & Time" () in
	let hboxsettings = (GPack.hbox ~border_width:10 
			      ~packing:(fun x -> ignore (notbook#append_page ~tab_label:label#coerce x)) ()) in
	let vboxy = GPack.vbox ~border_width:1 ~packing:(hboxsettings#pack ~expand:false ~fill:true ~padding:5) () in
	  (* size *)
	let hbox = GPack.hbox ~border_width:1 ~packing:(vboxy#pack ~expand:false ~fill:true ~padding:5) () in
	let labspin (hbox:GPack.box) markup adjref =
	  ignore (GMisc.label ~markup ~line_wrap:true ~justify:`RIGHT
		    ~packing:(hbox#pack ~expand:false ~fill:true ~padding:2) ());
	  ignore (GEdit.spin_button ~numeric:true ~update_policy:`IF_VALID ~digits:0 ~adjustment:(getoption !adjref)
		    ~packing:(hbox#pack ~expand:false ~padding:10) ()) 
	in
	  adjustmentx := Some (GData.adjustment ~lower:2. ~upper:50000. ~step_incr:1. 
				 ~page_incr:1. ~page_size:0. ());
	  adjustmenty := Some (GData.adjustment ~lower:2. ~upper:50000. ~step_incr:1. 
				 ~page_incr:1. ~page_size:0. ());
	  adjustmentcell := Some (GData.adjustment ~lower:1. ~upper:5000. ~step_incr:1. 
				    ~page_incr:1. ~page_size:0. ());
	  labspin hbox "<b>Width:</b>" adjustmentx;
	  labspin hbox "<b>Height:</b>" adjustmenty;
	  let hbox = GPack.hbox ~border_width:1 ~packing:(vboxy#pack ~expand:false ~fill:true ~padding:10) () in
	    fitwin := Some (GButton.button ~label:"Adjust to window" 
			      ~packing:(hbox#pack ~expand:false ~fill:true ~padding:5) ());
	    labspin hbox "<i>with cell size: </i>" adjustmentcell;
	    ignore (GMisc.separator `HORIZONTAL ~packing:(vboxy#pack ~expand:true ~fill:true)());
	    encorehbox := Some (GPack.hbox ~border_width:1 ~packing:(vboxy#pack ~expand:false
 								     ~fill:false ~padding:5) ()) ;
	    
	    (* choice fps/refresh *)
	    delayadjustment := Some (GData.adjustment ~value:( !cadence) ~lower:1. ~upper:1000. ~step_incr:1. 
				       ~page_incr:1. ~page_size:0. ());
	    ignore (GMisc.label ~text:"Desired steps/s:" ~packing:((getoption !encorehbox)#pack ~expand:false ~padding:5) ());
	    delayrange := Some (GEdit.spin_button ~numeric:true ~update_policy:`IF_VALID ~digits:0 
				  ~adjustment:(getoption !delayadjustment)
				  ~packing:((getoption !encorehbox)#pack ~expand:false ~padding:5) ());
	    let lab_of b = if b then "Cadence regulation enabled" else "Cadence regulation disabled" in
	    let tbou= GButton.toggle_button ~label:(lab_of !cadenced) ~active:(!cadenced) 
	      ~packing:((getoption !encorehbox)#pack ~expand:false ~fill:false ~padding:5) () in
 	      ignore (tbou#connect#toggled ~callback:(fun () -> cadenced:=tbou#active; 
							tbou#set_label (lab_of tbou#active)));
(* 	      ignore (GMisc.separator `HORIZONTAL ~packing:(cibox#pack ~expand:true ~fill:true ~padding:10)()); *)
		let tthbox = GPack.hbox ~border_width:1 ~packing:(vboxy#pack ~expand:true
 								    ~fill:true ~padding:5) () in
		let cursor = GData.adjustment ~value:(!interactivity_delay) 
		  ~lower:0. ~upper:0.1 ~step_incr:0.001 
		  ~page_incr:0.01 ~page_size:0. ()
		in
		  interactivitydelayadj := Some cursor;
		  ignore (GMisc.label  ~markup:"<b>Fast</b>" ~justify:`RIGHT ~packing:(tthbox#pack ~expand:false ~fill:true ~padding:1) ());
		  ignore (GRange.scale `HORIZONTAL ~adjustment:cursor ~draw_value:false
 			    ~packing:(tthbox#pack ~expand:true ~fill:true ~padding:1) ());
		  ignore (GMisc.label  ~markup:"<b>Slow</b>" ~justify:`LEFT ~packing:(tthbox#pack ~expand:false ~fill:true ~padding:1) ());
		  ignore (cursor#connect#value_changed ~callback:(fun () -> interactivity_delay := cursor#value))
      );
      (* Plugins *)
      (
	let label = GMisc.label ~text:"Plugins" () in
	let sw = (GBin.scrolled_window ~border_width:10 ~height:130
		    ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC 
		    ~packing:(fun x -> 
				ignore (notbook#append_page ~tab_label:label#coerce x)) ()) in
	vboxPlug := Some (GPack.vbox ~border_width:5  ~packing:sw#add_with_viewport ());
	displayedPlug := [GPack.hbox ~border_width:1 ~packing:((getoption !vboxPlug)#pack) ()];
      );    
      (* Memory *)
      (
	let label = GMisc.label ~text:"Memory" () in
	let prebox = GPack.hbox 
	  ~packing:(fun x -> 
		      ignore (notbook#append_page ~tab_label:label#coerce x)) () in
	let cibox = GPack.vbox ~border_width:10 ~packing:(prebox#pack ~expand:false ~fill:false) () in
	let curhbox = GPack.hbox ~border_width:1 ~packing:(cibox#pack ~expand:false ~fill:false ~padding:10) () in
	  acmeminfolabel := Some (GMisc.label ~markup:"<i>unknown AC mem info</i>" 
				    ~packing:(curhbox#pack ~expand:false ~fill:true ~padding:10) ());
	  ignore (GMisc.separator `HORIZONTAL ~packing:(cibox#pack ~expand:true ~fill:true)());
	  let curhbox = GPack.hbox ~border_width:1 ~packing:(cibox#pack ~expand:false ~fill:false ~padding:10) () in
	    hmeminfolabel := Some (GMisc.label ~markup:"<i>unknown history mem info</i>" 
				     ~packing:(curhbox#pack ~expand:false ~fill:true ~padding:10) ());
	    let bb = GButton.button ~label:"Clear history" 
	      ~packing:((curhbox)#pack ~expand:false ~fill:true ~padding:10) () in
  	      ignore (bb#connect#clicked ~callback:(fun () -> !gtk_sender Keymap.HistoryClear));
	      let lab_of_act b = if b then "Unlimited history" else "Limited history" in
	      let tbou= GButton.toggle_button ~label:(lab_of_act false) ~active:(false)
		~packing:((cibox)#pack ~expand:false ~fill:false ~padding:10) () in
		ignore (tbou#connect#toggled ~callback:(fun () -> !gtk_sender Keymap.Switchhistory; 
							  tbou#set_label (lab_of_act tbou#active)));
		ignore (GMisc.separator `HORIZONTAL ~packing:(cibox#pack ~expand:true ~fill:true)());
		let curhbox = GPack.hbox ~border_width:1 ~packing:(cibox#pack ~expand:false ~fill:false ~padding:10) () in
		let label_effort x = "<i>Max cache size (# of transitions): </i>"^(Kernel.infointnbtostring x) in
		let memklabel = GMisc.label ~markup:(label_effort (Kernel.get_mem_effort ()))
		  ~packing:(curhbox#pack ~expand:false ~fill:true ~padding:10) () in
		let img = GMisc.image ~stock:`GO_UP  () in
		let bu = GButton.button ~label:"2x" 
		  ~packing:((curhbox)#pack ~expand:false ~fill:true ~padding:10) () in
		  ignore (bu#set_image img#coerce);
		  let img = GMisc.image ~stock:`GO_DOWN  () in
		  let bd = GButton.button ~label:"/2"
		    ~packing:((curhbox)#pack ~expand:false ~fill:true ~padding:10) () in
		    ignore (bd#set_image img#coerce);
 		    let change_effort up =
		      let valid new_effort = 
			GtkThread.async memklabel#set_label (label_effort new_effort);
			Kernel.set_mem_effort new_effort
		      in
		      let past_effort = Kernel.get_mem_effort () in
			if up && past_effort <= max_int/2 then valid (past_effort lsl 1)
			else if not up then valid (past_effort lsr 1)
		    in
  		      ignore (bu#connect#clicked ~callback:(fun () -> change_effort true));
  		      ignore (bd#connect#clicked ~callback:(fun () -> change_effort false));
		      ignore (GMisc.separator `HORIZONTAL ~packing:(cibox#pack ~expand:true ~fill:true)());
		      let curhbox = GPack.hbox ~border_width:1 ~packing:(cibox#pack ~expand:false ~fill:false ~padding:10) () in
			gcmeminfolabel := Some (GMisc.label ~markup:"<i>unknown GC info</i>" 
						  ~packing:(curhbox#pack ~expand:false ~fill:true ~padding:10) ());
      );
      
      (* Bottom toolbar *)
      (
	toolbarBot := Some (GButton.toolbar
			      ~orientation:`VERTICAL
			      ~style:`BOTH
			      ~border_width:10
			      ~packing:(centerhbox#pack ~expand:false ~fill:false) ())
      );



      (* Progression/messages bar *)
      (
	hboxProg := Some (GPack.hbox ~packing:((getoption !vbox)#pack ~expand:false ~fill:false ~padding:10) ());
	let physbar = GMisc.statusbar ~packing:((getoption !hboxProg)#pack ~expand:true ~fill:true ~padding:2) () in
	let cntxt = physbar#new_context "ACML" in
	  sbar := Some cntxt;
	  ignore (cntxt#push ready_msg);
	  
	  pbar := Some (GRange.progress_bar ~pulse_step:0.02 ~packing:((getoption !hboxProg)#pack ~expand:false ~fill:false) ())
      )


      
    


let gtk_set_universe ac =
  assert  !gtk_enabled;
  let (x,y) = ac.Kernel.universe in
    GtkThread.sync (fun () -> try (
		       (getoption !adjustmentx)#set_value (float_of_int x);
		       (getoption !adjustmenty)#set_value (float_of_int y);
		       (getoption !adjustmentcell)#set_value (float_of_int (max 1 (max !scx !scy)))
		     ) with NothingToGet -> ()) ()
	
let gtk_compute_fit () =
  assert !gtk_enabled;
  GtkThread.async (fun () -> try (
		     let cs = max 1. (getoption !adjustmentcell)#value in
		       (getoption !adjustmentx)#set_value ((float_of_int !sx)/. cs);
		       (getoption !adjustmenty)#set_value ((float_of_int !sy)/. cs)
		   ) with NothingToGet -> ()) ()
    


let gtk_get_universe ac =
  assert !gtk_enabled;
  (* We must wait for Gtk to ensure the CA is correctly updated *)
  GtkThread.sync (fun () -> try (
		     let u = (int_of_float (getoption !adjustmentx)#value, int_of_float (getoption !adjustmenty)#value) in
		       ac.Kernel.universe <- u
		   ) with NothingToGet -> ()) ()

let wherewerewe = ref None

(* This is a callback *)
let gtk_file title sender =
  assert !gtk_enabled;
  let filew = match !wherewerewe with
      None -> GWindow.file_selection ~title ~height:500 ~border_width:10 () 
    | Some filename -> GWindow.file_selection ~title ~height:500 ~border_width:10 ~filename() 
  in
    ignore (filew#ok_button#connect#clicked ~callback:(fun () -> (sender filew#filename); 
							 wherewerewe := Some filew#filename;  filew#destroy ()));
    ignore (filew#cancel_button#connect#clicked ~callback:filew#destroy);
    filew#show ()

(* This is a callback *)
let gtk_builtin () =    
  assert !gtk_enabled;
  let page_int_fam = ref 0 in
  let page_int_obj = ref 0 in
  let page_ext = ref 0 in
  let window = GWindow.window ~height:600 ~type_hint:`NORMAL ~border_width:2 
    ~title:"Select what you want to open" () in
  let vbox = GPack.vbox ~border_width:1 ~height:400 ~width:800 ~packing:window#add () in
  let notbook = GPack.notebook ~tab_pos:`TOP ~tab_border:10
    ~packing:(vbox#pack ~expand:true ~fill:true) () in
    notbook#set_homogeneous_tabs true;    
    
    (*** Families ****)
    let label = GMisc.label  ~text:"Families" () in
    let hbox = GPack.hbox ~border_width:1 ~packing:(fun x -> page_int_fam:=(notbook#append_page ~tab_label:label#coerce x)) () in
    let scrolled_window = GBin.scrolled_window ~border_width:5
      ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ~packing:(hbox#pack ~expand:true ~fill:true) () in
    let view = GTree.view ~headers_visible:false ~height:200 ~packing:(scrolled_window#add) () in
    let cols = new GTree.column_list in
    let col_nick = cols#add Gobject.Data.string in
    let col_name = cols#add Gobject.Data.string in
    let col_in = cols#add Gobject.Data.string in
      (
	let liststore = GTree.list_store cols in
	let add (nick,name,isin,sav) = 
	  let toplevel = liststore#append () in
	    liststore#set ~row:toplevel ~column:col_name ("<i>"^name^"</i>");
	    liststore#set ~row:toplevel ~column:col_nick nick;
	    liststore#set ~row:toplevel ~column:col_in (if isin then GtkStock.convert_id `HOME else GtkStock.convert_id `FLOPPY);
	in
	  List.iter add (Kernel.get_all_combo ());
	  (* 	ignore (GRange.scrollbar `VERTICAL ~adjustment:view#vadjustment ~packing:(hbox#pack ~padding:10) ()); *)
	  let renderer_text = GTree.cell_renderer_text [] in
	  let col = GTree.view_column ~title:"Id"
	    ~renderer:(renderer_text, ["text", col_nick]) () in
	    ignore (view#append_column col);
	    let renderer_pixbuf = GTree.cell_renderer_pixbuf [] in
	    let col = GTree.view_column ~title:"Builtin"
	      ~renderer:(renderer_pixbuf, ["stock_id", col_in]) () in
	      ignore (view#append_column col);
	      let renderer_text = GTree.cell_renderer_text [] in
	      let col = GTree.view_column ~title:"Family Name"
		~renderer:(renderer_text, ["markup", col_name]) () in
		ignore (view#append_column col);
		view#set_model (Some (liststore#coerce));
		view#selection#set_mode `SINGLE;
      );
      (*** Objects ****)
      let label = GMisc.label  ~text:"Objects" () in
      let hbox = GPack.hbox ~border_width:1 ~packing:(fun x -> page_int_obj:=(notbook#append_page ~tab_label:label#coerce x)) () in
      let scrolled_window = GBin.scrolled_window ~border_width:5
	~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ~packing:(hbox#pack ~expand:true ~fill:true) () in
      let fview = GTree.view ~headers_visible:false ~height:200 ~packing:(scrolled_window#add) () in
      let cols = new GTree.column_list in
      let fcol_nick = cols#add Gobject.Data.string in
      let fcol_name = cols#add Gobject.Data.string in
      let fcol_in = cols#add Gobject.Data.string in
	(
	  let liststore = GTree.list_store cols in
	  let add (nick,name,isin) = 
	    let toplevel = liststore#append () in
	      liststore#set ~row:toplevel ~column:fcol_name ("<i>"^name^"</i>");
	      liststore#set ~row:toplevel ~column:fcol_nick nick;
	      liststore#set ~row:toplevel ~column:fcol_in 
		(if isin then GtkStock.convert_id `HOME else GtkStock.convert_id `REVERT_TO_SAVED);
	  in
	    List.iter add (Kernel.get_objects_deluxe ());
	    (* 	ignore (GRange.scrollbar `VERTICAL ~adjustment:view#vadjustment ~packing:(hbox#pack ~padding:10) ()); *)
	    let renderer_text = GTree.cell_renderer_text [] in
	    let col = GTree.view_column ~title:"Id"
	      ~renderer:(renderer_text, ["text", col_nick]) () in
	      ignore (fview#append_column col);
	      let renderer_pixbuf = GTree.cell_renderer_pixbuf [] in
	      let col = GTree.view_column ~title:"Builtin"
		~renderer:(renderer_pixbuf, ["stock_id", col_in]) () in
		ignore (fview#append_column col);
		let renderer_text = GTree.cell_renderer_text [] in
		let col = GTree.view_column ~title:"Family Name"
		  ~renderer:(renderer_text, ["markup", col_name]) () in
		  ignore (fview#append_column col);
		  fview#set_model (Some (liststore#coerce));
		  fview#selection#set_mode `SINGLE;
	);
	(****** EXTERNAL FILE *********)
	let label = GMisc.label  ~text:"External file" () in
	let hbox = GPack.hbox ~border_width:1 ~packing:(fun x -> page_ext:=(notbook#append_page ~tab_label:label#coerce x)) () in
	let vbox2 = GPack.vbox ~border_width:1 ~packing:(hbox#pack ~expand:true ~fill:true ~padding:20) () in
	let file_choose = GFile.chooser_widget ~action:`OPEN ~packing:(vbox2#pack ~expand:true ~fill:true ~padding:15) () in
	let liste = complete_all_file_filters () in
	let set_filter name = 
	  let patterns = try
			   fst (List.assoc name liste)
	    with Not_found -> ["*"]
	  in
	  file_choose#set_filter (GFile.filter ~patterns ()) in
	let get_final_name fi f =
	  (try
	    snd (List.assoc fi liste)
	  with Not_found -> (fun x -> x)) f
	in
	  (
	    let hbox =  (GPack.hbox ~border_width:1
	    	      ~packing:(vbox2#pack ~expand:false ~fill:true ~padding:10) ()) in
	    ignore (GMisc.label ~markup:"<b>Type of file to open: </b>" ~packing:(hbox#pack ~expand:false ~fill:false ~padding:10) ());
	    let combo = fst (GEdit.combo_box_entry_text ~width:200 
				    ~packing:(hbox#pack ~expand:false
						~fill:false ~padding:10) 
				    ~strings:(List.map fst liste) ()) in
	    let what = (fst (List.hd liste)) in
	    combo#entry#set_text what;
	    set_filter what;
	    ignore (combo#connect#changed 
		      ~callback:(fun () -> set_filter combo#entry#text));
	    ignore (file_choose#connect#file_activated 
		      ~callback:(fun () -> match file_choose#filename with
			  None -> ()
			| Some f -> (!gtk_sender (Keymap.Shortcut (get_final_name combo#entry#text f)); window#destroy ())
		      )
	    )
	  ); 
	
	  (*** buttons ***)
	  let hbox = GPack.hbox ~border_width:1 ~packing:(vbox#pack ~expand:false ~fill:false ~padding:10) () in
	  let boubou = GButton.button ~stock:`OPEN
	    ~packing:(hbox#pack ~expand:true ~fill:true ~padding:10) () in
	  let _ = GPack.hbox ~border_width:1 ~packing:(hbox#pack ~expand:true ~fill:true) () in
	  let boubad = GButton.button ~stock:`CANCEL
	    ~packing:(hbox#pack ~expand:true ~fill:true ~padding:10) () in
	  let action_if_one_selected  () =
	    match notbook#current_page with
		x when (x= !page_int_fam) -> (
		  let sel = view#selection#get_selected_rows in
		    if sel <> [] then let e = List.hd sel in
		    let row = view#model#get_iter e in
		    let nick = view#model#get ~row ~column:col_nick in
		      !gtk_sender (Keymap.Builtin nick)
		)
	      | x when (x= !page_int_obj) -> (
		  let sel = fview#selection#get_selected_rows in
		    if sel <> [] then let e = List.hd sel in
		    let row = fview#model#get_iter e in
		    let nick = fview#model#get ~row ~column:fcol_nick in
		      !gtk_sender (Keymap.Shortcut nick)
		)
	      | x when x = !page_ext -> (
		  match file_choose#filename with
		      None -> ()
		    | Some f -> (!gtk_sender (Keymap.Shortcut f); window#destroy ())
		)
	      | _ -> (window#destroy ())
	  in
	    ignore (boubou#connect#clicked ~callback:(fun () -> action_if_one_selected (); window#destroy ()));
	    ignore (view#connect#row_activated ~callback:(fun e _ -> 
							    let row = view#model#get_iter e in
							    let nick = view#model#get ~row ~column:col_nick in
							      !gtk_sender (Keymap.Builtin nick);
							      window#destroy ()
							 ));
	    ignore (fview#connect#row_activated ~callback:(fun e _ -> 
							    let row = fview#model#get_iter e in
							    let nick = fview#model#get ~row ~column:fcol_nick in
							      !gtk_sender (Keymap.Shortcut nick);
							      window#destroy ()
							 ));
	    ignore (boubad#connect#clicked ~callback:window#destroy);
	    window#show ()
	  
let gtk_refreshranges () =
  assert !gtk_enabled;
  GtkThread.async (fun () -> try (
		     (getoption !delayadjustment)#set_value ( !cadence);
		     (getoption !refreshadjustment)#set_value (float_of_int !showrate)
		   ) with NothingToGet -> ()) ()


let thebuttons =
  [
    (Some "Init", `GOTO_FIRST, toolbarMain, "Back to initial configuration", (fun sender -> sender Keymap.Revert));
    (Some "1 step", `GO_FORWARD, toolbarMain, "Run one step", (fun sender -> sender Keymap.Onestep));
    (Some "Make All", `REFRESH, toolbarAux, "Re-generate rule, configuration and everything else", (fun sender -> sender Keymap.Newboth));
    (Some "Make Conf", `REFRESH, toolbarAux, "Keep current rule but re-generate initial configuration and everything else", 
     (fun sender -> sender Keymap.Newinit));
    (Some "Make Other", `REFRESH, toolbarAux, "Keep current rule and current configuration but re-generate everything else", 
     (fun sender -> sender Keymap.Newneno));
    (Some "Prev", `UNDO, toolbarMain, "Revert to previous object", 
     (fun sender -> sender Keymap.Undo));
    (Some "Open...", `OPEN, toolbarBot, "Open a family, an object or a file...", 
     (fun sender -> gtk_builtin ()));
    (Some "Revert", `REVERT_TO_SAVED, toolbarBot, "Revert from file", (fun sender -> sender Keymap.Revert_from_file));
    (Some "Save", `SAVE, toolbarBot, "Save current object in a new file", 
     (fun sender -> sender Keymap.Saverule));
    (Some "Save...", `SAVE_AS, toolbarBot, "Save current object and asks where", 
     (fun sender -> gtk_file "Select file to save to" (fun f -> sender (Keymap.SaveAs f))));
(*     (Some "Random params", `STRIKETHROUGH, toolbarMain, "Choose parameters randomly", (fun sender -> sender Keymap.Randomparams)); *)
(*     (Some "Re-init PRNG", `REFRESH, toolbarMain, "(Re-)init PRNG state in a system dependent way", (fun sender -> sender Keymap.Newhazard)); *)
    (Some "Copy", `COPY, toolbarMain, "Copy configuration", (fun sender -> sender Keymap.Memorize));
    (Some "Paste", `PASTE, toolbarMain, "Paste configuration", (fun sender -> sender Keymap.Load));
    (Some "Snap", `COPY, toolbarMain, "Trigger a snapshot", (fun sender -> sender Keymap.Screenshot));
    (Some "Help", `HELP, toolbarBot, "Show help", (fun sender -> sender Keymap.Help));
(*     (Some "About", `ABOUT, toolbarBot, "Show copyleft info", (fun sender -> sender Keymap.Licence)); *)
    (Some "Quit", `QUIT, toolbarBot, "Guess what...", (fun sender -> sender Keymap.Askexit));
  ]


let gtk_runinfo t steps fps meminfo = 
  assert !gtk_enabled;
  if fps < 20. || t mod 3 = 0 then
    let dothat = (getoption !stepinfolabel)#set_label in
    let onthat = (Printf.sprintf "<i>step </i><big><b>%d</b></big><i>%s</i>" t
		    (if 0. < fps then (Printf.sprintf "\n(%.1f steps/s)" fps) else "\n"))
    in  (
	GtkThread.async dothat onthat;
	GtkThread.async (getoption !acmeminfolabel)#set_label 
	  ("<i>Current object: </i>"^(Kernel.infointtostring meminfo))
      )
    
(* These 2 functions must be called to update the icon of the corresponding button *)
let gtk_stop_button () =
  assert !gtk_enabled;
  GtkThread.sync (getoption !img)#set_stock `MEDIA_PAUSE
  
let gtk_start_button () =
  assert !gtk_enabled;
  GtkThread.sync (getoption !img)#set_stock `MEDIA_PLAY
      

let gtk_progress i =
  assert !gtk_enabled;
  GtkThread.sync (getoption !pbar)#set_fraction (max (min ((float_of_int i) /. 100.) 1.) 0.)
  



let gtk_infowindow title text justify =
  assert !gtk_enabled;
  GtkThread.async (fun () -> try (
		     let window = GWindow.window ~type_hint:`DIALOG ~position:`CENTER ~border_width:2 ~title () in
		     let vbox = GPack.vbox ~border_width:1 ~packing:window#add () in
		     let scrolled_window = GBin.scrolled_window ~border_width:10
		       ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ~packing:(vbox#pack ~expand:true) () in
		     let tv = GText.view ~editable:false ~width:630 ~height:200 ~packing:scrolled_window#add () in
		       tv#misc#modify_font (Pango.Font.from_string "monospace");
		       tv#buffer#set_text text;
		       tv#set_cursor_visible false;
		       (*   let lab = GMisc.label ~selectable:true ~text ~line_wrap:true ~justify ~packing:scrolled_window#add_with_viewport () in *)
		       let align = GBin.alignment ~xalign:0.5 ~yalign:0.5
			 ~xscale:0.0 ~yscale:0.0 ~packing:(vbox#pack ~expand:false) ()
		       in  let bxit = GButton.button  ~stock:`CLOSE 
			   ~packing:(align#add) () in
			 ignore (bxit#connect#clicked ~callback:window#destroy);			 
			 window#show ()
		   ) with NothingToGet -> ()) ()

let gtk_retry title message =
  assert !gtk_enabled;
  let communique = (Event.new_channel () : bool Event.channel) in
    GtkThread.sync (fun () -> 
		       let window = GWindow.window ~type_hint:`DIALOG ~position:`CENTER ~border_width:2 ~title () in
		       let vbox = GPack.vbox ~border_width:1 ~packing:window#add () in
		       let scrolled_window = GBin.scrolled_window ~border_width:10
			 ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ~packing:(vbox#pack ~expand:false) () in
		       let tv = GText.view ~editable:false ~width:600 ~height:100 ~packing:scrolled_window#add () in
			 tv#misc#modify_font (Pango.Font.from_string "monospace");
			 tv#buffer#set_text message;
			 tv#set_cursor_visible false;
			 let hbox = GPack.hbox ~border_width:20 ~packing:(vbox#pack ~expand:true) () in
			   ignore (GPack.hbox ~border_width:1 ~packing:(hbox#pack ~expand:true) ());
			   ignore (GMisc.label ~markup:"<big><b>Retry compilation?</b></big>" ~packing:(hbox#pack ~expand:true) ());	      
			   ignore (GPack.hbox ~border_width:1 ~packing:(hbox#pack ~expand:true) ());
			   let hbox = GPack.hbox ~border_width:1 ~packing:(vbox#pack ~expand:false) () in
			     ignore (GPack.hbox ~border_width:1 ~packing:(hbox#pack ~expand:true) ());
			     let bxit = GButton.button  ~stock:`YES
			       ~packing:(hbox#pack ~expand:true ~fill:true) () in
			       ignore (GPack.hbox ~border_width:1 ~packing:(hbox#pack ~expand:true) ());
			       let bcancel = GButton.button  ~stock:`NO
				 ~packing:(hbox#pack ~expand:true ~fill:true) () in
				 ignore (GPack.hbox ~border_width:1 ~packing:(hbox#pack ~expand:true) ());
				 ignore (bxit#connect#clicked ~callback:(fun () -> 
									   Event.sync (Event.send communique true); 
									   window#destroy ()));
				 ignore (bcancel#connect#clicked ~callback:(fun () -> 
									      Event.sync (Event.send communique false); 
									      window#destroy ()));
	 			 window#show ()
		    ) ();
    Event.sync (Event.receive communique)

let scrolleur line_wrap ttfont text (where:GPack.box) = 
  let scrolled_window = GBin.scrolled_window ~border_width:10 ~width:700 ~height:300
    ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ~packing:(where#pack ~expand:true ~fill:true) () in
    if ttfont then (
      let hbox2 = GPack.hbox ~border_width:0 ~packing:(scrolled_window#add_with_viewport) () in 
	ignore (GMisc.label ~selectable:true ~text ~line_wrap ~packing:(hbox2#pack ~expand:true ~fill:true ~padding:10) ());
	ignore (GPack.hbox ~border_width:0 ~packing:(hbox2#pack ~expand:true ~fill:true) ())
    ) else  
      let hbox2 = GPack.hbox ~border_width:0 ~packing:(scrolled_window#add_with_viewport) () in (
	  ignore (GMisc.label ~selectable:true ~markup:text ~line_wrap ~packing:(hbox2#pack ~expand:true ~fill:true ~padding:10) ());
	  ignore (GPack.hbox ~border_width:0 ~packing:(hbox2#pack ~expand:true ~fill:true) ())
	)
      

let gtk_launch_help ac =
  if !gtk_enabled then GtkThread.async 
    (fun () ->
      (* notebook *)
      let window = GWindow.window ~width:800 ~height:500 ~type_hint:`DIALOG ~border_width:2 ~title:"ACML HELP" () in
      let vbox = GPack.vbox ~border_width:1 ~packing:window#add () in
      let notbook = GPack.notebook ~tab_pos:`TOP ~tab_border:6
	~packing:(vbox#pack ~expand:true ~fill:true) () in
      notbook#set_homogeneous_tabs true;
      List.iter (fun (text,content) ->
	let label = GMisc.label ~text () in
	let vbox2 = GPack.vbox ~border_width:10 
	  ~packing:(fun x -> ignore (notbook#append_page ~tab_label:label#coerce x)) () in
	scrolleur true false content vbox2) HELP.rubriks;
      let label = GMisc.label ~text:"Keyboard shortcuts" () in
      let vbox2 = GPack.vbox ~border_width:10 ~packing:(fun x -> 
	ignore (notbook#append_page ~tab_label:label#coerce x)) () in
      scrolleur false false ("<i>This applies to the view window only.</i>\n\n<b>Single keystrokes:</b>\n<tt>"^
				(Glib.Markup.escape_text  !Keymap.keymaptext)^
				"</tt>\n\n<b>Key pressed while selecting region with mouse or 'control' + key to apply to the entire universe:</b>\n<tt>"^
				(List.fold_left (^) "" (Kernel.getactionnamefiltered ac (function Kernel.SrcKeyClick _ -> true | _ -> false)
							  (fun (s,nm) -> (match s with Kernel.SrcKeyClick t -> (t^" \t\t "^nm^"\n") | _ -> "BUG"))
				 ))^"</tt>"
	   ) vbox2;
	   let label = GMisc.label ~text:"Licence info" () in
	   let vbox2 = GPack.vbox ~border_width:10 ~packing:(fun x ->
							       ignore (notbook#append_page ~tab_label:label#coerce x)) () in
	     scrolleur false true LICENCE.text vbox2;
	     ignore (GPack.hbox ~border_width:10 ~packing:(vbox#pack ~expand:false) ());
	     let align = GBin.alignment ~xalign:0.5 ~yalign:0.5
	       ~xscale:0.0 ~yscale:0.0 ~packing:(vbox#pack ~expand:false) ()
	     in  let bxit = GButton.button  ~stock:`CLOSE 
		 ~packing:(align#add) () in
	       ignore (bxit#connect#clicked ~callback:window#destroy);			 
	       window#show ()
    ) ()

let gtk_launch_licence () = 
  if !gtk_enabled then GtkThread.async (gtk_infowindow "Copyleft info" LICENCE.text) `LEFT


let param_doc a name = match (Kernel.doc_about a name) with
    None -> ""
  | Some s -> "<i>"^s^"</i>\n"

let actions_info a =
  let actionitem_to_string (s,par) =
    Printf.sprintf "<b>%s</b>: %s\n%s\n" s (Kernel.kind_of_action par) (param_doc a s)
  in
    List.fold_left (fun acc x -> acc^(actionitem_to_string x)) "" a.Kernel.parameters.Kernel.actions

let paramitem_to_string (s,par) =
  let s = Kernel.clean_name s in
    match par with
	Kernel.ParInt p ->  Printf.sprintf "<b>%s</b>: integer between %d and %d" s 
	  p.Kernel.pmin p.Kernel.pmax
      |	Kernel.ParFloat p ->  Printf.sprintf "<b>%s</b>: float between %g and %g" s 
	  p.Kernel.pfmin p.Kernel.pfmax
      | Kernel.ParBool p ->  Printf.sprintf "<b>%s</b>: boolean" s 
      | Kernel.ParString p ->  Printf.sprintf "<b>%s</b>: string"  s 
      | Kernel.ParChoice p ->  Printf.sprintf "<b>%s</b>: choice among %s" 
	  s (List.fold_left (fun acc x -> acc^" \""^x^"\"") "" p.Kernel.pchoices)
    
let param_info a =
  let groups = List.rev (List.filter (function (s, Kernel.Grouparam _) -> true | _ -> false)
    a.Kernel.parameters.Kernel.metaparams) in
    (* content of groups *)
  let groupnames = List.map (fst) groups in
  let ingroups = List.flatten (List.map (function (s, Kernel.Grouparam l) -> l | _ -> []) groups) in
    (* parameters outside all group *)
  let outsiders = List.rev (List.filter (fun x -> not (List.mem (fst x) ingroups)) a.Kernel.parameters.Kernel.params) in
  let pre = List.fold_left (fun acc x -> acc^((paramitem_to_string x)^"\n"^(param_doc a (fst x))^"\n"))
    "" outsiders in
  let show_grp = function
      (name, Kernel.Grouparam l) -> (
	List.fold_left (fun acc x -> try
			  let y = List.assoc x a.Kernel.parameters.Kernel.params in
			    acc^"\t"^((paramitem_to_string (x, y))^"\n\t"^(param_doc a x)^"\n")
			with _ -> (
			  (* test whether this is another group *)
			  if List.mem x groupnames then acc^"\t <b>"^x^"</b>: group of parameters\n\n"
			  else acc
			)
		       )
	  ("<big><i>"^name^": </i></big>\n\n") l
      )
    | _ -> ""
  in pre^(List.fold_left (fun acc x -> acc^(show_grp x)) "" groups)

let object_info mat a =
  let koi = (Kernel.listacinfo a) in
  "<b>"^(List.fold_left (fun acc x -> acc^" / "^x) (List.hd koi) (List.tl koi))^"</b>\n\n"^
  "<i>Universe:</i> "^
    (if a.Kernel.dimension=1 then (Printf.sprintf " %d cells" (fst a.Kernel.universe)) 
     else (Printf.sprintf " %dx%d cells <small>(%s)</small>" (fst a.Kernel.universe) (snd a.Kernel.universe) (Kernel.info_mat mat)))
  ^"\n\n<i>Internal type of rule</i>: "^(Kernel.ruleinfo a)
  ^"\n\n<i>Memory cost</i>: "
       ^(Kernel.infointtostring (Kernel.memsize a))
  ^"\n\n<i>Rule determinism</i>: "^(match a.Kernel.determinization with
					Kernel.Stochastic -> "stochastic"
				      | _ -> "deterministic"
				   )
  ^"\n\n<i>Randomness</i>: "^(Kernel.rndratio a)

let description_info ac =
  match ac.Kernel.description with
      [] -> "<i>No description</i>"
    | li -> List.fold_left (fun acc x -> (Printf.sprintf "<big><b>%s</b>\n<i>%s</i>\n\n</big>" (fst x) (snd x))^acc) "" li

let gtk_info_on_ac mat ac = 
  if !gtk_enabled then GtkThread.async 
    (fun () ->
       (* notebook *)
       let window = GWindow.window ~type_hint:`DIALOG ~border_width:2 ~title:(("Family '")^(Kernel.whatsyourname ac)^"'") () in
       let vbox = GPack.vbox ~border_width:1 ~packing:window#add () in
       let notbook = GPack.notebook ~tab_pos:`TOP ~tab_border:6
	 ~packing:(vbox#pack ~expand:true ~fill:true) () in
	 notbook#set_homogeneous_tabs true;
	 let label = GMisc.label ~text:"Description" () in
	 let vbox2 = GPack.hbox ~border_width:10 
	   ~packing:(fun x -> 
		       ignore (notbook#append_page ~tab_label:label#coerce x)) () in
	   scrolleur true false (description_info ac) vbox2;
	   let label = GMisc.label ~text:"Parameters" () in
 	   let vbox2 = GPack.hbox ~border_width:10 
	     ~packing:(fun x -> 
			 ignore (notbook#append_page ~tab_label:label#coerce x)) () in
	     scrolleur true false (param_info ac) vbox2;
	     let label = GMisc.label ~text:"Actions" () in
 	     let vbox2 = GPack.hbox ~border_width:10 
	       ~packing:(fun x -> 
			   ignore (notbook#append_page ~tab_label:label#coerce x)) () in
	       scrolleur true false (actions_info ac) vbox2;
	     let label = GMisc.label ~text:"Source code" () in
	     let vbox2 = GPack.hbox ~border_width:10 
	       ~packing:(fun x -> 
			   ignore (notbook#append_page ~tab_label:label#coerce x)) () in
	       scrolleur false true (Kernel.whatsyoursource ac) vbox2;
	       let label = GMisc.label ~text:"Current object" () in
	       let vbox2 = GPack.hbox ~border_width:10 
		 ~packing:(fun x -> 
			     ignore (notbook#append_page ~tab_label:label#coerce x)) () in
		 scrolleur true false (object_info mat ac) vbox2;
		 ignore (GPack.hbox ~border_width:10 ~packing:(vbox#pack ~expand:false) ());
		 let align = GBin.alignment ~xalign:0.5 ~yalign:0.5
		   ~xscale:0.0 ~yscale:0.0 ~packing:(vbox#pack ~expand:false) ()
		 in  let bxit = GButton.button  ~stock:`CLOSE 
		     ~packing:(align#add) () in
		   ignore (bxit#connect#clicked ~callback:window#destroy);			 
		   window#show ()
    ) ()

let gtk_add_param tooltips ac packing name parm =
  assert !gtk_enabled;
  let hbox = GPack.hbox ~border_width:4 ~packing () in
    ignore (GPack.hbox ~packing:(hbox#pack ~expand:true ~fill:true) ());
    let tipfun o = 
      match Kernel.doc_about ac name with
	  None -> ()
	| Some text -> (tooltips:GData.tooltips)#set_tip o ~text
    in
      (match parm with
(* 	   Kernel.ParBool _ -> () *)
	 | _ -> tipfun (GMisc.label  ~markup:("<b>"^(Kernel.clean_name name)^":</b> ") 
			  ~line_wrap:true ~justify:`RIGHT
			  ~packing:(hbox#pack ~expand:false ~fill:true) ())#coerce
      );
      (match parm with
	   Kernel.ParInt which -> (
	     let lower = float_of_int which.Kernel.pmin in
	     let upper = float_of_int which.Kernel.pmax in
	     let value = float_of_int which.Kernel.pvalue in
	     let adjustment = GData.adjustment ~value ~lower ~upper 
	       ~step_incr:1. ~page_incr:1. ~page_size:0. () in
	       (
		 ignore (GEdit.spin_button ~numeric:true ~update_policy:`ALWAYS ~digits:0 ~adjustment
			   ~packing:(hbox#pack ~expand:false ~fill:true ~padding:5) ());
		 ignore (adjustment#connect#value_changed  
			   ~callback:(fun () -> 
					which.Kernel.pvalue <- int_of_float adjustment#value;
					debugline Dbgdev ("param "^name^" adjusted");
				     )
			)
	       ))
	| Kernel.ParFloat which -> (
	     let lower = which.Kernel.pfmin in
	     let upper = which.Kernel.pfmax in
	     let value = which.Kernel.pfvalue in
	     let adjustment = GData.adjustment ~value ~lower ~upper 
	        ~page_incr:1. ~page_size:0. () in
	       (
		 ignore (GRange.scale `HORIZONTAL ~digits:5 ~adjustment
			   ~packing:(hbox#pack ~expand:true ~fill:true ~padding:5) ());
		 ignore (adjustment#connect#value_changed  
			   ~callback:(fun () -> 
					which.Kernel.pfvalue <- adjustment#value;
					debugline Dbgdev ("param "^name^" adjusted");
				     )
			)
	       ))
	 | Kernel.ParChoice ki -> (
	     let (select,_) = 
	       GEdit.combo_box_entry_text ~packing:(hbox#pack ~expand:false ~fill:false ~padding:2) 
		 ~strings:ki.Kernel.pchoices () in
	       select#entry#set_text ki.Kernel.pcvalue;
 	       ignore (select#entry#connect#changed
			 ~callback:(fun () -> ki.Kernel.pcvalue <- select#entry#text))
	   )
 	 | Kernel.ParString ket -> 
	     if Kernel.param_is_text ac name then (
	       let buffer = GText.buffer ~text:"" () in
	       buffer#insert ~tags:[] ket.Kernel.psvalue;
		 ignore(GText.view ~editable:true ~accepts_tab:true ~buffer 
			  ~packing:(hbox#pack ~expand:true ~fill:true ~padding:10) ~show:true ());
		 ignore (buffer#connect#changed ~callback:(fun ()-> ket.Kernel.psvalue <- buffer#get_text ()))
	     ) else (
	       let shorten = if Kernel.param_is_file ac name then Filename.basename else fun x -> x in
	       let lab = GEdit.entry ~text:(shorten ket.Kernel.psvalue) ~width:120 
		 ~packing:(hbox#pack ~expand:false ~padding:2) () in
	       let on_click () =
		 let filew = GWindow.file_selection ~title:"Choose file" ~border_width:10 
		   ~filename:ket.Kernel.psvalue () 
		 in
		   ignore (filew#ok_button#connect#clicked 
			     ~callback:(fun () ->  ket.Kernel.psvalue <- filew#filename; 
			       ignore (lab#connect#changed ~callback:(fun ()-> ()));
			       lab#set_text (shorten filew#filename); 
			       ignore (lab#connect#changed ~callback:(fun ()-> ket.Kernel.psvalue <- lab#text));
			       filew#destroy ()));
		   ignore (filew#cancel_button#connect#clicked ~callback:filew#destroy);
		   filew#show ()
	       in 
		 if Kernel.param_is_file ac name then (
		   let bf = GButton.button ~label:"File..." ~packing:(hbox#pack ~expand:false 
									~fill:true ~padding:2) ()
		   in ignore (bf#connect#clicked  ~callback:on_click)); 
	       ignore (lab#connect#changed ~callback:(fun ()-> ket.Kernel.psvalue <- lab#text))
	     )
		 
	 | Kernel.ParBool pa -> 
	     let lab_of_activ b = if b then "Yes" else "No" in
	     let bt = GButton.toggle_button ~label:(lab_of_activ pa.Kernel.pbvalue)
	       ~packing:(hbox#pack ~expand:false ~fill:true ~padding:2) 
	       ~active:(pa.Kernel.pbvalue) ~draw_indicator:true ()
	     in (
		 ignore (bt#connect#toggled ~callback:(fun () -> pa.Kernel.pbvalue <- bt#active; 
							 bt#set_label (lab_of_activ pa.Kernel.pbvalue)));
		 tipfun  bt#coerce
	       )
      );
      debugline Dbgdev ("adding param '"^name^"' in GUI");
      hbox
    
let gtk_get_colormap ac =
  assert !gtk_enabled;
  GtkThread.sync
    (fun () -> ac.Kernel.colormap <- (getoption !colorselect)#entry#text) ()

let gtk_set_colormap ac =
  assert !gtk_enabled;
  GtkThread.async
    (fun () -> (getoption !colorselect)#entry#set_text (ac.Kernel.colormap)) ()

(* let gtk_get_modif ac = *)
(*   assert !gtk_enabled; *)
(*   GtkThread.sync *)
(*     (fun () -> let koi = (getoption !leftaction)#entry#text in *)
(*        ac.Kernel.action <- koi *)
(*     ) () *)

let kill_bill = function
    None -> ()
  | Some a -> (a#coerce)#destroy ()

let gtk_show_actions ac tooltips =
  assert !gtk_enabled;
  let showmeasurement name typ measure packing =
    let tipfun o = 
      match Kernel.doc_about ac name with
	  None -> ()
	| Some text -> (tooltips:GData.tooltips)#set_tip o ~text
    in
    let hbox = GPack.hbox ~border_width:5 ~packing () in
      ignore (GPack.hbox ~packing:(hbox#pack ~expand:true ~fill:true) ());
      displayedAct:= hbox::(!displayedAct);
      let lab = (GMisc.label  ~markup:("<b>"^name^":</b> ") 
		~line_wrap:true ~justify:`RIGHT
		~packing:(hbox#pack ~expand:false ~fill:true) ()) in
	tipfun lab#coerce;
	(match typ with
	     Kernel.MeTrigger -> (
	       let bt = GButton.button ~label:"Exec once"
		 ~packing:(hbox#pack ~expand:false ~fill:true ~padding:2) () in
		 ignore (bt#connect#clicked ~callback:(fun () -> !gtk_sender (Keymap.Measurement measure)))
	     )
	   | Kernel.MeOnOff bobo -> 
	       let lab_of_activ b = if b then "On" else "Off" in
	       let bt = GButton.toggle_button ~label:(lab_of_activ !bobo)
		 ~packing:(hbox#pack ~expand:false ~fill:true ~padding:2) 
		 ~active:(!bobo) ~draw_indicator:true ()
	       in (
		   ignore (bt#connect#toggled ~callback:(fun () -> bt#set_label (lab_of_activ bt#active);
							   bobo := bt#active
							)
			  )
		 )
	);
	debugline Dbgdev ("Action '"^name^"' displayed")
  in (
      (* Remove displayed actions *)
      List.iter (fun (obj : GPack.box) -> (getoption !vboxAct)#remove (obj#coerce); (obj#coerce)#destroy ()) 
	!displayedAct;
      displayedAct := [];
      List.iter (function (name, Kernel.Measurement (typ,acti)) ->  showmeasurement name typ acti (getoption !vboxAct)#pack
		   | _ -> ()
		) (Kernel.get_measures ac)
    )

let gtk_show_params mat ac =
  assert !gtk_enabled;
  GtkThread.sync
    (fun () -> (
       gtk_set_universe ac;
       debugline Dbgdev "handling parameters...";
       (* Remove all previous displayed params *)
       List.iter (fun (obj : GPack.box) -> (getoption !vboxParams)#remove (obj#coerce); (obj#coerce)#destroy ()) 
	 !displayedParams;
       displayedParams := [];
       let tooltips = GData.tooltips ~delay:0 () in 
       let groups = List.filter (function (s, Kernel.Grouparam _) -> true | _ -> false) 
	 ac.Kernel.parameters.Kernel.metaparams in
       let ingroups = List.flatten (List.map (function (s, Kernel.Grouparam l) -> l | _ -> []) groups) in
       let outsiders = List.rev (List.filter (fun x -> not (List.mem (fst x) ingroups)) ac.Kernel.parameters.Kernel.params) in
       let rootgroups = List.filter (fun x -> not (List.mem (fst x) ingroups)) groups in
	 (* Processing groups *)
       let rec mk_group is_root where label list =
	 let mb = GPack.hbox ~border_width:5 ~packing:(where) () in
	 let fr = GBin.frame ~packing:(mb#pack ~expand:true ~fill:true ~padding:10) () in
	 let exp = GBin.expander ~label () in 
	 let va = GPack.vbox () in
	 let vb = GPack.vbox ~packing:(va#pack ~expand:true ~fill:true ~padding:5) () in
	 (
           exp#set_expanded true;
	   fr#add exp#coerce;
	   exp#add va#coerce;
	   if is_root then displayedParams := mb::(!displayedParams);
	   List.iter (fun pname -> try 
		 let p = List.assoc pname ac.Kernel.parameters.Kernel.params in
		 ignore (gtk_add_param tooltips ac vb#pack pname p)
	       with Not_found -> (
		 (* this is not a param, maybe this is a group *)
		 try
		   (match (List.assoc pname groups) with
		      Kernel.Grouparam l -> mk_group false (vb#pack ~expand:true ~fill:true ~padding:5) pname l
		    | _ -> ()
		   )
		 with _ -> ()
	       )
	     ) list		
	 ) 
       in
	 (* Processing parameters outside groups  *)
	 List.iter (fun (name,c) -> 
		      let obj = gtk_add_param tooltips ac (getoption !vboxParams)#pack name c in
			displayedParams := obj::(!displayedParams)
		   ) outsiders;
	 (* Processing parameters inside groups  *)
	 List.iter (function (nm,Kernel.Grouparam l) -> mk_group true (getoption !vboxParams)#pack nm l | _ -> ()) rootgroups;
	 if !displayedParams = [] then displayedParams := [
	   let mb = GPack.hbox ~packing:((getoption !vboxParams)#pack ~expand:true ~fill:true) () in
	     ignore (GMisc.label ~markup:"<i>no parameter</i>" ~packing:(mb#pack ~expand:true ~padding:10) ()); mb
	 ];
	 debugline Dbgdev "parameters done.";
	 (* ACTIONS *)
	 gtk_show_actions ac tooltips;
 	 kill_bill !colorselect;
	 kill_bill !leftaction;
	 kill_bill !savechoice;
	 kill_bill !initchoice;
	 kill_bill !snapchoice;
	 kill_bill !rightaction;
	 (* VIEWER *)
	 colorselect := Some (fst (GEdit.combo_box_entry_text ~width:200 
				     ~packing:((getoption !colorbox)#pack ~expand:false 
						 ~fill:false ~padding:2) 
				     ~strings:(List.map fst ac.Kernel.parameters.Kernel.colormaps) ()));
	 (getoption !colorselect)#entry#set_text (ac.Kernel.colormap);
	 ignore ((getoption !colorselect)#connect#changed ~callback:(fun () -> !gtk_sender Keymap.Newcolormap));
	 debugline Dbgdev "Viewers done.";
	 (* INITCONF *)
	 initchoice := Some (fst (GEdit.combo_box_entry_text ~width:200 
				    ~packing:((getoption !initchoicebox)#pack ~expand:false 
						~fill:false ~padding:2) 
				    ~strings:(List.map fst ac.Kernel.initconfs) ()));
	 (getoption !initchoice)#entry#set_text (ac.Kernel.theinitconf);
	 ignore ((getoption !initchoice)#connect#changed 
		   ~callback:(fun () -> ac.Kernel.theinitconf <- (getoption !initchoice)#entry#text; 
				!gtk_sender Keymap.Newinit));
	 debugline Dbgdev ("initconfs done ("^ac.Kernel.theinitconf^")");
	 (* SNAPSHOT method *)
	 let thesnaps = Kernel.get_snaps ac in
	 let onesnap = Kernel.getactionname ac Kernel.SrcSnapshot in
	   snapchoice := Some (fst (GEdit.combo_box_entry_text ~width:200 
				      ~packing:((getoption !snapchoicebox)#pack ~expand:false 
						  ~fill:false ~padding:2) 
				      ~strings:(List.map fst thesnaps) ()));
	   (getoption !snapchoice)#entry#set_text onesnap;
	   Kernel.defaultaction ac Kernel.SrcSnapshot onesnap;
	   ignore ((getoption !snapchoice)#entry#connect#changed
		     ~callback:(fun () -> Kernel.setaction ac (Kernel.SrcSnapshot) (getoption !snapchoice)#entry#text));
	   debugline Dbgdev "Snapshots done.";
	   (* SAVING method *)
	   let thesavers = Kernel.get_savers ac in
	   let onesaver = Kernel.getactionname ac Kernel.SrcSave in
	     savechoice := Some (fst (GEdit.combo_box_entry_text ~width:200 
					~packing:((getoption !savechoicebox)#pack ~expand:false 
						    ~fill:false ~padding:2) 
					~strings:(List.map fst thesavers) ()));
	     (getoption !savechoice)#entry#set_text onesaver;
	     Kernel.defaultaction ac Kernel.SrcSave onesaver;
	     debugline Dbgdev "Savers done.";
	     (* MODIF method *)
	     let themodifs = Kernel.get_modifs ac in
 	     let leftmodif = Kernel.getactionname ac Kernel.SrcLeftclick in
 	     let rightmodif = Kernel.getactionname ac Kernel.SrcRightclick in
	       leftaction := Some (fst (GEdit.combo_box_entry_text ~width:140 
					  ~packing:((getoption !leftactionbox)#pack ~expand:false 
						      ~fill:false ~padding:2) 
					  ~strings:(List.map fst themodifs) ()));
	       (getoption !leftaction)#entry#set_text leftmodif;
	       rightaction := Some (fst (GEdit.combo_box_entry_text ~width:140 
					   ~packing:((getoption !rightactionbox)#pack ~expand:false 
						       ~fill:false ~padding:2) 
					   ~strings:(List.map fst themodifs) ()));
	       (getoption !rightaction)#entry#set_text rightmodif;
	       (*        (getoption !colorselect)#entry#set_text ac.Kernel.colormap; *)
	       ignore ((getoption !leftaction)#entry#connect#changed
			 ~callback:(fun () -> Kernel.setaction ac (Kernel.SrcLeftclick) (getoption !leftaction)#entry#text));
	       ignore ((getoption !rightaction)#entry#connect#changed
			 ~callback:(fun () -> Kernel.setaction ac (Kernel.SrcRightclick) (getoption !rightaction)#entry#text));
	       debugline Dbgdev "Modif methods done.";
	       (match !mysig with
		    Some sign -> (getoption !acinfobutton)#misc#disconnect sign
		  | _ -> ()); 
	       mysig := Some ((getoption !acinfobutton)#connect#clicked 
				~callback:(fun () -> gtk_info_on_ac mat ac));
     )) ()

(************ PLUGINS ******************)

let gtk_show_plugin_info pg =
  if !gtk_enabled then GtkThread.async 
    (fun () ->
       (* notebook *)
       let window = GWindow.window ~type_hint:`DIALOG ~border_width:2 
	 ~title:("Plugin '"^pg.Kernel.pg_name^"'") () in
       let vbox = GPack.vbox ~border_width:1 ~packing:window#add () in
       let notbook = GPack.notebook ~tab_pos:`TOP ~tab_border:6
	 ~packing:(vbox#pack ~expand:true ~fill:true) () in
	 notbook#set_homogeneous_tabs true;
	 let label = GMisc.label ~text:"Informations" () in
	 let vbox2 = GPack.hbox ~border_width:10 
	   ~packing:(fun x -> 
		       ignore (notbook#append_page ~tab_label:label#coerce x)) () in
	   scrolleur true false 
	     ("<b>When exectuted: </b><i>"^(Glib.Markup.escape_text (Kernel.string_from_repetition pg.Kernel.pg_repetition))
	      ^"</i>\n\n<b>Description: </b>\n"^(Glib.Markup.escape_text pg.Kernel.pg_description)) 
	     vbox2;
	   let label = GMisc.label ~text:"Source code" () in
	   let vbox2 = GPack.hbox ~border_width:10 
	     ~packing:(fun x -> 
			 ignore (notbook#append_page ~tab_label:label#coerce x)) () in
	     scrolleur false true (pg.Kernel.pg_src) vbox2;
	     ignore (GPack.hbox ~border_width:10 ~packing:(vbox#pack ~expand:false) ());
	     let align = GBin.alignment ~xalign:0.5 ~yalign:0.5
	       ~xscale:0.0 ~yscale:0.0 ~packing:(vbox#pack ~expand:false) ()
	     in  let bxit = GButton.button  ~stock:`CLOSE 
		 ~packing:(align#add) () in
	       ignore (bxit#connect#clicked ~callback:window#destroy);			 
	       window#show ()
    ) ()

let gtk_show_plugins ac =
  if !gtk_enabled then GtkThread.async 
    (fun () ->
       let showplugin pg packing =
	 let hbox = GPack.hbox ~border_width:5 ~packing () in
	   ignore (GPack.hbox ~packing:(hbox#pack ~expand:true ~fill:true) ());
	   displayedPlug:= hbox::(!displayedPlug);
	   ignore (GMisc.label  ~markup:("<b>"^pg.Kernel.pg_name^":</b> ") 
		     ~line_wrap:true ~justify:`RIGHT
		     ~packing:(hbox#pack ~expand:false ~fill:true) ());
	   let bt = GButton.button ~stock:`INFO
	     ~packing:(hbox#pack ~expand:false ~fill:true ~padding:2) () in
 	     ignore (bt#connect#clicked ~callback:(fun () -> gtk_show_plugin_info pg));
	     (match pg.Kernel.pg_repetition with
		  Kernel.Alademande -> (
		    let bt = GButton.button ~label:"Exec once"
		      ~packing:(hbox#pack ~expand:false ~fill:true ~padding:2) () in
		      ignore (bt#connect#clicked ~callback:(fun () -> pg.Kernel.pg_function ac))
		  )
		| _ -> 
		    let lab_of_activ b = if b then "On" else "Off" in
		    let bt = GButton.toggle_button ~label:(lab_of_activ pg.Kernel.pg_active)
		      ~packing:(hbox#pack ~expand:false ~fill:true ~padding:2) 
		      ~active:(pg.Kernel.pg_active) ~draw_indicator:true ()
		    in (
			ignore (bt#connect#toggled ~callback:(fun () -> bt#set_label (lab_of_activ bt#active);
								!gtk_sender (Keymap.SwitchPlugin (pg, bt#active))
							     ))
		      )
	     );
	     debugline Dbgdev ("Plugin '"^pg.Kernel.pg_name^"' displayed")
       in (
	   (* Remove displayed plugins *)
	   List.iter (fun (obj : GPack.box) -> (* (getoption !vboxPlug)#remove (obj#coerce); *) (obj#coerce)#destroy ()) 
	     !displayedPlug;
	   displayedPlug := [];
	   List.iter (fun pg -> 
			showplugin pg (getoption !vboxPlug)#pack
		     ) (Kernel.list_plugins ());
	   let h = GPack.hbox ~border_width:5 ~packing:(getoption !vboxPlug)#pack () in
	   displayedPlug := h::(!displayedPlug);
	   let f = GBin.frame ~label:"Hooks" ~packing:(h#pack ~expand:true) () in
	   let bbox =  GPack.vbox ~border_width:1 ~packing:f#add () in
	   List.iter (fun pg -> 
			showplugin pg bbox#pack
		     ) (Kernel.list_hook_plugins ());
	 )
    ) ()


let gtk_propose_aborting killer =
  assert !gtk_enabled
    
let gtk_discard_aborting () =
  assert !gtk_enabled

let gtk_overbook over = 
  debugline Dbgtk (if over then "OVER" else "UNDER")

(* GTK blocking listener for kernel events  *) 
let gtkObeyAcmlOnce ch =
  assert !gtk_enabled;
  let listen = Event.receive ch in
    match Event.sync listen with
      | HaraKiri -> (
(* 	    (getoption !window)#destroy (); *)
	    GMain.quit ();
	    debugline Dbgtk "down.";
	    Thread.exit ()
	  )
(*       | _ -> internalbug "Gtk kernel event handler incomplete" *)
	  

let (stress_gtk, howtostress) =
  let channel = ref (Event.new_channel ()) in
    ( 
      (fun () -> GtkThread.async gtkObeyAcmlOnce !channel; debugline Dbgtk "GTK is now stressed"),
      (fun c -> channel := c)
    )

(* Gtk main window thread *)
let gtkWindow channels =
  assert !gtk_enabled;
  let (fromk,tok) = channels in
    (* each time an event is sent, stress the SDL layer to force the SDL thread to sync with that event *)
  let sender x = GtkThread.sync (fun () -> stress_sdl (); Event.sync (Event.send tok x)) () in
    (
      gtk_sender := sender;
      howtostress fromk;
      ignore (GtkMain.Main.init ());
      let t = Thread.create GtkThread.main () in
      let (ma, mi, mu) = GMain.Main.version in
	debugline Dbgtk (Printf.sprintf "init ok (v%d.%d.%d)" ma mi mu);
	if !gtk_control then (
	  debugdoing Dbgtk "building interface";
	  GtkThread.sync (fun () -> 
			    (
			      gtk_build_interface ();
			      ignore ((getoption !window)#connect#destroy ~callback:(fun () -> exit 0));
			      ignore ((getoption !windowbis)#connect#destroy ~callback:(fun () -> exit 0));
			      ignore ((getoption !bss)#connect#clicked ~callback:(fun () -> sender Keymap.Stopstart));
			      (* 			    ignore ((getoption !colorbutton)#connect#clicked ~callback:(fun () -> sender Keymap.Newcolormap)); *)
			      ignore ((getoption !fitwin)#connect#clicked ~callback:gtk_compute_fit);
			      ignore ((getoption !delayadjustment)#connect#value_changed 
					~callback:(fun () -> 
						     cadence:=(getoption !delayadjustment)#value));
			      (* 			    ignore ((getoption !interactivitydelayadj)#connect#value_changed  *)
			      (* 				      ~callback:(fun () ->  *)
			      (* 						   interactivity_delay:=(getoption !interactivitydelayadj)#value /. 1000.)); *)
			      ignore ((getoption !refreshadjustment)#connect#value_changed 
					~callback:(fun () -> 
						     rate_changed := true;
						     showrate:=int_of_float (getoption !refreshadjustment)#value));
			      List.iter (fun (t,a,b,c,d) -> 
 					   let img = GMisc.image ~stock:(a : GtkStock.id) () in
					     match t with 
						 None -> 
						   ignore (((getoption !b) : GButton.toolbar)#insert_button ~tooltip:c 
							     ~icon:(img#coerce) ~callback:(fun () -> d sender) ())
					       | Some text -> ignore (((getoption !b) : GButton.toolbar)#insert_button 
									~text ~tooltip:c 
									~icon:(img#coerce) ~callback:(fun () -> d sender) ())
					)
				thebuttons;
			      
			      ignore ((getoption !window)#show ());
			    )) ();
	); 
	debugdone Dbgtk;
	t
    )
      

(********************************************)
(* GTK USER API CALLS                       *)
(********************************************)


type api_gui_element = ( GObj.widget -> unit) -> unit


(* WINDOW *)
let api_window title width height content_continuation =
  if !gtk_enabled then 
    let window = ref None in
      GtkThread.sync 
	(fun () ->
	   let w = GWindow.window ~height ~width ~type_hint:`DIALOG ~border_width:2 ~title () in
	   window := Some w;
	   content_continuation w#add;
	   ignore (w#show ());
	) ();
      (fun () ->  if !gtk_enabled then GtkThread.sync (fun () -> match !window with
							   None -> ()
 							 | Some w -> w#destroy ()
						      ) ())
  else (fun () -> ())


(* LABEL *)
let api_label text = fun x -> (
  ignore (GMisc.label ~text ~line_wrap:true ~packing:x ())
)

(* FULL LABEL *)
let api_full_label text = 
  let lbl = ref None in
    ((fun x -> (
	lbl := Some (GMisc.label ~text ~line_wrap:true ~packing:x ())
    )),
     (fun text -> match !lbl with
	  None -> ()
	| Some lab -> GtkThread.async (fun () -> lab#set_label text) ()
     )
    )

(* TEXT BUFFER *)
let api_textbuffer text = 
  let where = ref None in
    ((fun x -> (
       let scrolled_window = GBin.scrolled_window ~border_width:10 ~width:630 ~height:200
	 ~hpolicy:`AUTOMATIC ~vpolicy:`AUTOMATIC ~packing:x () in
       let vbox2 = GPack.vbox ~border_width:0 ~packing:(scrolled_window#add_with_viewport) () in
       let buffer = GText.buffer ~text () in
       let view = GText.view ~buffer ~width:500 ~editable:false ~packing:(vbox2#pack ~expand:true) () in
	 view#set_wrap_mode `WORD;
	 where := Some buffer
     )),
       (fun text ->
	  match !where with
	      None -> ()
	    | Some buff -> GtkThread.async (fun () -> (buff#insert text)) ()
		(* (GMisc.label ~selectable:true ~text ~line_wrap:true ~packing:(box#pack ~expand:true) ()) *)
       )
    )

(* VBOX *)
let api_vbox content_continuation_list = 
  fun x -> let vbox = GPack.vbox ~packing:x () in
  let vpack = vbox#pack ~padding:1 ~from:`START ~expand:false ~fill:false in
  let vfill () = ignore (GPack.vbox ~packing:(vbox#pack ~padding:1 ~expand:true ~fill:true) ()) in
    List.iter (fun cc -> cc vpack; vfill ()) content_continuation_list

(* HBOX *)
let api_hbox content_continuation_list = 
  fun x -> let hbox = GPack.hbox ~packing:x () in
  let hpack = hbox#pack ~padding:1 ~from:`START ~expand:false ~fill:false in
  let hfill () = ignore (GPack.hbox ~packing:(hbox#pack ~padding:1 ~expand:true ~fill:true) ()) in
    List.iter (fun cc -> cc hpack; hfill ()) content_continuation_list

(* BUTTON *)
let api_button label action = fun x -> (
  let bf = GButton.button ~label ~packing:x ()
  in ignore (bf#connect#clicked  ~callback:(fun () -> action ())); 
)  

type api_draw_context =
    {
      dc_size : unit-> int*int;
      dc_point: (int*int*int) -> int -> int -> unit;
      dc_rectangle : (int*int*int) -> int -> int -> int -> int -> unit;
      dc_clear : unit -> unit
    }

let normalize_color (a,b,c) =
  let norm_compo x =  (x land 255) lsl 8 in
    (norm_compo a, norm_compo b, norm_compo c)

(* DRAW AREA *)
let api_draw_area height width background = 
  let image = ref None in
  (* let pix = Gdk.Pixmap.create ~height ~width ~depth:24 () in *)
  let pixmap = GDraw.pixmap ~width ~height () in
  let draw_brush fg x y =
    let y = height-1 - y in
      pixmap#set_foreground fg;
      pixmap#rectangle ~x ~y ~width:1 ~height:1 ~filled:false ();
      (match !image with 
	   None -> ()
	 | Some i -> i#set_pixmap pixmap
      )
  in
  let draw_rect fg x y w h =
    let y = (height-1) - (y+h-1) in
      pixmap#set_foreground fg;
      pixmap#rectangle ~x ~y ~width:w ~height:h ~filled:true ();
      (match !image with 
	   None -> ()
	 | Some i -> i#set_pixmap pixmap
      )
  in
  let clear () =
    pixmap#set_foreground (`RGB (normalize_color background));
    pixmap#rectangle ~x:0 ~y:0 ~width ~height ~filled:true ()
  in
    (
      (fun x ->
	 (
	   let i = (GMisc.pixmap pixmap ~height ~width ~packing:x ~show:true ()) in
	     image := Some i;
	     pixmap#set_background (`RGB (normalize_color background));
	     clear ()
	 )),
      ({
	 dc_point = (fun color x y -> 
		       GtkThread.async (fun () -> draw_brush (`RGB (normalize_color color)) x y) ());
	 dc_rectangle = (fun color x y width height -> 
		       GtkThread.async (fun () -> draw_rect (`RGB (normalize_color color)) x y width height) ());
	 dc_size = (fun () -> (width,height));
	 dc_clear = (GtkThread.async (fun () -> clear ()))
       }
       )
    )
	

(********************************************)
(* STATUSBAR and MESSAGES                   *)
(********************************************)

let (statusbar, status, clearstatus) = 
  let info = ref "OK" in
  let changed = ref true in
    ((fun long sync -> 
	if !changed then (
	  caption_info !info;
	  if !gtk_enabled  && !gtk_control then (if sync then GtkThread.sync else GtkThread.async) 
	    (fun x -> 
	       (getoption !sbar)#pop ();
	       (* if long then *) ignore ((getoption !sbar)#push x) (* else (getoption !sbar)#flash ~delay:2000 x *)
	    ) !info;
	  changed:=false;
	)
     ),
     (fun x -> info := x; changed:=true),
     (fun () ->  
		    (
		      info := ""; changed:=true
		    )
     )
    )

let setstatus s = 
  status s; statusbar true false

let setworkingstatus s =
  status s; statusbar true true

let setimmediatestatus s =
  status s; statusbar true true
  
let drawborder on off onsize = ()

let infozoom () =  setstatus ("cell size="^(string_of_int (min !scx !scy)))

let retryorabort title message =
  if !gtk_enabled && !interactive then gtk_retry title message else false

let waittick () = if !gtk_enabled && !gtk_control then GtkThread.sync (getoption !pbar)#pulse ()

let (working, unworking, quickunworking) =
  let f () = () in 
    (f,
     (fun () -> if !gtk_enabled  && !gtk_control then GtkThread.async
	(fun () -> status ready_msg; (getoption !pbar)#set_fraction 0.0; statusbar true true;) ())
       ,f)

let progress i = 
  if !gtk_enabled  && !gtk_control then (quickunworking (); gtk_progress i)

let running () = if !gtk_enabled  && !gtk_control then GtkThread.async (getoption !pbar)#set_fraction 0.0


(********************************************)
(* GTK CALLS ENCAPSULATION                  *)
(********************************************)


let get_universe ac =
  if !gtk_enabled && !gtk_control then gtk_get_universe ac

let set_universe ac =
  if !gtk_enabled && !gtk_control then gtk_set_universe ac

let refreshranges () =
  if !gtk_enabled  && !gtk_control then gtk_refreshranges ()

let stop () =
  if !gtk_enabled  && !gtk_control then gtk_stop_button ()

let start () =
  if !gtk_enabled  && !gtk_control then gtk_start_button ()

let runinfo t steps fps mi =
  if !gtk_enabled  && !gtk_control then gtk_runinfo t steps fps mi

let launch_help ac = 
  if !gtk_enabled then gtk_launch_help ac

let launch_licence () = 
  if !gtk_enabled then gtk_launch_licence ()

let show_params mat ac = 
  if !gtk_enabled && !gtk_control then (
    setworkingstatus "Refreshing parameter list...";
    gtk_show_params mat ac;
  )

let show_plugins ac = 
  if !gtk_enabled && !gtk_control then (
    setworkingstatus "Refreshing plugin list...";
    gtk_show_plugins ac
  )

let startstopwilldostop () =
  if !gtk_enabled  && !gtk_control then gtk_stop_button ()

let startstopwilldostart () =
  if !gtk_enabled  && !gtk_control then gtk_start_button ()

let get_colormap ac =
  if !gtk_enabled  && !gtk_control then gtk_get_colormap ac

(* let get_modif ac = *)
(*   if !gtk_enabled then gtk_get_modif ac *)

(* let changeoverbook over = *)
(*   if !gtk_enabled then gtk_overbook over *)

(* let propose_aborting killer = *)
(*   if !gtk_enabled then gtk_propose_aborting killer *)

(* let discard_aborting () = *)
(*   if !gtk_enabled then gtk_discard_aborting () *)


      
let puttitle (hlength,hmem) ac = 
  if !gtk_enabled && !gtk_control then 
    GtkThread.async
      (fun () -> (
	 (getoption !labacname)#set_label ("<i>Family: </i><b><big>"^(Glib.Markup.escape_text (Kernel.whatsyourname ac))^"</big></b>");
	 (getoption !hmeminfolabel)#set_label ("<i>History of objects: </i>"^(Kernel.infointtostring hmem)
					       ^" ("^(string_of_int hlength)^" objects)");
	 let st = Gc.quick_stat () in
	 (getoption !gcmeminfolabel)#set_label ("<i>GC info: </i>#major="
						^(string_of_int (st.Gc.major_collections))
						^" / heap="^(Kernel.humanbiginttostring 
							       (Big_int.mult_big_int 
								  (Big_int.big_int_of_int st.Gc.heap_words)
								  (Big_int.big_int_of_int Sys.word_size)))^"b"
						^" / #chunks="^(string_of_int st.Gc.heap_chunks)
					       );
	  let koi = (Kernel.listacinfo ac) in
 	    (getoption !labacsubname)#set_label ("<i>Object: </i><b>"^(Glib.Markup.escape_text (Kernel.whatsyourobjname ac))^"</b>");
	    (getoption !labacspec)#set_label ("<small>"^(Glib.Markup.escape_text (List.fold_left (fun acc x -> acc^" / "^x) (List.hd koi) (List.tl koi)))
					      ^"</small>");
	    !quick_completion_update (List.append ((Kernel.get_objects ()))
					(Kernel.get_all_nicks ()));
 	    debugline Dbgdev "AC info displayed"
       )) 
      ()
	 
	
	      
let rec drawinfo draw_rect rgb =
  function
      [] -> ()
    | (Kernel.GenGrid (xs,ys,r,g,b))::q ->
 	(
	  for i = 0 to !x / (!ratex*xs) - 1 do
	    for j = 0 to !y / (!ratey*ys) - 1 do
	      draw_rect (!offx + !scx*(i*xs)) (!offy + !scy*(j*ys)) (!scx*xs-1) (!scy*ys-1) (rgb r g b)
 	    done
 	    done;
 	  drawinfo draw_rect rgb q
 	)
    | (Kernel.Markcell ((a,b),(r,g,bl)))::q -> (
	if a < !ox + !x  && a>= !ox && b < !oy + !y && b >= !oy && (a mod !ratex =0) && (b mod !ratey =0) then (
	  if true then draw_rect (!offx + !scx*(a - !ox)/(!ratex)) (!offy + !scy*(!y - 1 - b + !oy)/(!ratey)) 
	    (!scx-1) (!scy-1) (rgb r g bl);
	  (* the condition below implies that ratex=1 and ratey=1 so we don't care about them *)
	  if  !scx > 2 && !scy > 2 then 
	    draw_rect (!offx + !scx*(a - !ox) + 1) (!offy + !scy*(!y - 1 - b + !oy) + 1) 
	      (!scx - 3) (!scy - 3) (rgb r g bl));
	drawinfo draw_rect rgb q)
	      
	      
let hline surf x y w col =
  for x=x to x+w do
    put_pixel surf ~x ~y col
  done

let vline surf x y h col =
  for y=y to y+h do
    put_pixel surf ~x ~y col
  done

let clearscreen blackrect = 
  blackrect 0 0 !sx !sy
    
let colorcache a =
  match !thesurface with
      None -> (debugline Dbgdev "no visual in refresh")
    | Some q -> 
	if not (Kernel.complex_colors a) then 
	  (
	    debugdoing Dbgdev "Generating color cache";
	    setworkingstatus "Computing color cache...";
	    let t = Array.make a.Kernel.cardinality Int32.zero in
	    let dosingle i = function
		Kernel.Uni (r,g,b) -> (r,g,b)
	      | _ -> (
		  debugline Dbgdev (Printf.sprintf "state %d wrongly represented (red color)!" i);
		  (200,0,0)
		)
	    in
	    let mix x (d,e,f) = match x with
		Some (a,b,c) -> Some ((a+d)/2, (b+e)/2, (c+f)/2)
	      | _ -> Some (d,e,f)
	    in
	    let rec domultiple i cur = function
		[] -> (match cur with
			   None -> internalbug "Multiple color without color!"
			 | Some x -> x)
	      | t::q -> let n = dosingle i t in
		  domultiple i (mix cur n) q
	    in
	      for i = 0 to a.Kernel.cardinality-1 do
		t.(i) <- map_RGB q (match Kernel.colorofstate a i (fun a b -> 0) 0 0 with
					Kernel.Overdraw l -> domultiple i None l
				      | a -> dosingle i a
				   )
	      done;
	      a.Kernel.colorcache <- t;
	      debugdone Dbgdev
	  ) (* else raise (Kernel.BadACML "Too many states in this CA") *)

let refpos = ref 0
      
(* let motion_blur = *)
(*   let minicount = ref 0 in *)
(*     (fun fonc arg -> *)
(*        if !minicount >= !blur then (fonc arg; minicount:=0) *)
(*        else incr minicount *)
(*     ) *)


let refresh clear ac mat t fps steps run = 
  let cells = ac.Kernel.universe in
  let rate = !showrate in
  let therect = {r_x=0; r_y=0; r_w=0; r_h=0} in
    match !thesurface, !thebigarraysurface with
      | (Some q), (Some big) -> (
	  let pitch=(surface_info q).pitch in
	  let bpp=(surface_bpp q) lsr 3 in
	  let safe_rect x y w h col = (
	    therect.r_x <- x;
	    therect.r_y <- y;
	    therect.r_w <- w;
	    therect.r_h <- h;
 	    fill_rect ~rect:therect q col
	  ) in
	  let rect x y w h col = (
	    let pos = ref ((y*pitch+x*bpp) lsr 2) in
	      for j=0 to h-1 do
		pos := ((y+j)*pitch+x*bpp) lsr 2;
		for i = 0 to w-1 do
		  Bigarray.Array1.unsafe_set big !pos col;
		  pos:= !pos + (bpp lsr 2)
		done
	      done
	  ) in
	  let pixel x y col = Bigarray.Array1.unsafe_set big ((y*pitch+x*bpp) lsr 2) col in
	  let nfrect x y w h col = (
	    hline q x y w col;
	    vline q (x+w) y h col;
	    hline q x (y+h) w col;
	    vline q x y h col;
	  ) in
	  let rgb r g b = map_RGB q (r,g,b) in
	  let arrow_color = rgb 250 0 0 in
	  let dodrawcell = 
	    if Kernel.complex_colors ac then
	      (fun a b sta reader c d -> 
		 let rec doit i j = function
		     Kernel.Uni (r,g,b) -> rect (i*(!scx)+ !offx) (j*(!scy)+ !offy) !scx !scy (rgb r g b)
		   | Kernel.Up -> (
		       rect (i*(!scx)+ !offx) (j*(!scy)+ !offy + (!scy - !scy/3)) !scx (!scy/3) (arrow_color);
		       rect (i*(!scx)+ !scx/3+ !offx) (j*(!scy)+ !offy) (!scx/3) !scy (arrow_color);
		     )
		   | Kernel.Down -> (
		       rect (i*(!scx)+ !offx) (j*(!scy)+ !offy) !scx (!scy/3) (arrow_color);
		       rect (i*(!scx)+ !scx/3+ !offx) (j*(!scy)+ !offy) (!scx/3) !scy (arrow_color);
		     )
		   | Kernel.Left -> (
		       rect (i*(!scx)+ !offx) (j*(!scy)+ !offy + !scy/3) !scx (!scy/3) (arrow_color);
		       rect (i*(!scx)+ !scx - !scx/3 + !offx) (j*(!scy)+ !offy) (!scx/3) !scy (arrow_color);
		     )
		   | Kernel.Right -> (
		       rect (i*(!scx)+ !offx) (j*(!scy)+ !offy + !scy/3) !scx (!scy/3) (arrow_color);
		       rect (i*(!scx)+ !offx) (j*(!scy)+ !offy) (!scx/3) !scy (arrow_color);
		     )
		   | Kernel.Arround (r,g,b) -> (
		       rect (i*(!scx)+ !offx) (j*(!scy)+ !offy) !scx (!scy/6) (rgb r g b);
		       rect (i*(!scx)+ !offx) (j*(!scy)+ !scy - (!scy/6) + !offy) !scx (!scy/6) (rgb r g b);
		       rect (i*(!scx)+ !offx) (j*(!scy)+ !offy) (!scx/6) !scy (rgb r g b);
		       rect (i*(!scx)+ !scx - (!scx/6) + !offx) (j*(!scy)+ !offy) (!scx/6) !scy (rgb r g b); 
		     )
		   | Kernel.Overdraw lili -> List.iter (doit i j) lili
		   | Kernel.Nocolor -> ()
		 in doit a b (Kernel.colorofstate ac sta (fun x y -> reader x (y)) c d)
(* 		 in doit a b (Kernel.colorofstate ac sta (fun x y -> reader x (h-1-y)) a (h-1-b)) *)
	      )
	    else if !sx < !x || !sy < !y then 
	      (fun i j sta reader c d -> safe_rect (i*(!scx)+ !offx) (j*(!scy)+ !offy) !scx !scy (Array.unsafe_get ac.Kernel.colorcache sta))
	    else if !scx = 1 && !scy = 1 then
	      (fun i j sta reader c d -> pixel (i*(!scx)+ !offx) (j*(!scy)+ !offy) (Array.unsafe_get ac.Kernel.colorcache sta))
	    else
	      (fun i j sta reader c d -> rect (i*(!scx)+ !offx) (j*(!scy)+ !offy) !scx !scy (Array.unsafe_get ac.Kernel.colorcache sta))
	  in
	    (
 	      runinfo t steps fps (Kernel.memsize ac);
 	      if ((t mod rate = 0) && !dorefresh) || (not run) || clear then (	    
		if must_lock q then lock q;
		if clear then clearscreen (fun x y w h -> rect x y w h (rgb 0 0 0));
		let ux,uy = ac.Kernel.universe in
		(match ac.Kernel.dimension with
		     2 ->
		       let r = Kernel.getradius ac in
		       let thearray = Kernel.getcurrentbidimmatrix ac mat t in
		       let theoldarray = Kernel.getoldbidimmatrix ac mat t in
		       let readarray a b = Kernel.getter_from_fast_matrix thearray (a+r*steps) (b+r*steps) in
		       let safereader a b = readarray (((fst cells)+a) mod (fst cells)) ((2*(snd cells)+b) mod (snd cells)) in
		       let readoldarray a b = Kernel.getter_from_fast_matrix theoldarray (a+r*steps) (b+r*steps) in
			 if (rate < 2) && Kernel.is_sparse ac && (not clear) then (
			   let getbipos = function
			       Kernel.Plane (a,b) -> (a,b)
			     | _ -> internalbug "getbipos interactive"
			   in
			     Kernel.SpotSet.iter (fun pos -> let (i,j) = getbipos pos in
						    if i>= !ox && i <= !ox+ !x-1 && j>= !oy && j <= !oy+ !y-1
						    then dodrawcell (i- !ox) (!y - 1 -j + !oy) 
						      (readarray (i) (j)) (safereader) i j
						 ) ac.Kernel.modified
			 )
			 else
			   (
			     (* refresh only if state  change when we have access to previously displayed state *)
			     let dodo = if (rate < 2) && (not (Kernel.is_sparse ac)) 
			       && (not clear) && (not (Kernel.uses_view ac)) && (not !rate_changed) then 
 			       (fun a b c d -> if readarray (c) (d) <> readoldarray (c) (d) then
				  dodrawcell a b (readarray (c) (d)) safereader c d
			       ) 
			     else
			       (
				 fun a b c d -> dodrawcell a b (readarray (c) (d)) safereader c d
			       )
			     in
			       (
				 rate_changed := false;
				 for i = max 0 ((!ox)/(!ratex)) to min (ux-1) ((!ox+ !x-1)/(!ratex)) do
				   (* 				  debug Dbgdev "<"; *)
				   for j = max 0 ((!oy)/(!ratey)) to min (uy-1) ((!oy+ !y-1)/(!ratey)) do
				     (* 				    debug Dbgdev (Printf.sprintf "[%d]" j); *)
				     dodo (i- !ox/(!ratex)) ((!y - 1 + !oy)/(!ratey) -j) (!ratex*i) (!ratey*j) 
				   done;
				   (* 				  debug Dbgdev ">" *)
				 done;
			       )
			   )
		   | 1 ->  let b = snd cells in
		     let r = Kernel.getradius ac in
		     let (thearray,timeshift) = Kernel.getcurrentunidimmatrix mat in
		     let readarray a pres = thearray.(a+r*steps).((pres + timeshift+1) mod b) in
		     let safereader u v = readarray (((fst cells)+u) mod (fst cells)) (v) in
		       (
			 if (not clear) && run && (not !do1Dscroll) then (
			   let where = (b + timeshift - !refpos) mod b in
			     if where < !oy + !y && where > !oy - 1 then 
			       for i = (!ox)/(!ratex) to (!ox+ !x-1)/(!ratex) do
				 dodrawcell (i - !ox/(!ratex)) ((!oy + !y -1)/(!ratey) - where/(!ratey))
 				   (readarray (!ratex*i) (!oy+ !y-1)) safereader (!ratex*i) (!oy + !y -1)
			       done
			 ) else (
			   refpos := (timeshift + 1) mod b;
 			   for i = (!ox)/(!ratex) to (!ox+ !x-1)/(!ratex) do
			     for j = (!oy)/(!ratey) to (!oy+ !y-1)/(!ratey) do
			       dodrawcell (i - !ox/(!ratex)) ((!oy + !y -1)/(!ratey) -j)
				 (readarray (!ratex*i) (!ratey*j)) safereader (!ratex*i) (!ratey*j)
			     done
			   done
			 )
		       )
 		   | _ -> raise (Kernel.Unimplemented "higher dimension"));
		(* DECORATIONS *)
		drawinfo nfrect rgb (Kernel.getoverdrawings ac);
		if must_lock q then unlock q;
		flip q;
		(* if run then motion_blur flip q else flip q; *)
 		statusbar false false;
	      );
	    )
	)
      |	_ -> (debugline Dbgdev "surface problem in refresh")


	
let quickrefresh ac mat t steps run = ()
	
let help () = ()
    
let take_a_breath = sdl_mini_treat

type interface =
    {
      execThread : Thread.t;
      channelCom : kernelrequest Event.channel
    }

(* common channel to send info to kernel *)
let tokernel = ((Event.new_channel ()) : (Keymap.event Event.channel))


let ((releaseInterface : (((kernelrequest Event.channel) * (Keymap.event Event.channel) -> Thread.t) -> string -> unit)), 
     pollInteraction, waitInteraction, destroyInterfaces) =
  let myevent = ((Event.receive tokernel) : (Keymap.event Event.event)) in
  let threadList = ref [] in
    (   
      (fun funk name -> 
	 let fromkernel = Event.new_channel () in
	 let t = funk (fromkernel, tokernel) in
	   debugline Dbgdev ("new thread ("^name^" id="^(string_of_int (Thread.id t))^")");
	   threadList := (name,{channelCom=fromkernel; execThread =t})::(!threadList)
      ),
      (fun ac quick -> if quick then sdl_mini_treat myevent else sdl_treat_events ac false myevent
      ),
      (fun ac -> sdl_treat_events ac true myevent
      ),
      (fun () -> 
	(* Prepare GTK to listen to kernel *)
	if !gtk_enabled then stress_gtk ();
	(List.iter (fun (nam,ifce) -> (
	    debugline Dbgdev ("sending Hara-Kiri to thread "^nam);
	    Event.sync (Event.send ifce.channelCom HaraKiri); 
	  )
		   ) !threadList;
 	 List.iter (fun (nam,ifce) -> (
	     Thread.join ifce.execThread;
	     debugline Dbgdev ("thread "^nam^" (id="^(string_of_int (Thread.id ifce.execThread))
				^") terminated honorably.")
	   )
		   ) !threadList
	)
      )
    )
      
let closeAll () = 
  (
    debugline Dbgdev "Destroying interfaces...";
    destroyInterfaces ();
    debugline Dbgdev "There is a single thread left.";
    Sdl.quit ();
    debugline Dbgdev "Exiting...";
    exit 0
  )

let regulator_cost = 1.
let regulate_cadence  =
  let oldtick = ref 0. in 
  let oldim = ref 2 in
    (fun dim t  ->
       if dim <> !oldim then (
	 oldim:=dim;
	 showrate:=if dim=1 then max 1 (!y lsr 2) else 1;
	 if !gtk_enabled && !gtk_control then (getoption !refreshadjustment)#set_value (float_of_int !showrate)	   
       );
       if !cadenced then 
	 let newtick = float_of_int (Sdltimer.get_ticks ()) in
	 let diff = (1000. /. !cadence) -. newtick +. !oldtick  in
	   (
	     if diff > !interactivity_delay +. regulator_cost then (
	       Thread.delay ((diff-.regulator_cost) /. 1000.))
	     else (
	       Thread.delay !interactivity_delay;
	     );
	     oldtick := float_of_int (Sdltimer.get_ticks ())
	   )
       else if !interactivity_delay > 0. 
       then Thread.delay !interactivity_delay 
       else if !gTKsolidarity then (if t mod 7 = 0 then Thread.delay 0.001
				   else Thread.yield ()
				  )
    )
      

let wysiwyg filename =
  match !thesurface with
      None -> (debugline Dbgdev "no video surface to snap")
    | Some q -> (
	let q = get_video_surface () in
	let ba = Bigarray.Array2.create Bigarray.int32 Bigarray.c_layout !sy !sx in
	let (w,h,_) = surface_dims q in
	  debugline Dbgdev (Printf.sprintf "Copying displayed video content: %dx%d pixels" w h);
 	  for x=0 to w-1 do
	    for y=0 to h-1 do
	      let (red,green,blue) = get_pixel_color q ~x ~y in
		ba.{y,x} <- Int32.add (Int32.shift_left (Int32.of_int red) 16) (Int32.of_int  (green lsl 8 + blue))
	    done;
 	  done;
	  debugline Dbgdev "Writing to file...";
 	  Cairo_png.surface_write_to_file (Cairo_bigarray.of_bigarr_32 ~alpha:false ba) filename
      )

let raw_gray filename reader ac w h =
  let ba = Bigarray.Array2.create Bigarray.int32 Bigarray.c_layout h w in
  let get_color =
    if Kernel.complex_colors ac then
    (fun x y ->
      match Kernel.colorofstate ac (reader x y) reader x y with
	  Kernel.Uni (r,g,b) -> (r,g,b)
	| _ -> (0,0,0)
    )
    else (fun x y ->
      let h = Hashtbl.create ac.Kernel.cardinality in
      let v = reader x y in
      try Hashtbl.find h v with Not_found -> (
	let res =
	  match Kernel.colorofstate ac (reader x y) reader x y with
	      Kernel.Uni (r,g,b) -> (r,g,b)
	    | _ -> (0,0,0)
	in Hashtbl.add h v res; res
      )
    )
  in
    for x=0 to w-1 do
      for y=0 to h-1 do
	let (r,g,b) = get_color x y in
	ba.{h-1-y,x} <- Int32.add (Int32.shift_left (Int32.of_int r) 16) (Int32.of_int  (g lsl 8 + b))
      done;
    done;
    debugline Dbgdev "Writing to file...";
    Cairo_png.surface_write_to_file (Cairo_bigarray.of_bigarr_32 ~alpha:false ba) filename

