open Debug

type event =
    Newinit | Newneno | Switchgrid
  | Stop | Stopstart | Onestep | Nothing | Memorize | Load | Help | Unmark
  | Askexit | Output | Newboth | Saverule | Newcolormap | Nextcolormap | Screenshot | Zoomout
  | Zoomat of int*int | Accel | Slow | Redraw | Resize of int*int  | Shortcut of string
  | Fullscreen | Centerzoom | Licence | Forcedexit | Revert_from_file
  | Zone of (Kernel.actionsource * ((int*int)*(int*int)))
  | ZoneDiag of (Kernel.actionsource * ((int*int)*(int*int)))
  | ZoneSide of (Kernel.actionsource * ((int*int)*(int*int)))
  (* | Open of string  *)
  | Builtin of string | Undo | Revert | SaveAs of string | Randomparams
  | Newhazard | HistoryClear | AddCache | Switchhistory | SwitchPlugin of (Kernel.one_plug*bool)
  | Measurement of (int -> int -> int -> (int -> int -> int) -> (int -> int -> int -> unit) -> unit)
  | Newconfsize of (int*int)

let (clear_scenario, addto_scenario, play_scenario) =
  let scenario = (Queue.create (): event Queue.t) in
  (fun () -> Queue.clear scenario),
  (fun x -> Queue.push x scenario),
  (fun player -> Queue.iter player scenario)

let reserved = [
  (Sdlkey.KEY_a, "Re-generate, keeping rule, but maximizing space dimension wrt window size"); 
  (Sdlkey.KEY_d, "Re-generate, keeping rule, but with medium space dimension wrt window size");
  (Sdlkey.KEY_p, "Re-generate, keeping rule, but with small space dimension wrt window size")
]

let keymap = [
		   (Sdlkey.KEY_s,(Saverule, "Save current rule and initial conf"));
		   (Sdlkey.KEY_r,(Revert_from_file, "Revert from file (if current object comes from an external file)"));
		   (Sdlkey.KEY_g,(Screenshot, "Trigger a snapshot of current configuration"));
		   (Sdlkey.KEY_c,(Nextcolormap, "Change colormap (cycle among existing ones)"));
		   (Sdlkey.KEY_m,(Memorize, "Memorize current configuration"));
		   (Sdlkey.KEY_l,(Load, "Load a memorized configuration (if any)"));
		   (Sdlkey.KEY_b,(Newboth, "Re-generate everything including rule and initial configuration"));
		   (Sdlkey.KEY_u,(Undo, "Revert to previous AC and/or configuration"));
		   (Sdlkey.KEY_t,(Revert, "Revert to initial configuration"));
		   (Sdlkey.KEY_o,(Newneno, "Re-generate object but keep initial configuration and current rule"));
		   (Sdlkey.KEY_i,(Newinit, "Re-generate object but keep current rule"));
		   (Sdlkey.KEY_y,(Randomparams, "Set parameters randomly within admissible bounds"));
		   (Sdlkey.KEY_j,(Newhazard, "Re-init random seeds in a system dependent way"));
		   (Sdlkey.KEY_k,(Switchgrid, "Show/hide the grid"));
		   (Sdlkey.KEY_h,(Help, "Show this help"));
		   (Sdlkey.KEY_DOLLAR,(Licence, "Licence/copyright info"));
		   (Sdlkey.KEY_e,(Zoomout, "Zoom out"));
		   (Sdlkey.KEY_z,(Centerzoom, "Zoom in center"));
		   (Sdlkey.KEY_v,(Unmark, "Unmark all cells"));
		   (Sdlkey.KEY_f,(Fullscreen, "Switch between fullscreen and non-fullscrenn modes"));
		   (Sdlkey.KEY_q,(Accel, "Decrease screen updating rate"));
		   (Sdlkey.KEY_w,(Slow, "Increase screen updating rate"));
		   (Sdlkey.KEY_SPACE ,(Stopstart, "Start/stop execution"));
		   (Sdlkey.KEY_n,(Onestep, "Execute one step"));
		   (Sdlkey.KEY_ESCAPE,(Askexit, "Exit"));
		   (Sdlkey.KEY_x,(HistoryClear, "Clear CA history"));
		 ]

(** Check for multiple key binding *)
let check_multiple () =
  let rec doit = function
      [] -> ()
    | t::q -> if List.mem t q then debugline Dbgwarning ("several bindings for key \""^(Sdlkey.name t)^"\"")
      else doit q in
    doit ((List.map fst keymap) @ (List.map fst reserved))
    
let keymaptext = ref ""

let generate_help () =
  let mixmap = reserved @ (List.map (fun (a,(_,b)) -> (a,b)) keymap) in
  let lmax = List.fold_left 
    (fun acc (x,_) ->  let sz = Sdlkey.name x in max acc (String.length sz))
    0 mixmap in
    keymaptext := (List.fold_left
      (fun acc (k,t) -> let sz = String.length (Sdlkey.name k) in
	 (acc^"\n"^(String.make (lmax-sz) ' ')^(Sdlkey.name k)^"  --->  "^t))
      ""
      mixmap)^"\n";
    debugline Dbgdev "Key bindings help generated"
    
