(* This example illustrates the very basic way of using the ACML API.
   It defines the elementary cellular automaton whose local rule is
   the sum modulo 2 of the cell itself and its left and right
   neighbours (rule 150).

*)


(* An ACml source file is nothing else than a valid OCaml source file
   using the ACML API. *)
open Acml

(* The only thing that ACML will keep from your source at the end is
   not an Acml object but a function transforming a given Acml object
   into a new one. This is the key concept. I repeat: the API doesn't
   want an object but a function transforming an object into the
   desired one.  
   So let's define a function! *)
let transform a =

  (* At this point we don't know anything about what's in the Acml
     object 'a'. We are going to specify what is important for us and
     forget about the rest. The way to specify things is mostly
     imperative. *)

  (* First we specify dimension, number of states and
     neighbourhood.  *)
  set_dimension a 1;
  set_cardinality a 2;
  set_moore_neighbourhood a 1;

  (* Then we set the local rule, which is a function taking an array
     of int and returning an int. The array contains the value of each
     neighbour in some order (but the order doesn't matter for this
     example). *)

  set_rule a (fun x -> Array.fold_left (+) 0 x mod 2)


(* Finally, we register our transformation function to the API.
*)
let _ = valid transform

(* That's all! *)

