(* This example shows how to define new types of initial configurations and new ways to modify the current configuration.
*)

open Acml

let transform a =  

  (* We start from the 'addition modulo' rule. *)
  let n = param a "Number of states" 2 50 3  in
    set_cardinality a n;
    set_dimension a (param a "Dimension" 1 2 1);
    let r = param a "Radius" 1 50 1 in
      (match 
	 choice_param a "neighborhood" [ "Moore"; "von Neumann"]  
       with 
	   "von Neumann" -> set_vonneumann_neighbourhood a r
	 | _ -> set_moore_neighbourhood a r);   
      set_rule a (fun x -> (Array.fold_left (+) 0 x) mod n);

      (* Now we add a 'everywhere zero' initial configuration. This configuration type will be selectable through the user interface. *)
      add_init_conf a "Everywhere zero" (Uniform 0);

      (* Of course, we can use parameters in initial configurations. *)
      let s = param a "sparsity" 2 50 10 in
	add_init_conf a "Sparse configuration" (General (fun x y -> if x mod s = 0 && y mod s = 0 then (n-1) else 0));

	(* Finally, we add what's called a 'modif' (i.e. modification method) in the API: that's a function which is called when a zone is selected with the mouse and which can read and modify the content of selected cells. Such a modif method will be selectable via the user interface. Naturally, parameters can be used in modif methods. *)
	let p = param a "period" 1 50 10 in
	  add_modif a "Diagonal periodic" (SimpleZone (fun c x y -> ((x+y) mod p) mod n))

let _ = valid transform
